import { TypeMap } from '../type';
import Node, { SourceRange, NLAnnotationMap, AnnotationMap, AnnotationSpec } from './base';
import NodeVisitor from './visitor';
import { DeviceSelector } from './expression';
import { Expression } from './expression2';
import { AbstractSlot, OldSlot } from './slots';
import { Statement } from './statement';
import { Program } from './program';
import SchemaRetriever from '../schema';
import { TokenStream } from '../new-syntax/tokenstream';
/**
 * A single example (primitive template) in a ThingTalk dataset
 *
 */
export declare class Example extends Node {
    isExample: boolean;
    id: number;
    type: string;
    args: TypeMap;
    value: Expression;
    utterances: string[];
    preprocessed: string[];
    annotations: AnnotationMap;
    /**
     * Construct a new example.
     *
     * @param location - the position of this node in the source code
     * @param id - the ID of the example, or -1 if the example has no ID
     * @param {string} type - the type of this example, one of `stream`, `query`,
     *        `action`, or `program`
     * @param {Ast.Stream|Ast.Table|Ast.Action|Ast.Program} - the code this example
     *        maps to
     * @param {string[]} utterances - raw, unprocessed utterances for this example
     * @param {string[]} preprocessed - preprocessed (tokenized) utterances for this example
     * @param {Object.<string, any>} annotations - other annotations for this example
     */
    constructor(location: SourceRange | null, id: number, type: string, args: TypeMap, value: Expression, utterances: string[], preprocessed: string[], annotations: AnnotationMap);
    toSource(): TokenStream;
    visit(visitor: NodeVisitor): void;
    clone(): Example;
    /**
     * Typecheck this example.
     *
     * This method can be used to typecheck an example is isolation,
     * outside of a ThingTalk program. This is useful to typecheck a dataset
     * and discard examples that do not typecheck without failing the whole dataset.
     *
     * @param schemas - schema retriever object to retrieve Thingpedia information
     * @param [getMeta=false] - retrieve natural language metadata during typecheck
     */
    typecheck(schemas: SchemaRetriever, getMeta?: boolean): Promise<this>;
    /**
     * Iterate all slots (scalar value nodes) in this example.
     * @deprecated Use {@link Ast.Example.iterateSlots2} instead.
     */
    iterateSlots(): Generator<OldSlot, void>;
    /**
     * Iterate all slots (scalar value nodes) in this example.
     */
    iterateSlots2(): Generator<DeviceSelector | AbstractSlot, void>;
    /**
     * Convert a dataset example to a program.
     *
     * This will create a program that invokes the same code as the example value,
     * and will replace all parameters with slots.
     *
     * @return {Ast.Program} the new program
     */
    toProgram(): Program;
}
/**
 * A statement that declares a ThingTalk dataset (collection of primitive
 * templates).
 *
 */
export declare class Dataset extends Statement {
    name: string;
    examples: Example[];
    nl_annotations: NLAnnotationMap;
    impl_annotations: AnnotationMap;
    /**
     * Construct a new dataset.
     *
     * @param location - the position of this node in the source code
     * @param name - the name of this dataset
     * @param language - the language code of this dataset, as 2 letter ISO code
     * @param examples - the examples in this dataset
     * @param [annotations={}]- dataset annotations
     */
    constructor(location: SourceRange | null, name: string, examples: Example[], annotations?: AnnotationSpec);
    toSource(): TokenStream;
    get language(): string | undefined;
    visit(visitor: NodeVisitor): void;
    iterateSlots(): Generator<OldSlot, void>;
    iterateSlots2(): Generator<DeviceSelector | AbstractSlot, void>;
    optimize(): Dataset;
    clone(): Dataset;
}
