"use strict";
// -*- mode: typescript; indent-tabs-mode: nil; js-basic-offset: 4 -*-
//
// This file is part of ThingTalk
//
// Copyright 2018-2021 The Board of Trustees of the Leland Stanford Junior University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: Giovanni Campagna <gcampagn@cs.stanford.edu>
//         Silei Xu <silei@cs.stanford.edu>
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Dataset = exports.Example = void 0;
const assert_1 = __importDefault(require("assert"));
const base_1 = __importStar(require("./base"));
const values_1 = require("./values");
const expression2_1 = require("./expression2");
const statement_1 = require("./statement");
const Optimizer = __importStar(require("../optimize"));
const typecheck_1 = __importDefault(require("../typecheck"));
const declaration_like_to_program_1 = __importDefault(require("./declaration_like_to_program"));
const list_1 = __importDefault(require("../utils/list"));
/**
 * A single example (primitive template) in a ThingTalk dataset
 *
 */
class Example extends base_1.default {
    /**
     * Construct a new example.
     *
     * @param location - the position of this node in the source code
     * @param id - the ID of the example, or -1 if the example has no ID
     * @param {string} type - the type of this example, one of `stream`, `query`,
     *        `action`, or `program`
     * @param {Ast.Stream|Ast.Table|Ast.Action|Ast.Program} - the code this example
     *        maps to
     * @param {string[]} utterances - raw, unprocessed utterances for this example
     * @param {string[]} preprocessed - preprocessed (tokenized) utterances for this example
     * @param {Object.<string, any>} annotations - other annotations for this example
     */
    constructor(location, id, type, args, value, utterances, preprocessed, annotations) {
        super(location);
        this.isExample = true;
        assert_1.default(typeof id === 'number');
        this.id = id;
        this.type = type;
        assert_1.default(typeof args === 'object');
        this.args = args;
        assert_1.default(value instanceof expression2_1.Expression);
        this.value = value;
        assert_1.default(Array.isArray(utterances) && Array.isArray(preprocessed));
        this.utterances = utterances;
        this.preprocessed = preprocessed;
        assert_1.default(typeof annotations === 'object');
        this.annotations = annotations;
    }
    toSource() {
        const annotations = {};
        if (this.id >= 0)
            annotations.id = new values_1.Value.Number(this.id);
        Object.assign(annotations, this.annotations);
        const metadata = {};
        if (this.utterances.length > 0)
            metadata.utterances = this.utterances;
        if (this.preprocessed.length > 0)
            metadata.preprocessed = this.preprocessed;
        let args = list_1.default.Nil;
        let first = true;
        for (const argname in this.args) {
            const argtype = this.args[argname];
            if (first)
                first = false;
            else
                args = list_1.default.concat(args, ',');
            args = list_1.default.concat(args, argname, ':', argtype.toSource());
        }
        let list = list_1.default.singleton(this.type);
        if (args !== list_1.default.Nil)
            list = list_1.default.concat(list, ' ', '(', args, ')');
        list = list_1.default.concat(list, ' ', '=', ' ', this.value.toSource(), base_1.nlAnnotationsToSource(metadata), base_1.implAnnotationsToSource(annotations), ';');
        return list;
    }
    visit(visitor) {
        visitor.enter(this);
        if (visitor.visitExample(this))
            this.value.visit(visitor);
        visitor.exit(this);
    }
    clone() {
        return new Example(this.location, this.id, this.type, Object.assign({}, this.args), this.value.clone(), this.utterances.slice(0), this.preprocessed.slice(0), Object.assign({}, this.annotations));
    }
    /**
     * Typecheck this example.
     *
     * This method can be used to typecheck an example is isolation,
     * outside of a ThingTalk program. This is useful to typecheck a dataset
     * and discard examples that do not typecheck without failing the whole dataset.
     *
     * @param schemas - schema retriever object to retrieve Thingpedia information
     * @param [getMeta=false] - retrieve natural language metadata during typecheck
     */
    async typecheck(schemas, getMeta = false) {
        const typeChecker = new typecheck_1.default(schemas, getMeta);
        await typeChecker.typeCheckExample(this);
        return this;
    }
    /**
     * Iterate all slots (scalar value nodes) in this example.
     * @deprecated Use {@link Ast.Example.iterateSlots2} instead.
     */
    *iterateSlots() {
        yield* this.value.iterateSlots({});
    }
    /**
     * Iterate all slots (scalar value nodes) in this example.
     */
    *iterateSlots2() {
        yield* this.value.iterateSlots2({});
    }
    /**
     * Convert a dataset example to a program.
     *
     * This will create a program that invokes the same code as the example value,
     * and will replace all parameters with slots.
     *
     * @return {Ast.Program} the new program
     */
    toProgram() {
        return declaration_like_to_program_1.default(this);
    }
}
exports.Example = Example;
/**
 * A statement that declares a ThingTalk dataset (collection of primitive
 * templates).
 *
 */
class Dataset extends statement_1.Statement {
    /**
     * Construct a new dataset.
     *
     * @param location - the position of this node in the source code
     * @param name - the name of this dataset
     * @param language - the language code of this dataset, as 2 letter ISO code
     * @param examples - the examples in this dataset
     * @param [annotations={}]- dataset annotations
     */
    constructor(location, name, examples, annotations = {}) {
        super(location);
        assert_1.default(typeof name === 'string');
        assert_1.default(!name.startsWith('@'));
        this.name = name;
        assert_1.default(Array.isArray(examples)); // of Example
        this.examples = examples;
        this.impl_annotations = annotations.impl || {};
        this.nl_annotations = annotations.nl || {};
    }
    toSource() {
        let list = list_1.default.concat('dataset', '@' + this.name, base_1.nlAnnotationsToSource(this.nl_annotations), base_1.implAnnotationsToSource(this.impl_annotations), ' ', '{', '\n', '\t+');
        let first = true;
        for (const ex of this.examples) {
            // force an additional \n between examples
            if (first)
                first = false;
            else
                list = list_1.default.concat(list, '\n');
            list = list_1.default.concat(list, ex.toSource(), '\n');
        }
        list = list_1.default.concat(list, '\t-', '}');
        return list;
    }
    get language() {
        const language = this.impl_annotations.language;
        if (language)
            return String(language.toJS());
        else
            return undefined;
    }
    visit(visitor) {
        visitor.enter(this);
        if (visitor.visitDataset(this)) {
            for (const example of this.examples)
                example.visit(visitor);
        }
        visitor.exit(this);
    }
    *iterateSlots() {
        for (const ex of this.examples)
            yield* ex.iterateSlots();
    }
    *iterateSlots2() {
        for (const ex of this.examples)
            yield* ex.iterateSlots2();
    }
    optimize() {
        return Optimizer.optimizeDataset(this);
    }
    clone() {
        const newMetadata = {};
        Object.assign(newMetadata, this.nl_annotations);
        const newAnnotations = {};
        Object.assign(newAnnotations, this.impl_annotations);
        return new Dataset(this.location, this.name, this.examples.map((e) => e.clone()), { nl: newMetadata, impl: newAnnotations });
    }
}
exports.Dataset = Dataset;
//# sourceMappingURL=dataset.js.map