import interpolate from 'string-interp';
import { LocationLike } from '../builtin/values';
import { FormattedObjectSpec, FormattedObject } from './format_objects';
import * as I18n from '../i18n';
import type SchemaRetriever from '../schema';
/**
 * Namespace for format objects.
 *
 * Classes in this namespace are not accessible directly, but objects
 * of this classes are returned by {@link Formatter} methods.
 *
 * @name FormatObjects
 * @namespace
 */
declare type PlainObject = {
    [key: string]: unknown;
};
interface FunctionSpec {
    type: 'code';
    code: string;
}
interface TextSpec {
    type: 'text';
    text: string;
}
declare type FormatFunction = (value: PlainObject, hint: string, formatter: Formatter) => FormatSpecChunk | FormatSpecChunk[];
declare type FormatSpecChunk = string | FormattedObjectSpec | TextSpec | FunctionSpec | FormatFunction;
export declare type FormatSpec = FormatSpecChunk[];
declare type FormattedChunk = string | FormattedObject;
/**
 * An object that is able to convert structured ThingTalk results
 * into something suitable for display to the user.
 *
 * @extends FormatUtils
 */
declare class Formatter extends interpolate.Formatter {
    private _schemas;
    private _interp;
    private _;
    /**
     * Construct a new formatter.
     *
     * @param {string} locale - the user's locale, as a BCP47 tag
     * @param {string} timezone - the user's timezone, as a string in the IANA timezone database (e.g. America/Los_Angeles, Europe/Rome)
     * @param {SchemaRetriever} schemaRetriever - the interface to access Thingpedia for formatting information
     * @param {Gettext} [gettext] - gettext instance; this is optional if {@link I18n.init} has been called
     */
    constructor(locale: string, timezone: string, schemaRetriever: SchemaRetriever, gettext?: I18n.Gettext);
    private _displayKey;
    private _displayKeyPart;
    _replaceInString(str: unknown, argMap: PlainObject): string | null;
    private _getFunctionDef;
    private _getArgImportance;
    private _formatFallbackValue;
    private _formatFallbackObject;
    private _formatFallback;
    private _formatAggregation;
    formatForType(outputType: string, outputValue: PlainObject, hint: 'string'): Promise<string>;
    formatForType(outputType: string, outputValue: PlainObject, hint: 'messages'): Promise<FormattedChunk[]>;
    formatForType(outputType: string, outputValue: PlainObject, hint: string): Promise<string | FormattedChunk[]>;
    private _toFormatObjects;
    format(formatspec: FormatSpecChunk[], argMap: PlainObject, hint: 'string'): string;
    format(formatspec: FormatSpecChunk[], argMap: PlainObject, hint: 'messages'): FormattedChunk[];
    format(formatspec: FormatSpecChunk[], argMap: PlainObject, hint: string): string | FormattedChunk[];
    private _normalize;
    private _applyHint;
    anyToString(o: unknown): string;
    /**
     * Convert a location to a string.
     *
     * @param {Builtin.Location} loc - the location to display
     * @return {string} the formatted location
     * @deprecated Use {@link Builtin.Location#toLocaleString} instead.
     */
    locationToString(loc: LocationLike): string;
}
export default Formatter;
declare type FormatFunction2 = FormatFunction;
declare type FormatSpecChunk2 = FormatSpecChunk;
declare type FormatSpec2 = FormatSpec;
declare type FormattedChunk2 = FormattedChunk;
declare namespace Formatter {
    type FormatFunction = FormatFunction2;
    type FormatSpecChunk = FormatSpecChunk2;
    type FormatSpec = FormatSpec2;
    type FormattedChunk = FormattedChunk2;
}
