"use strict";
// -*- mode: typescript; indent-tabs-mode: nil; js-basic-offset: 4 -*-
//
// This file is part of ThingTalk
//
// Copyright 2020 The Board of Trustees of the Leland Stanford Junior University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: Giovanni Campagna <gcampagn@cs.stanford.edu>
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = __importDefault(require("assert"));
const Ast = __importStar(require("../ast"));
const type_1 = __importDefault(require("../type"));
const keywords_1 = require("./keywords");
const date_utils_1 = require("../utils/date_utils");
function makeInput($, statements, annotations) {
    const classes = [];
    const datasets = [];
    const declarations = [];
    const executable = [];
    for (const stmt of statements) {
        if (stmt instanceof Ast.ClassDef) {
            classes.push(stmt);
        }
        else if (stmt instanceof Ast.Dataset) {
            datasets.push(stmt);
        }
        else if (stmt instanceof Ast.FunctionDeclaration) {
            declarations.push(stmt);
        }
        else {
            assert_1.default(stmt instanceof Ast.ExpressionStatement || stmt instanceof Ast.Assignment);
            executable.push(stmt);
        }
    }
    if (datasets.length > 0) {
        if (declarations.length > 0 || executable.length > 0)
            $.error(`Cannot mix dataset with ThingTalk statements other than class`);
        return new Ast.Library($.location, classes, datasets);
    }
    if (classes.length > 0 && declarations.length === 0 && executable.length === 0)
        return new Ast.Library($.location, classes, datasets);
    return new Ast.Program($.location, classes, declarations, executable, annotations);
}
function makeClassDef($, name, _extends, members, annotations, is_abstract) {
    const queries = {};
    const actions = {};
    const imports = [];
    const entities = [];
    for (const ast of members) {
        if (ast instanceof Ast.FunctionDef) {
            switch (ast.functionType) {
                case 'query':
                    queries[ast.name] = ast;
                    break;
                case 'action':
                    actions[ast.name] = ast;
                    break;
            }
        }
        else if (ast instanceof Ast.MixinImportStmt) {
            imports.push(ast);
        }
        else if (ast instanceof Ast.EntityDef) {
            entities.push(ast);
        }
    }
    const options = { is_abstract: !!is_abstract };
    return new Ast.ClassDef($.location, name, _extends, { queries, actions, imports, entities }, annotations, options);
}
function splitStatements(stmts) {
    const declarations = [];
    const executable = [];
    for (const stmt of stmts) {
        if (stmt instanceof Ast.FunctionDeclaration)
            declarations.push(stmt);
        else
            executable.push(stmt);
    }
    return [declarations, executable];
}
function isIdentifier(name) {
    return /^[A-Za-z_][A-Za-z0-9_]*$/.test(name) && !keywords_1.KEYWORDS.has(name) && !keywords_1.FORBIDDEN_KEYWORDS.has(name);
}
function checkExampleAnnotations($, annotations) {
    let id = -1;
    const utterances = [];
    const preprocessed = [];
    const other = {};
    for (const annotation of annotations) {
        if (annotation.type === 'impl') {
            if (annotation.name === 'id') {
                if (!(annotation.value instanceof Ast.NumberValue))
                    return $.error(`Example id annotation must be a number`);
                id = annotation.value.toJS();
            }
            else {
                other[annotation.name] = annotation.value;
            }
        }
        else {
            if (!(annotation.value instanceof Ast.ArrayValue) && !(annotation.value instanceof Ast.StringValue))
                return $.error(`An array is expected for ${annotation.name} annotation`);
            if (annotation.value instanceof Ast.StringValue)
                annotation.value = new Ast.ArrayValue([annotation.value]);
            switch (annotation.name) {
                case 'utterances':
                    utterances.push(...annotation.value.toJS());
                    break;
                case 'preprocessed':
                    preprocessed.push(...annotation.value.toJS());
                    break;
                default:
                    return $.error(`Invalid example annotation ${annotation.name}, expected utterances or preprocessed`);
            }
        }
    }
    return [id, utterances, preprocessed, other];
}
// split annotations between nl and impl
function splitAnnotations(annotations) {
    const nl_annotations = {};
    const impl_annotations = {};
    for (const a of annotations) {
        if (a.type === 'nl')
            nl_annotations[a.name] = a.value.toJS();
        else if (a.type === 'impl')
            impl_annotations[a.name] = a.value;
        else
            throw new TypeError();
    }
    return { nl: nl_annotations, impl: impl_annotations };
}
function parseIncompleteDate($, params, time) {
    const parts = [];
    // params is the raw sequence of numbers and commas
    // (which is unfortunately too messy to model as SLR grammar)
    // convert it to a sequence of possibly empty parameters
    let hadNumber = false;
    for (const p of params) {
        if (p === ',') {
            if (hadNumber)
                hadNumber = false;
            else
                parts.push(null);
        }
        else {
            if (hadNumber)
                return $.error(`Expected , between numbers in Date constructor`);
            else
                parts.push(p);
            hadNumber = true;
        }
    }
    if (parts.length > 6 || (time && parts.length > 3))
        return $.error(`Too many values in Date constructor`);
    if (parts.length === 4) // 4 would be year month day hour without minute or second, which is not allowed
        return $.error(`Wrong number of values in Date constructor (must be 1, 2, 3, 5, or 6)`);
    if (parts.length === 0 || parts.every((p) => p === null))
        return $.error(`Not enough values in Date constructor`);
    if (parts[0] !== null) {
        // absolute date, or incomplete date where the incomplete parts
        // are all to the right of the specified parts
        // (these are exactly equivalent because the incomplete bits are
        // set to 1 or 0)
        const d = new Date;
        if (parts.length === 1 && !time) {
            let unix_or_year = parts[0];
            // treat it as year between 10000 BC and 3000 AD
            // (in Unix, that would be a 10 seconds before midnight 31/12/1969
            // to 3 seconds later - we can leave with that window of time not
            // covered properly...)
            if (unix_or_year >= -10000 && unix_or_year < 3000) {
                if (unix_or_year >= 0 && unix_or_year < 100) { // then assume 1950-2050
                    if (unix_or_year >= 50)
                        unix_or_year += 1900;
                    else
                        unix_or_year += 2000;
                }
                d.setFullYear(unix_or_year);
                d.setMonth(0, 1);
                d.setHours(0, 0, 0, 0);
            }
            else {
                d.setTime(unix_or_year);
            }
        }
        else {
            let year = parts[0];
            if (year >= 0 && year < 100) { // then assume 1950-2050
                if (year >= 50)
                    year += 1900;
                else
                    year += 2000;
            }
            d.setFullYear(year);
            // set both the month and the date at the same time
            // otherwise when today's date is the 31st and the chosen
            // month has 30 days, the Date will be adjusted to the
            // first day of the subsequent month, which is wrong
            d.setMonth((parts[1] || 1) - 1, parts[2] || 1);
            if (time)
                d.setHours(time.hour, time.minute, time.second, 0);
            else
                d.setHours(parts[3] || 0, parts[4] || 0, parts[5] || 0, 0);
        }
        return new Ast.Value.Date(d);
    }
    else {
        const year = parts[0] || null;
        const month = parts[1] || null;
        const day = parts[2] || null;
        if (parts.length > 3)
            time = new Ast.Time.Absolute(parts[3] || 0, parts[4] || 0, parts[5] || 0);
        return new Ast.Value.Date(new Ast.DatePiece(year, month, day, time));
    }
}
function checkDialogueHistoryAnnotationSpec($, spec) {
    if (!spec.impl)
        return spec;
    if (spec.impl.confirm)
        $.error(`Duplicate #[confirm] annotation`);
    if (spec.impl.results || spec.impl.more || spec.impl.count || spec.impl.error)
        $.error(`Invalid dialogue history item annotation`);
    return spec;
}
const TERMINAL_IDS = { " 0PAD": 0, " 1EOF": 1, " 2START": 2, "!": 3, "#[": 4, "#_[": 5, "$?": 6, "$answer": 7, "$choice": 8, "$context": 9, "$debug": 10, "$dialogue": 11, "$end_of": 12, "$failed": 13, "$help": 14, "$location": 15, "$nevermind": 16, "$no": 17, "$now": 18, "$policy": 19, "$program_id": 20, "$result": 21, "$source": 22, "$start_of": 23, "$stop": 24, "$time": 25, "$train": 26, "$type": 27, "$undefined": 28, "$wakeup": 29, "$yes": 30, "%": 31, "&&": 32, "(": 33, ")": 34, "*": 35, "**": 36, "+": 37, ",": 38, "-": 39, ".": 40, "/": 41, ":": 42, ":=": 43, ";": 44, "<=": 45, "=": 46, "==": 47, "=>": 48, "=~": 49, ">=": 50, "Any": 51, "ArgMap": 52, "Array": 53, "Boolean": 54, "CLASS_OR_FUNCTION_REF": 55, "CURRENCY": 56, "Compound": 57, "Currency": 58, "DATE": 59, "DOLLARIDENTIFIER": 60, "DURATION": 61, "Date": 62, "EMAIL_ADDRESS": 63, "ENTITY_NAME": 64, "Entity": 65, "Enum": 66, "GENERIC_ENTITY": 67, "HASHTAG": 68, "IDENTIFIER": 69, "LOCATION": 70, "Location": 71, "MEASURE": 72, "Measure": 73, "NUMBER": 74, "Number": 75, "Object": 76, "PATH_NAME": 77, "PHONE_NUMBER": 78, "PICTURE": 79, "QUOTED_STRING": 80, "RecurrentTimeSpecification": 81, "SLOT": 82, "String": 83, "TIME": 84, "TYPE_ANNOT": 85, "Time": 86, "URL": 87, "USERNAME": 88, "[": 89, "]": 90, "abstract": 91, "action": 92, "aggregate": 93, "all": 94, "any": 95, "as": 96, "asc": 97, "avg": 98, "await": 99, "beginDate": 100, "beginTime": 101, "bookkeeping": 102, "break": 103, "case": 104, "catch": 105, "class": 106, "compute": 107, "confirm": 108, "const": 109, "contains": 110, "contains~": 111, "continue": 112, "count": 113, "dataset": 114, "dayOfWeek": 115, "debugger": 116, "delete": 117, "desc": 118, "distance": 119, "do": 120, "edge": 121, "endDate": 122, "endTime": 123, "ends_with": 124, "entity": 125, "enum": 126, "error": 127, "export": 128, "extends": 129, "false": 130, "filter": 131, "finally": 132, "for": 133, "frequency": 134, "from": 135, "function": 136, "if": 137, "implements": 138, "import": 139, "in": 140, "in_array": 141, "in_array~": 142, "instanceof": 143, "interface": 144, "interval": 145, "join": 146, "language": 147, "let": 148, "list": 149, "max": 150, "min": 151, "mixin": 152, "monitor": 153, "monitorable": 154, "more": 155, "new": 156, "notify": 157, "now": 158, "null": 159, "of": 160, "oninput": 161, "opt": 162, "out": 163, "package": 164, "prefix_of": 165, "private": 166, "program": 167, "protected": 168, "public": 169, "query": 170, "req": 171, "results": 172, "return": 173, "set_time": 174, "sort": 175, "starts_with": 176, "static": 177, "stream": 178, "subtract": 179, "suffix_of": 180, "sum": 181, "super": 182, "switch": 183, "this": 184, "throw": 185, "true": 186, "try": 187, "typeof": 188, "var": 189, "void": 190, "while": 191, "with": 192, "yield": 193, "{": 194, "||": 195, "}": 196, "~=": 197, "~contains": 198, "~in_array": 199 };
const RULE_NON_TERMINALS = [99, 99, 99, 99, 128, 128, 128, 128, 128, 128, 128, 128, 128, 58, 58, 58, 114, 114, 132, 132, 72, 72, 72, 72, 72, 72, 72, 72, 66, 66, 65, 65, 65, 65, 69, 69, 68, 67, 67, 67, 67, 67, 67, 70, 70, 71, 71, 109, 110, 110, 110, 111, 111, 111, 111, 111, 9, 7, 7, 8, 8, 130, 130, 135, 135, 134, 134, 134, 134, 134, 129, 129, 129, 129, 17, 25, 25, 23, 23, 23, 23, 22, 22, 24, 24, 27, 27, 92, 92, 92, 92, 92, 92, 92, 73, 73, 74, 74, 26, 26, 26, 26, 26, 26, 26, 26, 26, 63, 63, 64, 64, 62, 62, 62, 62, 139, 139, 140, 140, 138, 12, 12, 12, 60, 60, 60, 60, 59, 59, 79, 79, 18, 18, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 141, 141, 81, 81, 20, 133, 133, 21, 21, 115, 115, 115, 118, 118, 118, 116, 116, 117, 117, 82, 82, 82, 5, 5, 98, 98, 98, 112, 112, 112, 112, 112, 112, 112, 112, 112, 131, 131, 131, 96, 96, 97, 97, 95, 95, 127, 127, 108, 108, 6, 6, 143, 143, 143, 28, 28, 103, 103, 103, 102, 84, 84, 86, 86, 83, 83, 83, 87, 87, 87, 87, 85, 85, 88, 88, 88, 88, 88, 88, 88, 88, 88, 29, 29, 29, 29, 29, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 4, 4, 4, 4, 126, 126, 19, 19, 19, 19, 19, 19, 19, 113, 113, 113, 113, 113, 113, 113, 113, 113, 147, 119, 119, 146, 14, 14, 89, 89, 3, 3, 3, 101, 101, 101, 101, 80, 80, 125, 125, 125, 144, 144, 77, 77, 77, 77, 56, 16, 16, 107, 107, 107, 106, 106, 51, 124, 124, 123, 122, 122, 121, 121, 121, 121, 121, 121, 121, 121, 121, 55, 55, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 15, 15, 34, 34, 11, 52, 100, 10, 10, 49, 49, 49, 50, 35, 35, 48, 48, 48, 48, 48, 2, 2, 2, 2, 53, 53, 36, 36, 36, 61, 61, 61, 61, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 47, 76, 76, 46, 46, 46, 46, 45, 40, 44, 42, 38, 41, 43, 37, 39, 33, 33, 13, 13, 136, 136, 137, 137, 137, 137, 137, 137, 137, 137, 137, 137, 137, 137, 137, 137, 137, 137, 32, 32, 31, 31, 30, 30, 75, 75, 93, 93, 57, 57, 57, 57, 57, 104, 104, 104, 104, 104, 104, 104, 104, 104, 104, 104, 104, 104, 104, 104, 104, 104, 104, 104, 148, 148, 148, 148, 148, 148, 105, 105, 105, 105, 105, 91, 91, 91, 91, 91, 91, 142, 142, 120, 120, 145, 145, 145, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 0];
const ARITY = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 5, 5, 1, 2, 1, 3, 4, 7, 3, 6, 5, 8, 4, 7, 1, 2, 2, 3, 7, 8, 1, 2, 1, 5, 5, 5, 5, 5, 5, 2, 3, 1, 3, 7, 3, 3, 3, 1, 3, 1, 3, 3, 1, 5, 5, 1, 2, 1, 2, 1, 2, 1, 1, 1, 2, 1, 1, 2, 1, 3, 5, 1, 2, 3, 5, 4, 6, 2, 3, 1, 3, 1, 2, 1, 2, 3, 3, 2, 1, 2, 1, 1, 1, 3, 6, 3, 4, 5, 6, 4, 5, 6, 7, 2, 3, 1, 3, 2, 3, 3, 4, 2, 3, 1, 3, 2, 2, 2, 1, 3, 4, 5, 6, 2, 3, 1, 2, 1, 1, 6, 5, 5, 4, 6, 5, 5, 4, 6, 5, 5, 4, 6, 5, 5, 4, 8, 7, 7, 6, 6, 7, 1, 3, 1, 1, 3, 1, 3, 1, 5, 5, 1, 3, 1, 1, 3, 1, 3, 1, 3, 3, 1, 3, 1, 6, 4, 3, 6, 6, 6, 4, 4, 6, 2, 1, 2, 4, 5, 2, 3, 1, 3, 3, 3, 2, 2, 1, 3, 1, 3, 1, 1, 2, 3, 6, 1, 1, 2, 3, 3, 1, 1, 3, 1, 3, 3, 1, 3, 3, 3, 1, 3, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 6, 9, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 1, 3, 1, 1, 3, 1, 3, 3, 1, 3, 3, 3, 1, 3, 4, 4, 4, 1, 1, 1, 1, 1, 1, 4, 2, 3, 2, 3, 4, 3, 5, 5, 1, 3, 3, 1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 2, 1, 3, 9, 7, 5, 1, 9, 8, 7, 1, 3, 1, 4, 1, 1, 1, 2, 2, 1, 5, 5, 7, 5, 7, 11, 10, 9, 4, 4, 1, 4, 2, 5, 5, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 5, 2, 4, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 4, 4, 4, 4, 3, 1, 2, 1, 3, 2, 3, 1, 3, 1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2];
const GOTO = [{ "4": 44, "5": 39, "7": 23, "8": 22, "9": 21, "17": 31, "21": 35, "23": 27, "25": 82, "57": 51, "58": 3, "60": 28, "72": 5, "81": 30, "82": 37, "98": 40, "99": 1, "104": 56, "109": 4, "112": 41, "114": 2, "115": 36, "128": 8, "131": 48, "133": 33, "134": 26, "135": 20, "141": 29, "142": 47 }, {}, {}, {}, {}, {}, { "132": 85 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 44, "5": 39, "17": 31, "21": 35, "23": 27, "25": 82, "57": 51, "60": 28, "81": 30, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "134": 91, "141": 29, "142": 47 }, { "4": 44, "5": 39, "17": 31, "21": 35, "23": 27, "25": 82, "57": 51, "60": 28, "81": 30, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "134": 26, "135": 92, "141": 29, "142": 47 }, { "7": 93 }, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 94 }, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 113 }, {}, {}, {}, {}, {}, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 115 }, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "4": 146, "6": 207, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 126, "85": 128, "86": 125, "87": 127, "88": 129, "90": 145, "100": 202, "104": 56, "107": 135, "108": 120, "116": 123, "117": 124, "118": 121, "119": 143, "125": 132, "126": 147, "143": 208, "144": 133, "147": 137, "148": 144 }, {}, {}, {}, { "4": 44, "5": 39, "20": 217, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, {}, {}, {}, {}, { "96": 223 }, {}, { "96": 225 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "57": 51, "104": 56, "142": 227 }, {}, {}, {}, {}, { "7": 23, "8": 22, "9": 233 }, {}, { "1": 194, "2": 190, "4": 146, "6": 207, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 238, "85": 128, "87": 127, "88": 129, "90": 237, "100": 202, "104": 56, "107": 135, "108": 235, "119": 143, "125": 132, "126": 147, "143": 208, "144": 133, "147": 137, "148": 144 }, {}, { "100": 239 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 240, "147": 255, "148": 144 }, {}, { "4": 44, "5": 39, "17": 31, "21": 35, "23": 27, "25": 82, "57": 51, "60": 28, "81": 30, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "134": 91, "141": 29, "142": 47 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 44, "5": 39, "21": 35, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 264, "142": 47 }, { "4": 44, "5": 39, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 266, "131": 48, "142": 47 }, { "1": 194, "2": 190, "4": 146, "6": 207, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 238, "85": 128, "87": 127, "88": 129, "90": 237, "100": 202, "104": 56, "107": 135, "108": 267, "119": 143, "125": 132, "126": 147, "143": 208, "144": 133, "147": 137, "148": 144 }, { "1": 194, "2": 190, "4": 146, "6": 207, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 238, "85": 128, "87": 127, "88": 129, "90": 237, "100": 202, "104": 56, "107": 135, "108": 268, "119": 143, "125": 132, "126": 147, "143": 208, "144": 133, "147": 137, "148": 144 }, {}, {}, {}, {}, {}, {}, { "29": 278 }, {}, {}, {}, { "1": 194, "2": 190, "4": 146, "6": 207, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 290, "84": 288, "85": 128, "86": 292, "87": 127, "88": 129, "90": 237, "100": 202, "104": 56, "107": 135, "108": 289, "119": 291, "125": 132, "126": 147, "143": 208, "144": 133, "147": 137, "148": 144 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "15": 299, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 300, "147": 255, "148": 144 }, { "4": 98, "57": 96, "90": 97, "104": 56, "106": 302, "119": 303, "126": 99, "148": 144 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 397, "57": 395, "90": 396, "91": 312, "94": 311, "104": 56, "126": 398 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "145": 408 }, {}, { "100": 412 }, {}, {}, {}, {}, {}, {}, {}, { "19": 414, "90": 416 }, {}, {}, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 419 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "15": 421, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 420, "147": 255, "148": 144 }, {}, {}, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "127": 424, "144": 250, "146": 425, "147": 255, "148": 144 }, { "4": 98, "57": 96, "90": 97, "104": 56, "119": 426, "126": 99, "148": 144 }, { "4": 429, "5": 39, "20": 428, "21": 218, "57": 432, "82": 37, "90": 97, "98": 40, "104": 56, "112": 41, "115": 36, "119": 427, "126": 99, "131": 48, "142": 47, "148": 144 }, { "4": 438, "5": 39, "20": 433, "21": 218, "57": 437, "82": 37, "90": 97, "98": 40, "104": 56, "112": 41, "115": 36, "119": 435, "120": 434, "126": 99, "131": 48, "142": 47, "148": 144 }, {}, { "4": 98, "57": 96, "90": 97, "95": 442, "97": 441, "104": 56, "126": 99, "148": 444 }, {}, { "4": 397, "57": 395, "90": 396, "91": 446, "104": 56, "126": 398 }, { "139": 447 }, { "7": 23, "8": 22, "9": 450, "59": 449 }, { "7": 23, "8": 22, "9": 455, "22": 453 }, {}, { "4": 44, "5": 39, "21": 35, "57": 51, "65": 458, "66": 457, "81": 459, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "142": 47 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "15": 460, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 300, "147": 255, "148": 144 }, {}, { "4": 397, "57": 395, "90": 396, "91": 462, "104": 56, "126": 398 }, {}, { "1": 194, "2": 190, "4": 146, "6": 207, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 126, "84": 288, "85": 128, "86": 292, "87": 127, "88": 129, "90": 145, "100": 202, "104": 56, "107": 135, "108": 289, "119": 464, "125": 132, "126": 147, "143": 208, "144": 133, "147": 137, "148": 144 }, {}, { "29": 278 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 475, "147": 255, "148": 144 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 476, "147": 255, "148": 144 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 477, "147": 255, "148": 144 }, { "4": 44, "5": 39, "20": 478, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, {}, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "4": 146, "6": 480, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 238, "85": 128, "87": 127, "88": 129, "90": 237, "100": 202, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "143": 208, "144": 133, "147": 137, "148": 144 }, {}, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 483, "85": 128, "86": 125, "87": 127, "88": 129, "90": 97, "100": 202, "104": 56, "107": 135, "116": 482, "117": 124, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 483, "85": 128, "86": 125, "87": 127, "88": 129, "90": 97, "100": 202, "104": 56, "107": 135, "117": 485, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 486 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "19": 489, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 491, "85": 128, "87": 127, "88": 129, "90": 237, "100": 202, "102": 488, "103": 487, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "85": 128, "87": 492, "88": 129, "90": 97, "100": 202, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "85": 128, "87": 493, "88": 129, "90": 97, "100": 202, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, { "1": 194, "2": 190, "3": 494, "4": 146, "10": 201, "11": 187, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 143, "125": 249, "126": 147, "144": 250, "147": 255, "148": 144 }, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "85": 496, "88": 129, "90": 97, "100": 202, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "85": 497, "88": 129, "90": 97, "100": 202, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "85": 498, "88": 129, "90": 97, "100": 202, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "88": 499, "90": 97, "100": 202, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, {}, {}, { "29": 278 }, {}, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 503 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 504, "147": 255, "148": 144 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "15": 505, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 300, "147": 255, "148": 144 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "15": 506, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 300, "147": 255, "148": 144 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "15": 507, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 300, "147": 255, "148": 144 }, {}, {}, {}, {}, {}, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "119": 513, "126": 99, "148": 144 }, {}, {}, {}, {}, {}, { "4": 397, "57": 395, "90": 396, "91": 312, "94": 519, "104": 56, "126": 398 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 520 }, {}, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 523 }, { "145": 524 }, { "145": 525 }, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 526 }, {}, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 238, "85": 128, "87": 127, "88": 129, "90": 237, "100": 202, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "143": 527, "144": 133, "147": 137, "148": 144 }, {}, { "1": 194, "2": 190, "4": 146, "6": 207, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 238, "85": 128, "87": 127, "88": 129, "90": 237, "100": 202, "104": 56, "107": 135, "108": 289, "119": 143, "125": 132, "126": 147, "143": 208, "144": 133, "147": 137, "148": 144 }, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "119": 528, "126": 99, "148": 144 }, { "4": 44, "5": 39, "20": 529, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, {}, {}, {}, {}, { "4": 44, "5": 39, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 266, "131": 48, "142": 47 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 397, "57": 395, "90": 396, "91": 545, "104": 56, "126": 398 }, { "96": 546 }, { "7": 23, "8": 22, "9": 548 }, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "138": 551, "140": 550, "148": 552 }, {}, { "59": 553 }, { "11": 554 }, { "78": 557, "79": 556 }, {}, { "24": 562 }, { "22": 564 }, { "26": 567, "27": 566, "92": 570 }, { "4": 44, "5": 39, "21": 35, "57": 51, "65": 575, "81": 459, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "142": 47 }, {}, { "67": 580, "68": 577, "69": 579 }, {}, { "4": 44, "5": 39, "21": 35, "57": 51, "65": 458, "66": 583, "81": 459, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "142": 47 }, {}, { "110": 584, "111": 586, "132": 589 }, {}, {}, {}, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 592, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 143, "125": 249, "126": 147, "144": 250, "147": 255, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "19": 489, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 491, "85": 128, "87": 127, "88": 129, "90": 237, "100": 202, "102": 488, "103": 593, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "90": 97, "100": 202, "101": 594, "104": 56, "107": 252, "113": 247, "119": 143, "125": 249, "126": 147, "144": 250, "147": 255, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "90": 97, "100": 202, "101": 595, "104": 56, "107": 252, "113": 247, "119": 143, "125": 249, "126": 147, "144": 250, "147": 255, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 596, "90": 97, "100": 202, "104": 56, "107": 252, "113": 247, "119": 143, "125": 249, "126": 147, "144": 250, "147": 255, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 597, "90": 97, "100": 202, "104": 56, "107": 252, "113": 247, "119": 143, "125": 249, "126": 147, "144": 250, "147": 255, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 598, "90": 97, "100": 202, "104": 56, "107": 252, "113": 247, "119": 143, "125": 249, "126": 147, "144": 250, "147": 255, "148": 144 }, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "90": 97, "100": 202, "104": 56, "107": 252, "113": 599, "119": 143, "125": 249, "126": 147, "144": 250, "147": 255, "148": 144 }, {}, {}, {}, {}, { "4": 44, "5": 39, "57": 51, "82": 604, "98": 40, "104": 56, "112": 41, "131": 48, "142": 47 }, {}, { "4": 44, "5": 39, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 605, "131": 48, "142": 47 }, {}, {}, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 483, "84": 288, "85": 128, "86": 292, "87": 127, "88": 129, "90": 97, "100": 202, "104": 56, "107": 135, "119": 464, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, {}, {}, {}, {}, {}, { "19": 606, "90": 416 }, { "29": 607 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 136, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 483, "85": 128, "86": 609, "87": 127, "88": 129, "90": 97, "100": 202, "104": 56, "107": 135, "119": 143, "125": 132, "126": 147, "144": 133, "147": 137, "148": 144 }, {}, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 614, "147": 255, "148": 144 }, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "119": 616, "126": 99, "148": 144 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 617, "147": 255, "148": 144 }, { "136": 618 }, { "4": 98, "13": 622, "57": 96, "90": 97, "104": 56, "126": 99, "148": 623 }, { "123": 625, "124": 624 }, { "10": 627, "100": 202 }, { "10": 632, "11": 629, "61": 630, "76": 631, "100": 202 }, { "10": 634, "11": 635, "100": 202 }, {}, {}, { "11": 637 }, { "11": 638 }, {}, {}, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 643, "147": 255, "148": 144 }, {}, { "4": 44, "5": 39, "20": 644, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, {}, {}, { "4": 44, "5": 39, "20": 645, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, { "4": 44, "5": 39, "20": 646, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, {}, {}, { "4": 44, "5": 39, "20": 647, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, { "4": 98, "57": 96, "90": 97, "104": 56, "119": 648, "126": 99, "148": 144 }, {}, { "4": 98, "57": 96, "90": 97, "95": 649, "104": 56, "126": 99, "148": 444 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 650, "147": 255, "148": 144 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 651, "147": 255, "148": 144 }, { "96": 652 }, {}, { "4": 44, "5": 39, "17": 657, "21": 35, "57": 51, "81": 656, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "129": 654, "130": 653, "131": 48, "133": 33, "141": 655, "142": 47 }, {}, {}, {}, {}, { "136": 662 }, {}, { "7": 23, "8": 22, "9": 664, "59": 663 }, {}, { "78": 666 }, {}, { "18": 668, "139": 667 }, { "18": 672, "139": 671 }, { "18": 674, "139": 673 }, { "18": 676, "139": 675 }, { "7": 23, "8": 22, "9": 678, "22": 677 }, {}, {}, {}, { "26": 681, "92": 570 }, {}, { "4": 397, "57": 395, "90": 396, "91": 683, "93": 682, "104": 56, "126": 398 }, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 684 }, { "4": 397, "57": 395, "90": 396, "91": 685, "104": 56, "126": 398 }, {}, {}, {}, {}, {}, {}, {}, { "4": 701, "90": 700, "104": 699, "105": 697, "126": 702 }, { "67": 703 }, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 705 }, { "7": 23, "8": 22, "9": 707 }, { "4": 44, "5": 39, "21": 35, "57": 51, "65": 575, "81": 459, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "142": 47 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "1": 194, "2": 190, "4": 146, "10": 201, "11": 187, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "90": 97, "100": 202, "104": 56, "107": 252, "113": 714, "119": 143, "125": 249, "126": 147, "144": 250, "147": 255, "148": 144 }, { "4": 44, "5": 39, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 715, "131": 48, "142": 47 }, {}, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 716, "147": 255, "148": 144 }, {}, {}, {}, {}, {}, {}, {}, {}, { "137": 719 }, {}, {}, {}, { "136": 738 }, {}, {}, { "121": 742, "122": 741 }, {}, {}, {}, { "10": 755, "100": 202 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 44, "5": 39, "17": 657, "21": 35, "57": 51, "81": 656, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "129": 768, "131": 48, "133": 33, "141": 655, "142": 47 }, {}, {}, {}, {}, { "4": 44, "5": 39, "20": 770, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, { "4": 44, "5": 39, "17": 657, "21": 35, "57": 51, "81": 656, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "129": 654, "130": 771, "131": 48, "133": 33, "141": 655, "142": 47 }, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "138": 772, "148": 552 }, {}, {}, { "59": 773 }, {}, {}, { "18": 774 }, { "4": 44, "5": 39, "20": 775, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, {}, {}, { "18": 776 }, { "4": 44, "5": 39, "20": 777, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, { "18": 778 }, { "4": 44, "5": 39, "20": 779, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, { "18": 780 }, { "4": 44, "5": 39, "20": 781, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, {}, { "22": 783 }, {}, {}, {}, {}, {}, { "7": 23, "8": 22, "9": 788 }, { "63": 790 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 701, "90": 700, "104": 699, "105": 697, "126": 702 }, {}, { "4": 44, "5": 39, "21": 35, "57": 51, "65": 458, "66": 802, "81": 459, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "142": 47 }, {}, {}, { "111": 805, "132": 589 }, { "111": 807, "132": 589 }, { "4": 397, "57": 395, "90": 396, "91": 462, "104": 56, "126": 398 }, { "1": 194, "2": 190, "4": 146, "6": 207, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 238, "85": 128, "87": 127, "88": 129, "90": 237, "100": 202, "104": 56, "107": 135, "108": 809, "119": 143, "125": 132, "126": 147, "143": 208, "144": 133, "147": 137, "148": 144 }, { "1": 194, "2": 190, "4": 146, "6": 207, "10": 201, "11": 187, "16": 136, "19": 210, "28": 209, "33": 173, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 131, "56": 134, "57": 96, "76": 176, "77": 138, "83": 238, "85": 128, "87": 127, "88": 129, "90": 237, "100": 202, "104": 56, "107": 135, "108": 810, "119": 143, "125": 132, "126": 147, "143": 208, "144": 133, "147": 137, "148": 144 }, {}, {}, {}, {}, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 814, "147": 255, "148": 144 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 98, "30": 821, "31": 820, "32": 819, "57": 96, "90": 97, "104": 56, "126": 99, "148": 822 }, {}, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 823 }, {}, {}, { "123": 824 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "10": 835, "100": 202 }, {}, {}, { "2": 836 }, {}, {}, { "2": 838, "10": 840, "50": 839, "100": 202 }, { "10": 841, "100": 202 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 44, "5": 39, "17": 657, "21": 35, "57": 51, "81": 656, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "129": 768, "131": 48, "133": 33, "141": 655, "142": 47 }, {}, {}, { "4": 44, "5": 39, "20": 844, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, { "7": 23, "8": 845 }, { "4": 44, "5": 39, "20": 847, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, { "7": 23, "8": 848 }, { "4": 44, "5": 39, "20": 850, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, { "7": 23, "8": 851 }, { "4": 44, "5": 39, "20": 853, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, { "7": 23, "8": 855 }, { "4": 44, "5": 39, "20": 857, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, {}, {}, {}, { "4": 397, "57": 395, "90": 396, "91": 859, "104": 56, "126": 398 }, {}, {}, { "4": 98, "57": 96, "73": 862, "74": 861, "90": 97, "104": 56, "126": 99, "148": 863 }, { "7": 23, "8": 22, "9": 866 }, { "4": 397, "57": 395, "90": 396, "91": 683, "93": 867, "104": 56, "126": 398 }, { "4": 98, "12": 872, "57": 96, "62": 870, "64": 869, "90": 97, "104": 56, "126": 99, "148": 871 }, {}, {}, { "76": 875 }, { "70": 876 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 878, "147": 255, "148": 144 }, {}, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 880, "147": 255, "148": 144 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 881, "147": 255, "148": 144 }, { "1": 194, "2": 190, "3": 244, "4": 146, "10": 201, "11": 187, "14": 241, "16": 253, "33": 173, "34": 257, "35": 157, "36": 159, "37": 168, "38": 165, "39": 169, "40": 163, "41": 166, "42": 164, "43": 170, "44": 167, "45": 162, "46": 161, "47": 171, "48": 158, "49": 155, "50": 156, "51": 172, "52": 154, "53": 160, "54": 258, "55": 254, "56": 251, "57": 96, "76": 176, "77": 256, "80": 246, "89": 243, "90": 97, "100": 202, "101": 245, "104": 56, "107": 252, "113": 247, "119": 242, "125": 249, "126": 147, "144": 250, "146": 882, "147": 255, "148": 144 }, { "4": 44, "5": 39, "21": 35, "57": 51, "65": 575, "81": 459, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "142": 47 }, { "4": 44, "5": 39, "21": 35, "57": 51, "65": 458, "66": 883, "81": 459, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "142": 47 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 44, "5": 39, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 884, "131": 48, "142": 47 }, {}, { "137": 885 }, {}, { "145": 887 }, { "4": 397, "57": 395, "75": 888, "90": 396, "91": 312, "94": 889, "104": 56, "126": 398 }, {}, {}, {}, { "136": 892 }, { "136": 893 }, {}, {}, { "121": 894 }, { "1": 895 }, { "1": 897 }, { "2": 898 }, { "2": 899 }, { "100": 900 }, { "10": 902, "49": 901, "100": 202 }, { "76": 903 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "7": 23, "8": 913 }, { "7": 93 }, {}, { "7": 23, "8": 916 }, { "7": 93 }, {}, { "7": 23, "8": 919 }, { "7": 93 }, {}, { "7": 23, "8": 922 }, { "4": 44, "5": 39, "20": 924, "21": 218, "57": 51, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "142": 47 }, { "7": 93 }, {}, {}, { "96": 927 }, {}, {}, { "7": 23, "8": 22, "9": 929 }, {}, {}, {}, {}, {}, { "63": 932 }, {}, {}, {}, { "136": 935 }, { "4": 98, "57": 96, "90": 97, "104": 56, "126": 99, "148": 936 }, {}, {}, {}, {}, { "71": 942, "107": 943 }, {}, {}, {}, {}, {}, { "4": 44, "5": 39, "21": 35, "57": 51, "65": 575, "81": 459, "82": 37, "98": 40, "104": 56, "112": 41, "115": 36, "131": 48, "133": 33, "142": 47 }, {}, {}, {}, {}, {}, {}, {}, { "4": 98, "30": 955, "57": 96, "90": 97, "104": 56, "126": 99, "148": 822 }, { "7": 23, "8": 22, "9": 956 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "145": 408 }, {}, {}, {}, { "10": 958, "100": 202 }, {}, {}, {}, { "10": 840, "50": 960, "100": 202 }, { "11": 961 }, {}, { "7": 93 }, {}, {}, { "7": 93 }, {}, {}, { "7": 93 }, {}, {}, { "7": 93 }, {}, {}, {}, { "7": 23, "8": 967 }, {}, {}, {}, { "4": 98, "57": 96, "73": 971, "90": 97, "104": 56, "126": 99, "148": 863 }, {}, { "7": 23, "8": 22, "9": 973 }, {}, { "4": 98, "12": 872, "57": 96, "62": 974, "90": 97, "104": 56, "126": 99, "148": 871 }, { "7": 23, "8": 22, "9": 975 }, { "136": 976 }, {}, {}, { "7": 23, "8": 22, "9": 978 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "4": 397, "57": 395, "90": 396, "91": 312, "94": 982, "104": 56, "126": 398 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, { "7": 23, "8": 988 }, { "7": 93 }, {}, {}, {}, {}, {}, {}, {}, {}, { "7": 23, "8": 22, "9": 992 }, {}, {}, {}, { "107": 994 }, {}, {}, { "10": 632, "11": 629, "61": 630, "76": 631, "100": 202 }, {}, { "10": 840, "50": 995, "100": 202 }, {}, {}, { "7": 93 }, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}];
const PARSER_ACTION = [{ "4": [1, 25], "5": [1, 24], "7": [1, 10], "8": [1, 9], "10": [1, 15], "11": [1, 6], "13": [1, 13], "14": [1, 18], "16": [1, 16], "17": [1, 12], "19": [1, 7], "24": [1, 17], "26": [1, 14], "29": [1, 19], "30": [1, 11], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "91": [1, 84], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "106": [1, 83], "108": [1, 53], "113": [1, 45], "114": [1, 81], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "136": [1, 80], "145": [1, 72], "147": [1, 60], "148": [1, 32], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "1": [0, 0] }, { "1": [2, 0] }, { "1": [2, 1] }, { "1": [2, 2] }, { "1": [2, 3] }, { "55": [1, 86] }, { "194": [1, 87] }, { "44": [1, 88] }, { "33": [1, 89] }, { "33": [1, 90] }, { "44": [2, 4] }, { "44": [2, 5] }, { "44": [2, 6] }, { "44": [2, 7] }, { "44": [2, 8] }, { "44": [2, 9] }, { "44": [2, 10] }, { "44": [2, 11] }, { "44": [2, 12] }, { "1": [2, 16], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "91": [1, 84], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "106": [1, 83], "108": [1, 53], "113": [1, 45], "114": [1, 81], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "136": [1, 80], "145": [1, 72], "147": [1, 60], "148": [1, 32], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "91": [1, 84], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "106": [1, 83], "108": [1, 53], "113": [1, 45], "114": [1, 81], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "136": [1, 80], "145": [1, 72], "147": [1, 60], "148": [1, 32], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [1, 25], "5": [1, 24], "33": [2, 56], "34": [2, 56], "38": [2, 56], "44": [2, 56], "55": [2, 56], "69": [2, 56], "89": [2, 56], "91": [2, 56], "92": [2, 56], "97": [2, 56], "98": [2, 56], "100": [2, 56], "101": [2, 56], "106": [2, 56], "108": [2, 56], "113": [2, 56], "114": [2, 56], "115": [2, 56], "118": [2, 56], "122": [2, 56], "123": [2, 56], "125": [2, 56], "127": [2, 56], "134": [2, 56], "135": [2, 56], "136": [2, 56], "145": [2, 56], "147": [2, 56], "148": [2, 56], "149": [2, 56], "150": [2, 56], "151": [2, 56], "153": [2, 56], "154": [2, 56], "155": [2, 56], "158": [2, 56], "167": [2, 56], "170": [2, 56], "172": [2, 56], "175": [2, 56], "178": [2, 56], "179": [2, 56], "181": [2, 56], "194": [2, 56], "196": [2, 56] }, { "4": [2, 59], "5": [2, 59], "33": [2, 59], "34": [2, 59], "38": [2, 59], "44": [2, 59], "55": [2, 59], "69": [2, 59], "89": [2, 59], "91": [2, 59], "92": [2, 59], "97": [2, 59], "98": [2, 59], "100": [2, 59], "101": [2, 59], "106": [2, 59], "108": [2, 59], "113": [2, 59], "114": [2, 59], "115": [2, 59], "118": [2, 59], "122": [2, 59], "123": [2, 59], "125": [2, 59], "127": [2, 59], "134": [2, 59], "135": [2, 59], "136": [2, 59], "145": [2, 59], "147": [2, 59], "148": [2, 59], "149": [2, 59], "150": [2, 59], "151": [2, 59], "153": [2, 59], "154": [2, 59], "155": [2, 59], "158": [2, 59], "167": [2, 59], "170": [2, 59], "172": [2, 59], "175": [2, 59], "178": [2, 59], "179": [2, 59], "181": [2, 59], "194": [2, 59], "196": [2, 59] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "1": [2, 63], "33": [2, 63], "55": [2, 63], "69": [2, 63], "89": [2, 63], "91": [2, 63], "92": [2, 63], "97": [2, 63], "98": [2, 63], "100": [2, 63], "101": [2, 63], "106": [2, 63], "108": [2, 63], "113": [2, 63], "114": [2, 63], "115": [2, 63], "118": [2, 63], "122": [2, 63], "123": [2, 63], "125": [2, 63], "127": [2, 63], "134": [2, 63], "135": [2, 63], "136": [2, 63], "145": [2, 63], "147": [2, 63], "148": [2, 63], "149": [2, 63], "150": [2, 63], "151": [2, 63], "153": [2, 63], "154": [2, 63], "155": [2, 63], "158": [2, 63], "167": [2, 63], "170": [2, 63], "172": [2, 63], "175": [2, 63], "178": [2, 63], "179": [2, 63], "181": [2, 63] }, { "1": [2, 65], "33": [2, 65], "55": [2, 65], "69": [2, 65], "89": [2, 65], "91": [2, 65], "92": [2, 65], "97": [2, 65], "98": [2, 65], "100": [2, 65], "101": [2, 65], "106": [2, 65], "108": [2, 65], "113": [2, 65], "114": [2, 65], "115": [2, 65], "118": [2, 65], "122": [2, 65], "123": [2, 65], "125": [2, 65], "127": [2, 65], "134": [2, 65], "135": [2, 65], "136": [2, 65], "145": [2, 65], "147": [2, 65], "148": [2, 65], "149": [2, 65], "150": [2, 65], "151": [2, 65], "153": [2, 65], "154": [2, 65], "155": [2, 65], "158": [2, 65], "167": [2, 65], "170": [2, 65], "172": [2, 65], "175": [2, 65], "178": [2, 65], "179": [2, 65], "181": [2, 65] }, { "1": [2, 66], "33": [2, 66], "55": [2, 66], "69": [2, 66], "89": [2, 66], "91": [2, 66], "92": [2, 66], "97": [2, 66], "98": [2, 66], "100": [2, 66], "101": [2, 66], "106": [2, 66], "108": [2, 66], "113": [2, 66], "114": [2, 66], "115": [2, 66], "118": [2, 66], "122": [2, 66], "123": [2, 66], "125": [2, 66], "127": [2, 66], "134": [2, 66], "135": [2, 66], "136": [2, 66], "145": [2, 66], "147": [2, 66], "148": [2, 66], "149": [2, 66], "150": [2, 66], "151": [2, 66], "153": [2, 66], "154": [2, 66], "155": [2, 66], "158": [2, 66], "167": [2, 66], "170": [2, 66], "172": [2, 66], "175": [2, 66], "178": [2, 66], "179": [2, 66], "181": [2, 66] }, { "1": [2, 67], "33": [2, 67], "55": [2, 67], "69": [2, 67], "89": [2, 67], "91": [2, 67], "92": [2, 67], "97": [2, 67], "98": [2, 67], "100": [2, 67], "101": [2, 67], "106": [2, 67], "108": [2, 67], "113": [2, 67], "114": [2, 67], "115": [2, 67], "118": [2, 67], "122": [2, 67], "123": [2, 67], "125": [2, 67], "127": [2, 67], "134": [2, 67], "135": [2, 67], "136": [2, 67], "145": [2, 67], "147": [2, 67], "148": [2, 67], "149": [2, 67], "150": [2, 67], "151": [2, 67], "153": [2, 67], "154": [2, 67], "155": [2, 67], "158": [2, 67], "167": [2, 67], "170": [2, 67], "172": [2, 67], "175": [2, 67], "178": [2, 67], "179": [2, 67], "181": [2, 67] }, { "44": [1, 114] }, { "1": [2, 69], "33": [2, 69], "55": [2, 69], "69": [2, 69], "89": [2, 69], "91": [2, 69], "92": [2, 69], "97": [2, 69], "98": [2, 69], "100": [2, 69], "101": [2, 69], "106": [2, 69], "108": [2, 69], "113": [2, 69], "114": [2, 69], "115": [2, 69], "118": [2, 69], "122": [2, 69], "123": [2, 69], "125": [2, 69], "127": [2, 69], "134": [2, 69], "135": [2, 69], "136": [2, 69], "145": [2, 69], "147": [2, 69], "148": [2, 69], "149": [2, 69], "150": [2, 69], "151": [2, 69], "153": [2, 69], "154": [2, 69], "155": [2, 69], "158": [2, 69], "167": [2, 69], "170": [2, 69], "172": [2, 69], "175": [2, 69], "178": [2, 69], "179": [2, 69], "181": [2, 69] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "4": [2, 155], "5": [2, 155], "44": [2, 155] }, { "48": [1, 116] }, { "4": [2, 158], "5": [2, 158], "44": [2, 158], "48": [1, 117] }, { "4": [2, 160], "5": [2, 160], "34": [2, 160], "44": [2, 160], "48": [2, 160], "196": [2, 160] }, { "4": [2, 162], "5": [2, 162], "34": [2, 162], "38": [1, 118], "44": [2, 162], "48": [2, 162], "131": [1, 119], "196": [2, 162] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 130], "35": [1, 122], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "4": [2, 172], "5": [2, 172], "34": [2, 172], "38": [2, 172], "44": [2, 172], "48": [2, 172], "96": [1, 215], "131": [2, 172], "196": [2, 172] }, { "4": [2, 175], "5": [2, 175], "34": [2, 175], "38": [2, 175], "44": [2, 175], "48": [2, 175], "96": [2, 175], "131": [2, 175], "196": [2, 175] }, { "4": [2, 177], "5": [2, 177], "34": [2, 177], "38": [2, 177], "44": [2, 177], "48": [2, 177], "89": [1, 216], "96": [2, 177], "131": [2, 177], "196": [2, 177] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "33": [1, 219] }, { "33": [1, 220] }, { "33": [1, 221] }, { "33": [1, 222] }, { "33": [1, 224] }, { "4": [2, 188], "5": [2, 188], "34": [2, 188], "38": [2, 188], "44": [2, 188], "48": [2, 188], "89": [2, 188], "96": [2, 188], "131": [2, 188], "196": [2, 188] }, { "33": [1, 224], "40": [1, 226] }, { "33": [2, 477] }, { "33": [2, 478] }, { "4": [2, 436], "5": [2, 436], "31": [2, 436], "32": [2, 436], "33": [2, 436], "34": [2, 436], "35": [2, 436], "36": [2, 436], "37": [2, 436], "38": [2, 436], "39": [2, 436], "40": [2, 436], "41": [2, 436], "42": [2, 436], "44": [2, 436], "45": [2, 436], "46": [2, 436], "47": [2, 436], "48": [2, 436], "49": [2, 436], "50": [2, 436], "85": [2, 436], "90": [2, 436], "96": [2, 436], "97": [2, 436], "118": [2, 436], "129": [2, 436], "131": [2, 436], "135": [2, 436], "160": [2, 436], "195": [2, 436], "196": [2, 436], "197": [2, 436] }, { "4": [2, 437], "5": [2, 437], "31": [2, 437], "32": [2, 437], "33": [2, 437], "34": [2, 437], "35": [2, 437], "36": [2, 437], "37": [2, 437], "38": [2, 437], "39": [2, 437], "40": [2, 437], "41": [2, 437], "42": [2, 437], "44": [2, 437], "45": [2, 437], "46": [2, 437], "47": [2, 437], "48": [2, 437], "49": [2, 437], "50": [2, 437], "85": [2, 437], "90": [2, 437], "96": [2, 437], "97": [2, 437], "118": [2, 437], "129": [2, 437], "131": [2, 437], "135": [2, 437], "160": [2, 437], "195": [2, 437], "196": [2, 437], "197": [2, 437] }, { "4": [2, 438], "5": [2, 438], "31": [2, 438], "32": [2, 438], "33": [2, 438], "34": [2, 438], "35": [2, 438], "36": [2, 438], "37": [2, 438], "38": [2, 438], "39": [2, 438], "40": [2, 438], "41": [2, 438], "42": [2, 438], "44": [2, 438], "45": [2, 438], "46": [2, 438], "47": [2, 438], "48": [2, 438], "49": [2, 438], "50": [2, 438], "85": [2, 438], "90": [2, 438], "96": [2, 438], "97": [2, 438], "118": [2, 438], "129": [2, 438], "131": [2, 438], "135": [2, 438], "160": [2, 438], "195": [2, 438], "196": [2, 438], "197": [2, 438] }, { "4": [2, 439], "5": [2, 439], "31": [2, 439], "32": [2, 439], "33": [2, 439], "34": [2, 439], "35": [2, 439], "36": [2, 439], "37": [2, 439], "38": [2, 439], "39": [2, 439], "40": [2, 439], "41": [2, 439], "42": [2, 439], "44": [2, 439], "45": [2, 439], "46": [2, 439], "47": [2, 439], "48": [2, 439], "49": [2, 439], "50": [2, 439], "85": [2, 439], "90": [2, 439], "96": [2, 439], "97": [2, 439], "118": [2, 439], "129": [2, 439], "131": [2, 439], "135": [2, 439], "160": [2, 439], "195": [2, 439], "196": [2, 439], "197": [2, 439] }, { "4": [2, 440], "5": [2, 440], "31": [2, 440], "32": [2, 440], "33": [2, 440], "34": [2, 440], "35": [2, 440], "36": [2, 440], "37": [2, 440], "38": [2, 440], "39": [2, 440], "40": [2, 440], "41": [2, 440], "42": [2, 440], "44": [2, 440], "45": [2, 440], "46": [2, 440], "47": [2, 440], "48": [2, 440], "49": [2, 440], "50": [2, 440], "85": [2, 440], "90": [2, 440], "96": [2, 440], "97": [2, 440], "118": [2, 440], "129": [2, 440], "131": [2, 440], "135": [2, 440], "160": [2, 440], "195": [2, 440], "196": [2, 440], "197": [2, 440] }, { "4": [2, 441], "5": [2, 441], "31": [2, 441], "32": [2, 441], "33": [2, 441], "34": [2, 441], "35": [2, 441], "36": [2, 441], "37": [2, 441], "38": [2, 441], "39": [2, 441], "40": [2, 441], "41": [2, 441], "42": [2, 441], "44": [2, 441], "45": [2, 441], "46": [2, 441], "47": [2, 441], "48": [2, 441], "49": [2, 441], "50": [2, 441], "85": [2, 441], "90": [2, 441], "96": [2, 441], "97": [2, 441], "118": [2, 441], "129": [2, 441], "131": [2, 441], "135": [2, 441], "160": [2, 441], "195": [2, 441], "196": [2, 441], "197": [2, 441] }, { "4": [2, 442], "5": [2, 442], "31": [2, 442], "32": [2, 442], "33": [2, 442], "34": [2, 442], "35": [2, 442], "36": [2, 442], "37": [2, 442], "38": [2, 442], "39": [2, 442], "40": [2, 442], "41": [2, 442], "42": [2, 442], "44": [2, 442], "45": [2, 442], "46": [2, 442], "47": [2, 442], "48": [2, 442], "49": [2, 442], "50": [2, 442], "85": [2, 442], "90": [2, 442], "96": [2, 442], "97": [2, 442], "118": [2, 442], "129": [2, 442], "131": [2, 442], "135": [2, 442], "160": [2, 442], "195": [2, 442], "196": [2, 442], "197": [2, 442] }, { "4": [2, 443], "5": [2, 443], "31": [2, 443], "32": [2, 443], "33": [2, 443], "34": [2, 443], "35": [2, 443], "36": [2, 443], "37": [2, 443], "38": [2, 443], "39": [2, 443], "40": [2, 443], "41": [2, 443], "42": [2, 443], "44": [2, 443], "45": [2, 443], "46": [2, 443], "47": [2, 443], "48": [2, 443], "49": [2, 443], "50": [2, 443], "85": [2, 443], "90": [2, 443], "96": [2, 443], "97": [2, 443], "118": [2, 443], "129": [2, 443], "131": [2, 443], "135": [2, 443], "160": [2, 443], "195": [2, 443], "196": [2, 443], "197": [2, 443] }, { "4": [2, 444], "5": [2, 444], "31": [2, 444], "32": [2, 444], "33": [2, 444], "34": [2, 444], "35": [2, 444], "36": [2, 444], "37": [2, 444], "38": [2, 444], "39": [2, 444], "40": [2, 444], "41": [2, 444], "42": [2, 444], "44": [2, 444], "45": [2, 444], "46": [2, 444], "47": [2, 444], "48": [2, 444], "49": [2, 444], "50": [2, 444], "85": [2, 444], "90": [2, 444], "96": [2, 444], "97": [2, 444], "118": [2, 444], "129": [2, 444], "131": [2, 444], "135": [2, 444], "160": [2, 444], "195": [2, 444], "196": [2, 444], "197": [2, 444] }, { "4": [2, 445], "5": [2, 445], "31": [2, 445], "32": [2, 445], "33": [2, 445], "34": [2, 445], "35": [2, 445], "36": [2, 445], "37": [2, 445], "38": [2, 445], "39": [2, 445], "40": [2, 445], "41": [2, 445], "42": [2, 445], "44": [2, 445], "45": [2, 445], "46": [2, 445], "47": [2, 445], "48": [2, 445], "49": [2, 445], "50": [2, 445], "85": [2, 445], "90": [2, 445], "96": [2, 445], "97": [2, 445], "118": [2, 445], "129": [2, 445], "131": [2, 445], "135": [2, 445], "160": [2, 445], "195": [2, 445], "196": [2, 445], "197": [2, 445] }, { "4": [2, 446], "5": [2, 446], "31": [2, 446], "32": [2, 446], "33": [2, 446], "34": [2, 446], "35": [2, 446], "36": [2, 446], "37": [2, 446], "38": [2, 446], "39": [2, 446], "40": [2, 446], "41": [2, 446], "42": [2, 446], "44": [2, 446], "45": [2, 446], "46": [2, 446], "47": [2, 446], "48": [2, 446], "49": [2, 446], "50": [2, 446], "85": [2, 446], "90": [2, 446], "96": [2, 446], "97": [2, 446], "118": [2, 446], "129": [2, 446], "131": [2, 446], "135": [2, 446], "160": [2, 446], "195": [2, 446], "196": [2, 446], "197": [2, 446] }, { "4": [2, 447], "5": [2, 447], "31": [2, 447], "32": [2, 447], "33": [2, 447], "34": [2, 447], "35": [2, 447], "36": [2, 447], "37": [2, 447], "38": [2, 447], "39": [2, 447], "40": [2, 447], "41": [2, 447], "42": [2, 447], "44": [2, 447], "45": [2, 447], "46": [2, 447], "47": [2, 447], "48": [2, 447], "49": [2, 447], "50": [2, 447], "85": [2, 447], "90": [2, 447], "96": [2, 447], "97": [2, 447], "118": [2, 447], "129": [2, 447], "131": [2, 447], "135": [2, 447], "160": [2, 447], "195": [2, 447], "196": [2, 447], "197": [2, 447] }, { "4": [2, 448], "5": [2, 448], "31": [2, 448], "32": [2, 448], "33": [2, 448], "34": [2, 448], "35": [2, 448], "36": [2, 448], "37": [2, 448], "38": [2, 448], "39": [2, 448], "40": [2, 448], "41": [2, 448], "42": [2, 448], "44": [2, 448], "45": [2, 448], "46": [2, 448], "47": [2, 448], "48": [2, 448], "49": [2, 448], "50": [2, 448], "85": [2, 448], "90": [2, 448], "96": [2, 448], "97": [2, 448], "118": [2, 448], "129": [2, 448], "131": [2, 448], "135": [2, 448], "160": [2, 448], "195": [2, 448], "196": [2, 448], "197": [2, 448] }, { "4": [2, 449], "5": [2, 449], "31": [2, 449], "32": [2, 449], "33": [2, 449], "34": [2, 449], "35": [2, 449], "36": [2, 449], "37": [2, 449], "38": [2, 449], "39": [2, 449], "40": [2, 449], "41": [2, 449], "42": [2, 449], "44": [2, 449], "45": [2, 449], "46": [2, 449], "47": [2, 449], "48": [2, 449], "49": [2, 449], "50": [2, 449], "85": [2, 449], "90": [2, 449], "96": [2, 449], "97": [2, 449], "118": [2, 449], "129": [2, 449], "131": [2, 449], "135": [2, 449], "160": [2, 449], "195": [2, 449], "196": [2, 449], "197": [2, 449] }, { "4": [2, 450], "5": [2, 450], "31": [2, 450], "32": [2, 450], "33": [2, 450], "34": [2, 450], "35": [2, 450], "36": [2, 450], "37": [2, 450], "38": [2, 450], "39": [2, 450], "40": [2, 450], "41": [2, 450], "42": [2, 450], "44": [2, 450], "45": [2, 450], "46": [2, 450], "47": [2, 450], "48": [2, 450], "49": [2, 450], "50": [2, 450], "85": [2, 450], "90": [2, 450], "96": [2, 450], "97": [2, 450], "118": [2, 450], "129": [2, 450], "131": [2, 450], "135": [2, 450], "160": [2, 450], "195": [2, 450], "196": [2, 450], "197": [2, 450] }, { "4": [2, 451], "5": [2, 451], "31": [2, 451], "32": [2, 451], "33": [2, 451], "34": [2, 451], "35": [2, 451], "36": [2, 451], "37": [2, 451], "38": [2, 451], "39": [2, 451], "40": [2, 451], "41": [2, 451], "42": [2, 451], "44": [2, 451], "45": [2, 451], "46": [2, 451], "47": [2, 451], "48": [2, 451], "49": [2, 451], "50": [2, 451], "85": [2, 451], "90": [2, 451], "96": [2, 451], "97": [2, 451], "118": [2, 451], "129": [2, 451], "131": [2, 451], "135": [2, 451], "160": [2, 451], "195": [2, 451], "196": [2, 451], "197": [2, 451] }, { "4": [2, 452], "5": [2, 452], "31": [2, 452], "32": [2, 452], "33": [2, 452], "34": [2, 452], "35": [2, 452], "36": [2, 452], "37": [2, 452], "38": [2, 452], "39": [2, 452], "40": [2, 452], "41": [2, 452], "42": [2, 452], "44": [2, 452], "45": [2, 452], "46": [2, 452], "47": [2, 452], "48": [2, 452], "49": [2, 452], "50": [2, 452], "85": [2, 452], "90": [2, 452], "96": [2, 452], "97": [2, 452], "118": [2, 452], "129": [2, 452], "131": [2, 452], "135": [2, 452], "160": [2, 452], "195": [2, 452], "196": [2, 452], "197": [2, 452] }, { "4": [2, 453], "5": [2, 453], "31": [2, 453], "32": [2, 453], "33": [2, 453], "34": [2, 453], "35": [2, 453], "36": [2, 453], "37": [2, 453], "38": [2, 453], "39": [2, 453], "40": [2, 453], "41": [2, 453], "42": [2, 453], "44": [2, 453], "45": [2, 453], "46": [2, 453], "47": [2, 453], "48": [2, 453], "49": [2, 453], "50": [2, 453], "85": [2, 453], "90": [2, 453], "96": [2, 453], "97": [2, 453], "118": [2, 453], "129": [2, 453], "131": [2, 453], "135": [2, 453], "160": [2, 453], "195": [2, 453], "196": [2, 453], "197": [2, 453] }, { "4": [2, 454], "5": [2, 454], "31": [2, 454], "32": [2, 454], "33": [2, 454], "34": [2, 454], "35": [2, 454], "36": [2, 454], "37": [2, 454], "38": [2, 454], "39": [2, 454], "40": [2, 454], "41": [2, 454], "42": [2, 454], "44": [2, 454], "45": [2, 454], "46": [2, 454], "47": [2, 454], "48": [2, 454], "49": [2, 454], "50": [2, 454], "85": [2, 454], "90": [2, 454], "96": [2, 454], "97": [2, 454], "118": [2, 454], "129": [2, 454], "131": [2, 454], "135": [2, 454], "160": [2, 454], "195": [2, 454], "196": [2, 454], "197": [2, 454] }, { "4": [2, 455], "5": [2, 455], "31": [2, 455], "32": [2, 455], "33": [2, 455], "34": [2, 455], "35": [2, 455], "36": [2, 455], "37": [2, 455], "38": [2, 455], "39": [2, 455], "40": [2, 455], "41": [2, 455], "42": [2, 455], "44": [2, 455], "45": [2, 455], "46": [2, 455], "47": [2, 455], "48": [2, 455], "49": [2, 455], "50": [2, 455], "85": [2, 455], "90": [2, 455], "96": [2, 455], "97": [2, 455], "118": [2, 455], "129": [2, 455], "131": [2, 455], "135": [2, 455], "160": [2, 455], "195": [2, 455], "196": [2, 455], "197": [2, 455] }, { "4": [2, 456], "5": [2, 456], "31": [2, 456], "32": [2, 456], "33": [2, 456], "34": [2, 456], "35": [2, 456], "36": [2, 456], "37": [2, 456], "38": [2, 456], "39": [2, 456], "40": [2, 456], "41": [2, 456], "42": [2, 456], "44": [2, 456], "45": [2, 456], "46": [2, 456], "47": [2, 456], "48": [2, 456], "49": [2, 456], "50": [2, 456], "85": [2, 456], "90": [2, 456], "96": [2, 456], "97": [2, 456], "118": [2, 456], "129": [2, 456], "131": [2, 456], "135": [2, 456], "160": [2, 456], "195": [2, 456], "196": [2, 456], "197": [2, 456] }, { "4": [2, 457], "5": [2, 457], "31": [2, 457], "32": [2, 457], "33": [2, 457], "34": [2, 457], "35": [2, 457], "36": [2, 457], "37": [2, 457], "38": [2, 457], "39": [2, 457], "40": [2, 457], "41": [2, 457], "42": [2, 457], "44": [2, 457], "45": [2, 457], "46": [2, 457], "47": [2, 457], "48": [2, 457], "49": [2, 457], "50": [2, 457], "85": [2, 457], "90": [2, 457], "96": [2, 457], "97": [2, 457], "118": [2, 457], "129": [2, 457], "131": [2, 457], "135": [2, 457], "160": [2, 457], "195": [2, 457], "196": [2, 457], "197": [2, 457] }, { "4": [2, 458], "5": [2, 458], "31": [2, 458], "32": [2, 458], "33": [2, 458], "34": [2, 458], "35": [2, 458], "36": [2, 458], "37": [2, 458], "38": [2, 458], "39": [2, 458], "40": [2, 458], "41": [2, 458], "42": [2, 458], "44": [2, 458], "45": [2, 458], "46": [2, 458], "47": [2, 458], "48": [2, 458], "49": [2, 458], "50": [2, 458], "85": [2, 458], "90": [2, 458], "96": [2, 458], "97": [2, 458], "118": [2, 458], "129": [2, 458], "131": [2, 458], "135": [2, 458], "160": [2, 458], "195": [2, 458], "196": [2, 458], "197": [2, 458] }, { "4": [2, 459], "5": [2, 459], "31": [2, 459], "32": [2, 459], "33": [2, 459], "34": [2, 459], "35": [2, 459], "36": [2, 459], "37": [2, 459], "38": [2, 459], "39": [2, 459], "40": [2, 459], "41": [2, 459], "42": [2, 459], "44": [2, 459], "45": [2, 459], "46": [2, 459], "47": [2, 459], "48": [2, 459], "49": [2, 459], "50": [2, 459], "85": [2, 459], "90": [2, 459], "96": [2, 459], "97": [2, 459], "118": [2, 459], "129": [2, 459], "131": [2, 459], "135": [2, 459], "160": [2, 459], "195": [2, 459], "196": [2, 459], "197": [2, 459] }, { "4": [2, 250], "5": [2, 250], "31": [2, 250], "32": [2, 250], "33": [2, 250], "34": [2, 250], "35": [2, 250], "36": [2, 250], "37": [2, 250], "38": [2, 250], "39": [2, 250], "40": [2, 250], "41": [2, 250], "42": [2, 250], "44": [2, 250], "45": [2, 250], "46": [2, 250], "47": [2, 250], "48": [2, 250], "49": [2, 250], "50": [2, 250], "85": [2, 250], "90": [2, 250], "96": [2, 250], "97": [2, 250], "118": [2, 250], "129": [2, 250], "131": [2, 250], "135": [2, 250], "160": [2, 250], "195": [2, 250], "196": [2, 250], "197": [2, 250] }, { "4": [2, 251], "5": [2, 251], "31": [2, 251], "32": [2, 251], "33": [2, 251], "34": [2, 251], "35": [2, 251], "36": [2, 251], "37": [2, 251], "38": [2, 251], "39": [2, 251], "40": [2, 251], "41": [2, 251], "42": [2, 251], "44": [2, 251], "45": [2, 251], "46": [2, 251], "47": [2, 251], "48": [2, 251], "49": [2, 251], "50": [2, 251], "85": [2, 251], "90": [2, 251], "96": [2, 251], "97": [2, 251], "118": [2, 251], "129": [2, 251], "131": [2, 251], "135": [2, 251], "160": [2, 251], "195": [2, 251], "196": [2, 251], "197": [2, 251] }, { "4": [2, 252], "5": [2, 252], "31": [2, 252], "32": [2, 252], "33": [2, 252], "34": [2, 252], "35": [2, 252], "36": [2, 252], "37": [2, 252], "38": [2, 252], "39": [2, 252], "40": [2, 252], "41": [2, 252], "42": [2, 252], "44": [2, 252], "45": [2, 252], "46": [2, 252], "47": [2, 252], "48": [2, 252], "49": [2, 252], "50": [2, 252], "85": [2, 252], "90": [2, 252], "96": [2, 252], "97": [2, 252], "118": [2, 252], "129": [2, 252], "131": [2, 252], "135": [2, 252], "160": [2, 252], "195": [2, 252], "196": [2, 252], "197": [2, 252] }, { "4": [2, 253], "5": [2, 253], "31": [2, 253], "32": [2, 253], "33": [2, 253], "34": [2, 253], "35": [2, 253], "36": [2, 253], "37": [2, 253], "38": [2, 253], "39": [2, 253], "40": [2, 253], "41": [2, 253], "42": [2, 253], "44": [2, 253], "45": [2, 253], "46": [2, 253], "47": [2, 253], "48": [2, 253], "49": [2, 253], "50": [2, 253], "85": [2, 253], "90": [2, 253], "96": [2, 253], "97": [2, 253], "118": [2, 253], "129": [2, 253], "131": [2, 253], "135": [2, 253], "160": [2, 253], "195": [2, 253], "196": [2, 253], "197": [2, 253] }, { "69": [1, 50], "92": [1, 57], "97": [1, 74], "100": [1, 66], "101": [1, 67], "108": [1, 53], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "178": [1, 65], "179": [1, 73] }, { "55": [1, 228] }, { "55": [1, 229] }, { "55": [2, 75] }, { "106": [1, 230] }, { "4": [1, 25], "5": [1, 24], "33": [1, 232], "44": [1, 231] }, { "4": [2, 18], "5": [2, 18], "33": [2, 18], "38": [2, 18], "40": [1, 234], "44": [2, 18], "48": [2, 18], "131": [2, 18] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "1": [2, 13] }, { "74": [1, 204] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "1": [2, 64], "33": [2, 64], "55": [2, 64], "69": [2, 64], "89": [2, 64], "91": [2, 64], "92": [2, 64], "97": [2, 64], "98": [2, 64], "100": [2, 64], "101": [2, 64], "106": [2, 64], "108": [2, 64], "113": [2, 64], "114": [2, 64], "115": [2, 64], "118": [2, 64], "122": [2, 64], "123": [2, 64], "125": [2, 64], "127": [2, 64], "134": [2, 64], "135": [2, 64], "136": [2, 64], "145": [2, 64], "147": [2, 64], "148": [2, 64], "149": [2, 64], "150": [2, 64], "151": [2, 64], "153": [2, 64], "154": [2, 64], "155": [2, 64], "158": [2, 64], "167": [2, 64], "170": [2, 64], "172": [2, 64], "175": [2, 64], "178": [2, 64], "179": [2, 64], "181": [2, 64] }, { "1": [2, 17], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "91": [1, 84], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "106": [1, 83], "108": [1, 53], "113": [1, 45], "114": [1, 81], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "136": [1, 80], "145": [1, 72], "147": [1, 60], "148": [1, 32], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [2, 60], "5": [2, 60], "33": [2, 60], "34": [2, 60], "38": [2, 60], "44": [2, 60], "55": [2, 60], "69": [2, 60], "89": [2, 60], "91": [2, 60], "92": [2, 60], "97": [2, 60], "98": [2, 60], "100": [2, 60], "101": [2, 60], "106": [2, 60], "108": [2, 60], "113": [2, 60], "114": [2, 60], "115": [2, 60], "118": [2, 60], "122": [2, 60], "123": [2, 60], "125": [2, 60], "127": [2, 60], "134": [2, 60], "135": [2, 60], "136": [2, 60], "145": [2, 60], "147": [2, 60], "148": [2, 60], "149": [2, 60], "150": [2, 60], "151": [2, 60], "153": [2, 60], "154": [2, 60], "155": [2, 60], "158": [2, 60], "167": [2, 60], "170": [2, 60], "172": [2, 60], "175": [2, 60], "178": [2, 60], "179": [2, 60], "181": [2, 60], "194": [2, 60], "196": [2, 60] }, { "46": [1, 261] }, { "4": [2, 460], "5": [2, 460], "31": [2, 460], "32": [2, 460], "34": [2, 460], "35": [2, 460], "36": [2, 460], "37": [2, 460], "38": [2, 460], "39": [2, 460], "40": [2, 460], "41": [2, 460], "42": [2, 460], "44": [2, 460], "45": [2, 460], "46": [2, 460], "47": [2, 460], "48": [2, 460], "49": [2, 460], "50": [2, 460], "85": [2, 460], "90": [2, 460], "96": [2, 460], "97": [2, 460], "118": [2, 460], "129": [2, 460], "131": [2, 460], "160": [2, 460], "195": [2, 460], "196": [2, 460], "197": [2, 460] }, { "4": [2, 461], "5": [2, 461], "31": [2, 461], "32": [2, 461], "34": [2, 461], "35": [2, 461], "36": [2, 461], "37": [2, 461], "38": [2, 461], "39": [2, 461], "40": [2, 461], "41": [2, 461], "42": [2, 461], "44": [2, 461], "45": [2, 461], "46": [2, 461], "47": [2, 461], "48": [2, 461], "49": [2, 461], "50": [2, 461], "85": [2, 461], "90": [2, 461], "96": [2, 461], "97": [2, 461], "118": [2, 461], "129": [2, 461], "131": [2, 461], "160": [2, 461], "195": [2, 461], "196": [2, 461], "197": [2, 461] }, { "4": [2, 462], "5": [2, 462], "31": [2, 462], "32": [2, 462], "34": [2, 462], "35": [2, 462], "36": [2, 462], "37": [2, 462], "38": [2, 462], "39": [2, 462], "40": [2, 462], "41": [2, 462], "42": [2, 462], "44": [2, 462], "45": [2, 462], "46": [2, 462], "47": [2, 462], "48": [2, 462], "49": [2, 462], "50": [2, 462], "85": [2, 462], "90": [2, 462], "96": [2, 462], "97": [2, 462], "118": [2, 462], "129": [2, 462], "131": [2, 462], "160": [2, 462], "195": [2, 462], "196": [2, 462], "197": [2, 462] }, { "4": [2, 463], "5": [2, 463], "31": [2, 463], "32": [2, 463], "34": [2, 463], "35": [2, 463], "36": [2, 463], "37": [2, 463], "38": [2, 463], "39": [2, 463], "40": [2, 463], "41": [2, 463], "42": [2, 463], "44": [2, 463], "45": [2, 463], "46": [2, 463], "47": [2, 463], "48": [2, 463], "49": [2, 463], "50": [2, 463], "85": [2, 463], "90": [2, 463], "96": [2, 463], "97": [2, 463], "118": [2, 463], "129": [2, 463], "131": [2, 463], "160": [2, 463], "195": [2, 463], "196": [2, 463], "197": [2, 463] }, { "4": [2, 464], "5": [2, 464], "31": [2, 464], "32": [2, 464], "34": [2, 464], "35": [2, 464], "36": [2, 464], "37": [2, 464], "38": [2, 464], "39": [2, 464], "40": [2, 464], "41": [2, 464], "42": [2, 464], "44": [2, 464], "45": [2, 464], "46": [2, 464], "47": [2, 464], "48": [2, 464], "49": [2, 464], "50": [2, 464], "85": [2, 464], "90": [2, 464], "96": [2, 464], "97": [2, 464], "118": [2, 464], "129": [2, 464], "131": [2, 464], "160": [2, 464], "195": [2, 464], "196": [2, 464], "197": [2, 464] }, { "4": [2, 465], "5": [2, 465], "31": [2, 465], "32": [2, 465], "34": [2, 465], "35": [2, 465], "36": [2, 465], "37": [2, 465], "38": [2, 465], "39": [2, 465], "40": [2, 465], "41": [2, 465], "42": [2, 465], "44": [2, 465], "45": [2, 465], "46": [2, 465], "47": [2, 465], "48": [2, 465], "49": [2, 465], "50": [2, 465], "85": [2, 465], "90": [2, 465], "96": [2, 465], "97": [2, 465], "118": [2, 465], "129": [2, 465], "131": [2, 465], "160": [2, 465], "195": [2, 465], "196": [2, 465], "197": [2, 465] }, { "4": [2, 254], "5": [2, 254], "31": [2, 254], "32": [2, 254], "33": [2, 254], "34": [2, 254], "35": [2, 254], "36": [2, 254], "37": [2, 254], "38": [2, 254], "39": [2, 254], "40": [2, 254], "41": [2, 254], "42": [2, 254], "44": [2, 254], "45": [2, 254], "46": [2, 254], "47": [2, 254], "48": [2, 254], "49": [2, 254], "50": [2, 254], "85": [2, 254], "90": [2, 254], "96": [2, 254], "97": [2, 254], "118": [2, 254], "129": [2, 254], "131": [2, 254], "135": [2, 254], "160": [2, 254], "195": [2, 254], "196": [2, 254], "197": [2, 254] }, { "4": [2, 255], "5": [2, 255], "31": [2, 255], "32": [2, 255], "33": [2, 255], "34": [2, 255], "35": [2, 255], "36": [2, 255], "37": [2, 255], "38": [2, 255], "39": [2, 255], "40": [2, 255], "41": [2, 255], "42": [2, 255], "44": [2, 255], "45": [2, 255], "46": [2, 255], "47": [2, 255], "48": [2, 255], "49": [2, 255], "50": [2, 255], "85": [2, 255], "90": [2, 255], "96": [2, 255], "97": [2, 255], "118": [2, 255], "129": [2, 255], "131": [2, 255], "135": [2, 255], "160": [2, 255], "195": [2, 255], "196": [2, 255], "197": [2, 255] }, { "4": [2, 240], "5": [2, 240], "31": [2, 240], "32": [2, 240], "33": [2, 240], "34": [2, 240], "35": [2, 240], "36": [2, 240], "37": [2, 240], "38": [2, 240], "39": [2, 240], "40": [2, 240], "41": [2, 240], "42": [2, 240], "44": [2, 240], "45": [2, 240], "46": [2, 240], "47": [2, 240], "48": [2, 240], "49": [2, 240], "50": [2, 240], "85": [2, 240], "90": [2, 240], "96": [2, 240], "97": [2, 240], "118": [2, 240], "129": [2, 240], "131": [2, 240], "135": [2, 240], "160": [2, 240], "195": [2, 240], "196": [2, 240], "197": [2, 240] }, { "4": [2, 241], "5": [2, 241], "31": [2, 241], "32": [2, 241], "33": [2, 241], "34": [2, 241], "35": [2, 241], "36": [2, 241], "37": [2, 241], "38": [2, 241], "39": [2, 241], "40": [2, 241], "41": [2, 241], "42": [2, 241], "44": [2, 241], "45": [2, 241], "46": [2, 241], "47": [2, 241], "48": [2, 241], "49": [2, 241], "50": [2, 241], "85": [2, 241], "90": [2, 241], "96": [2, 241], "97": [2, 241], "118": [2, 241], "129": [2, 241], "131": [2, 241], "135": [2, 241], "160": [2, 241], "195": [2, 241], "196": [2, 241], "197": [2, 241] }, { "4": [2, 242], "5": [2, 242], "31": [2, 242], "32": [2, 242], "33": [2, 242], "34": [2, 242], "35": [2, 242], "36": [2, 242], "37": [2, 242], "38": [2, 242], "39": [2, 242], "40": [2, 242], "41": [2, 242], "42": [2, 242], "44": [2, 242], "45": [2, 242], "46": [2, 242], "47": [2, 242], "48": [2, 242], "49": [2, 242], "50": [2, 242], "85": [2, 242], "90": [2, 242], "96": [2, 242], "97": [2, 242], "118": [2, 242], "129": [2, 242], "131": [2, 242], "135": [2, 242], "160": [2, 242], "195": [2, 242], "196": [2, 242], "197": [2, 242] }, { "4": [2, 243], "5": [2, 243], "31": [2, 243], "32": [2, 243], "33": [2, 243], "34": [2, 243], "35": [2, 243], "36": [2, 243], "37": [2, 243], "38": [2, 243], "39": [2, 243], "40": [2, 243], "41": [2, 243], "42": [2, 243], "44": [2, 243], "45": [2, 243], "46": [2, 243], "47": [2, 243], "48": [2, 243], "49": [2, 243], "50": [2, 243], "85": [2, 243], "90": [2, 243], "96": [2, 243], "97": [2, 243], "118": [2, 243], "129": [2, 243], "131": [2, 243], "135": [2, 243], "160": [2, 243], "195": [2, 243], "196": [2, 243], "197": [2, 243] }, { "4": [2, 244], "5": [2, 244], "31": [2, 244], "32": [2, 244], "33": [2, 244], "34": [2, 244], "35": [2, 244], "36": [2, 244], "37": [2, 244], "38": [2, 244], "39": [2, 244], "40": [2, 244], "41": [2, 244], "42": [2, 244], "44": [2, 244], "45": [2, 244], "46": [2, 244], "47": [2, 244], "48": [2, 244], "49": [2, 244], "50": [2, 244], "85": [2, 244], "90": [2, 244], "96": [2, 244], "97": [2, 244], "118": [2, 244], "129": [2, 244], "131": [2, 244], "135": [2, 244], "160": [2, 244], "195": [2, 244], "196": [2, 244], "197": [2, 244] }, { "4": [2, 245], "5": [2, 245], "31": [2, 245], "32": [2, 245], "33": [2, 245], "34": [2, 245], "35": [2, 245], "36": [2, 245], "37": [2, 245], "38": [2, 245], "39": [2, 245], "40": [2, 245], "41": [2, 245], "42": [2, 245], "44": [2, 245], "45": [2, 245], "46": [2, 245], "47": [2, 245], "48": [2, 245], "49": [2, 245], "50": [2, 245], "85": [2, 245], "90": [2, 245], "96": [2, 245], "97": [2, 245], "118": [2, 245], "129": [2, 245], "131": [2, 245], "135": [2, 245], "160": [2, 245], "195": [2, 245], "196": [2, 245], "197": [2, 245] }, { "4": [2, 246], "5": [2, 246], "31": [2, 246], "32": [2, 246], "33": [2, 246], "34": [2, 246], "35": [2, 246], "36": [2, 246], "37": [2, 246], "38": [2, 246], "39": [2, 246], "40": [2, 246], "41": [2, 246], "42": [2, 246], "44": [2, 246], "45": [2, 246], "46": [2, 246], "47": [2, 246], "48": [2, 246], "49": [2, 246], "50": [2, 246], "85": [2, 246], "90": [2, 246], "96": [2, 246], "97": [2, 246], "118": [2, 246], "129": [2, 246], "131": [2, 246], "135": [2, 246], "160": [2, 246], "195": [2, 246], "196": [2, 246], "197": [2, 246] }, { "4": [2, 247], "5": [2, 247], "31": [2, 247], "32": [2, 247], "33": [2, 247], "34": [2, 247], "35": [2, 247], "36": [2, 247], "37": [2, 247], "38": [2, 247], "39": [2, 247], "40": [2, 247], "41": [2, 247], "42": [2, 247], "44": [2, 247], "45": [2, 247], "46": [2, 247], "47": [2, 247], "48": [2, 247], "49": [2, 247], "50": [2, 247], "85": [2, 247], "90": [2, 247], "96": [2, 247], "97": [2, 247], "118": [2, 247], "129": [2, 247], "131": [2, 247], "135": [2, 247], "160": [2, 247], "195": [2, 247], "196": [2, 247], "197": [2, 247] }, { "4": [2, 248], "5": [2, 248], "31": [2, 248], "32": [2, 248], "33": [2, 248], "34": [2, 248], "35": [2, 248], "36": [2, 248], "37": [2, 248], "38": [2, 248], "39": [2, 248], "40": [2, 248], "41": [2, 248], "42": [2, 248], "44": [2, 248], "45": [2, 248], "46": [2, 248], "47": [2, 248], "48": [2, 248], "49": [2, 248], "50": [2, 248], "85": [2, 248], "90": [2, 248], "96": [2, 248], "97": [2, 248], "118": [2, 248], "129": [2, 248], "131": [2, 248], "135": [2, 248], "160": [2, 248], "195": [2, 248], "196": [2, 248], "197": [2, 248] }, { "4": [2, 249], "5": [2, 249], "31": [2, 249], "32": [2, 249], "33": [2, 249], "34": [2, 249], "35": [2, 249], "36": [2, 249], "37": [2, 249], "38": [2, 249], "39": [2, 249], "40": [2, 249], "41": [2, 249], "42": [2, 249], "44": [2, 249], "45": [2, 249], "46": [2, 249], "47": [2, 249], "48": [2, 249], "49": [2, 249], "50": [2, 249], "85": [2, 249], "90": [2, 249], "96": [2, 249], "97": [2, 249], "118": [2, 249], "129": [2, 249], "131": [2, 249], "135": [2, 249], "160": [2, 249], "195": [2, 249], "196": [2, 249], "197": [2, 249] }, { "46": [1, 262] }, { "1": [2, 68], "33": [2, 68], "55": [2, 68], "69": [2, 68], "89": [2, 68], "91": [2, 68], "92": [2, 68], "97": [2, 68], "98": [2, 68], "100": [2, 68], "101": [2, 68], "106": [2, 68], "108": [2, 68], "113": [2, 68], "114": [2, 68], "115": [2, 68], "118": [2, 68], "122": [2, 68], "123": [2, 68], "125": [2, 68], "127": [2, 68], "134": [2, 68], "135": [2, 68], "136": [2, 68], "145": [2, 68], "147": [2, 68], "148": [2, 68], "149": [2, 68], "150": [2, 68], "151": [2, 68], "153": [2, 68], "154": [2, 68], "155": [2, 68], "158": [2, 68], "167": [2, 68], "170": [2, 68], "172": [2, 68], "175": [2, 68], "178": [2, 68], "179": [2, 68], "181": [2, 68] }, { "46": [1, 263] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "157": [1, 265], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "90": [1, 269], "195": [1, 270] }, { "90": [1, 271] }, { "38": [1, 272], "90": [2, 165] }, { "38": [1, 273], "90": [2, 167] }, { "38": [2, 168], "90": [2, 168] }, { "38": [2, 170], "90": [2, 170], "96": [1, 274] }, { "34": [2, 215], "37": [1, 276], "38": [2, 215], "39": [1, 277], "45": [1, 281], "47": [1, 279], "49": [1, 282], "50": [1, 280], "90": [2, 215], "96": [2, 215], "131": [1, 275], "197": [1, 283] }, { "31": [1, 286], "34": [2, 217], "35": [1, 284], "37": [2, 217], "38": [2, 217], "39": [2, 217], "41": [1, 285], "45": [2, 217], "47": [2, 217], "49": [2, 217], "50": [2, 217], "90": [2, 217], "96": [2, 217], "131": [2, 217], "197": [2, 217] }, { "31": [2, 220], "34": [2, 220], "35": [2, 220], "36": [1, 287], "37": [2, 220], "38": [2, 220], "39": [2, 220], "41": [2, 220], "45": [2, 220], "47": [2, 220], "49": [2, 220], "50": [2, 220], "90": [2, 220], "96": [2, 220], "131": [2, 220], "197": [2, 220] }, { "31": [2, 224], "34": [2, 224], "35": [2, 224], "36": [2, 224], "37": [2, 224], "38": [2, 224], "39": [2, 224], "41": [2, 224], "45": [2, 224], "47": [2, 224], "49": [2, 224], "50": [2, 224], "90": [2, 224], "96": [2, 224], "131": [2, 224], "197": [2, 224] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "31": [2, 227], "34": [2, 227], "35": [2, 227], "36": [2, 227], "37": [2, 227], "38": [2, 227], "39": [2, 227], "41": [2, 227], "45": [2, 227], "47": [2, 227], "49": [2, 227], "50": [2, 227], "90": [2, 227], "96": [2, 227], "131": [2, 227], "197": [2, 227] }, { "31": [2, 228], "34": [2, 228], "35": [2, 228], "36": [2, 228], "37": [2, 228], "38": [2, 228], "39": [2, 228], "41": [2, 228], "45": [2, 228], "47": [2, 228], "49": [2, 228], "50": [2, 228], "90": [2, 228], "96": [2, 228], "131": [2, 228], "197": [2, 228] }, { "31": [2, 229], "34": [2, 229], "35": [2, 229], "36": [2, 229], "37": [2, 229], "38": [2, 229], "39": [2, 229], "41": [2, 229], "45": [2, 229], "47": [2, 229], "49": [2, 229], "50": [2, 229], "90": [2, 229], "96": [2, 229], "131": [2, 229], "197": [2, 229] }, { "31": [2, 230], "34": [2, 230], "35": [2, 230], "36": [2, 230], "37": [2, 230], "38": [2, 230], "39": [2, 230], "41": [2, 230], "45": [2, 230], "47": [2, 230], "49": [2, 230], "50": [2, 230], "90": [2, 230], "96": [2, 230], "131": [2, 230], "197": [2, 230] }, { "31": [2, 231], "34": [2, 231], "35": [2, 231], "36": [2, 231], "37": [2, 231], "38": [2, 231], "39": [2, 231], "41": [2, 231], "45": [2, 231], "47": [2, 231], "49": [2, 231], "50": [2, 231], "90": [2, 231], "96": [2, 231], "131": [2, 231], "197": [2, 231] }, { "31": [2, 232], "34": [2, 232], "35": [2, 232], "36": [2, 232], "37": [2, 232], "38": [2, 232], "39": [2, 232], "41": [2, 232], "45": [2, 232], "47": [2, 232], "49": [2, 232], "50": [2, 232], "90": [2, 232], "96": [2, 232], "131": [2, 232], "197": [2, 232] }, { "31": [2, 233], "34": [2, 233], "35": [2, 233], "36": [2, 233], "37": [2, 233], "38": [2, 233], "39": [2, 233], "41": [2, 233], "45": [2, 233], "47": [2, 233], "49": [2, 233], "50": [2, 233], "90": [2, 233], "96": [2, 233], "131": [2, 233], "197": [2, 233] }, { "31": [2, 234], "34": [2, 234], "35": [2, 234], "36": [2, 234], "37": [2, 234], "38": [2, 234], "39": [2, 234], "41": [2, 234], "45": [2, 234], "47": [2, 234], "49": [2, 234], "50": [2, 234], "90": [2, 234], "96": [2, 234], "131": [2, 234], "197": [2, 234] }, { "4": [2, 294], "5": [2, 294], "31": [2, 294], "32": [2, 294], "34": [2, 294], "35": [2, 294], "36": [2, 294], "37": [2, 294], "38": [2, 294], "39": [2, 294], "41": [2, 294], "42": [2, 294], "44": [2, 294], "45": [2, 294], "47": [2, 294], "48": [2, 294], "49": [2, 294], "50": [2, 294], "90": [2, 294], "96": [2, 294], "97": [2, 294], "118": [2, 294], "131": [2, 294], "195": [2, 294], "196": [2, 294], "197": [2, 294] }, { "4": [2, 295], "5": [2, 295], "31": [2, 295], "32": [2, 295], "34": [2, 295], "35": [2, 295], "36": [2, 295], "37": [2, 295], "38": [2, 295], "39": [2, 295], "41": [2, 295], "42": [2, 295], "44": [2, 295], "45": [2, 295], "47": [2, 295], "48": [2, 295], "49": [2, 295], "50": [2, 295], "90": [2, 295], "96": [2, 295], "97": [2, 295], "118": [2, 295], "131": [2, 295], "195": [2, 295], "196": [2, 295], "197": [2, 295] }, { "4": [2, 296], "5": [2, 296], "31": [2, 296], "32": [2, 296], "34": [2, 296], "35": [2, 296], "36": [2, 296], "37": [2, 296], "38": [2, 296], "39": [2, 296], "41": [2, 296], "42": [2, 296], "44": [2, 296], "45": [2, 296], "47": [2, 296], "48": [2, 296], "49": [2, 296], "50": [2, 296], "90": [2, 296], "96": [2, 296], "97": [2, 296], "118": [2, 296], "131": [2, 296], "195": [2, 296], "196": [2, 296], "197": [2, 296] }, { "4": [2, 297], "5": [2, 297], "31": [2, 297], "32": [2, 297], "34": [2, 297], "35": [2, 297], "36": [2, 297], "37": [2, 297], "38": [2, 297], "39": [2, 297], "41": [2, 297], "42": [2, 297], "44": [2, 297], "45": [2, 297], "47": [2, 297], "48": [2, 297], "49": [2, 297], "50": [2, 297], "90": [2, 297], "96": [2, 297], "97": [2, 297], "118": [2, 297], "131": [2, 297], "195": [2, 297], "196": [2, 297], "197": [2, 297] }, { "4": [2, 272], "5": [2, 272], "31": [2, 272], "32": [2, 272], "34": [2, 272], "35": [2, 272], "36": [2, 272], "37": [2, 272], "38": [2, 272], "39": [2, 272], "40": [1, 293], "41": [2, 272], "42": [2, 272], "44": [2, 272], "45": [2, 272], "47": [2, 272], "48": [2, 272], "49": [2, 272], "50": [2, 272], "90": [2, 272], "96": [2, 272], "97": [2, 272], "118": [2, 272], "131": [2, 272], "195": [2, 272], "196": [2, 272], "197": [2, 272] }, { "4": [2, 273], "5": [2, 273], "31": [2, 273], "32": [2, 273], "34": [2, 273], "35": [2, 273], "36": [2, 273], "37": [2, 273], "38": [2, 273], "39": [2, 273], "40": [2, 273], "41": [2, 273], "42": [2, 273], "44": [2, 273], "45": [2, 273], "46": [2, 273], "47": [2, 273], "48": [2, 273], "49": [2, 273], "50": [2, 273], "85": [2, 273], "90": [2, 273], "96": [2, 273], "97": [2, 273], "118": [2, 273], "131": [2, 273], "160": [2, 273], "195": [2, 273], "196": [2, 273], "197": [2, 273] }, { "4": [2, 462], "5": [2, 462], "31": [2, 462], "32": [2, 462], "33": [1, 294], "34": [2, 462], "35": [2, 462], "36": [2, 462], "37": [2, 462], "38": [2, 462], "39": [2, 462], "40": [2, 462], "41": [2, 462], "42": [2, 462], "44": [2, 462], "45": [2, 462], "46": [2, 462], "47": [2, 462], "48": [2, 462], "49": [2, 462], "50": [2, 462], "85": [2, 462], "90": [2, 462], "96": [2, 462], "97": [2, 462], "118": [2, 462], "129": [2, 462], "131": [2, 462], "160": [2, 462], "195": [2, 462], "196": [2, 462], "197": [2, 462] }, { "4": [2, 463], "5": [2, 463], "31": [2, 463], "32": [2, 463], "33": [1, 295], "34": [2, 463], "35": [2, 463], "36": [2, 463], "37": [2, 463], "38": [2, 463], "39": [2, 463], "40": [2, 463], "41": [2, 463], "42": [2, 463], "44": [2, 463], "45": [2, 463], "46": [2, 463], "47": [2, 463], "48": [2, 463], "49": [2, 463], "50": [2, 463], "85": [2, 463], "90": [2, 463], "96": [2, 463], "97": [2, 463], "118": [2, 463], "129": [2, 463], "131": [2, 463], "160": [2, 463], "195": [2, 463], "196": [2, 463], "197": [2, 463] }, { "4": [2, 464], "5": [2, 464], "31": [2, 464], "32": [2, 464], "33": [1, 296], "34": [2, 464], "35": [2, 464], "36": [2, 464], "37": [2, 464], "38": [2, 464], "39": [2, 464], "40": [2, 464], "41": [2, 464], "42": [2, 464], "44": [2, 464], "45": [2, 464], "46": [2, 464], "47": [2, 464], "48": [2, 464], "49": [2, 464], "50": [2, 464], "85": [2, 464], "90": [2, 464], "96": [2, 464], "97": [2, 464], "118": [2, 464], "129": [2, 464], "131": [2, 464], "160": [2, 464], "195": [2, 464], "196": [2, 464], "197": [2, 464] }, { "4": [2, 465], "5": [2, 465], "31": [2, 465], "32": [2, 465], "33": [1, 297], "34": [2, 465], "35": [2, 465], "36": [2, 465], "37": [2, 465], "38": [2, 465], "39": [2, 465], "40": [2, 465], "41": [2, 465], "42": [2, 465], "44": [2, 465], "45": [2, 465], "46": [2, 465], "47": [2, 465], "48": [2, 465], "49": [2, 465], "50": [2, 465], "85": [2, 465], "90": [2, 465], "96": [2, 465], "97": [2, 465], "118": [2, 465], "129": [2, 465], "131": [2, 465], "160": [2, 465], "195": [2, 465], "196": [2, 465], "197": [2, 465] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "90": [1, 298], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "196": [1, 301], "198": [1, 109], "199": [1, 112] }, { "40": [1, 304] }, { "4": [2, 292], "5": [2, 292], "31": [2, 292], "32": [2, 292], "34": [2, 292], "35": [2, 292], "36": [2, 292], "37": [2, 292], "38": [2, 292], "39": [2, 292], "41": [2, 292], "42": [2, 292], "44": [2, 292], "45": [2, 292], "47": [2, 292], "48": [2, 292], "49": [2, 292], "50": [2, 292], "90": [2, 292], "96": [2, 292], "97": [2, 292], "118": [2, 292], "131": [2, 292], "195": [2, 292], "196": [2, 292], "197": [2, 292] }, { "4": [2, 293], "5": [2, 293], "31": [2, 293], "32": [2, 293], "34": [2, 293], "35": [2, 293], "36": [2, 293], "37": [2, 293], "38": [2, 293], "39": [2, 293], "41": [2, 293], "42": [2, 293], "44": [2, 293], "45": [2, 293], "47": [2, 293], "48": [2, 293], "49": [2, 293], "50": [2, 293], "90": [2, 293], "96": [2, 293], "97": [2, 293], "118": [2, 293], "131": [2, 293], "195": [2, 293], "196": [2, 293], "197": [2, 293] }, { "4": [2, 323], "5": [2, 323], "31": [2, 323], "32": [2, 323], "34": [2, 323], "35": [2, 323], "36": [2, 323], "37": [2, 323], "38": [2, 323], "39": [2, 323], "41": [2, 323], "42": [2, 323], "44": [2, 323], "45": [2, 323], "47": [2, 323], "48": [2, 323], "49": [2, 323], "50": [2, 323], "90": [2, 323], "96": [2, 323], "97": [2, 323], "118": [2, 323], "131": [2, 323], "195": [2, 323], "196": [2, 323], "197": [2, 323] }, { "4": [2, 324], "5": [2, 324], "31": [2, 324], "32": [2, 324], "34": [2, 324], "35": [2, 324], "36": [2, 324], "37": [2, 324], "38": [2, 324], "39": [2, 324], "41": [2, 324], "42": [2, 324], "44": [2, 324], "45": [2, 324], "47": [2, 324], "48": [2, 324], "49": [2, 324], "50": [2, 324], "90": [2, 324], "96": [2, 324], "97": [2, 324], "118": [2, 324], "131": [2, 324], "195": [2, 324], "196": [2, 324], "197": [2, 324] }, { "4": [2, 325], "5": [2, 325], "31": [2, 325], "32": [2, 325], "34": [2, 325], "35": [2, 325], "36": [2, 325], "37": [2, 325], "38": [2, 325], "39": [2, 325], "41": [2, 325], "42": [2, 325], "44": [2, 325], "45": [2, 325], "47": [2, 325], "48": [2, 325], "49": [2, 325], "50": [2, 325], "90": [2, 325], "96": [2, 325], "97": [2, 325], "118": [2, 325], "131": [2, 325], "195": [2, 325], "196": [2, 325], "197": [2, 325] }, { "4": [2, 326], "5": [2, 326], "31": [2, 326], "32": [2, 326], "34": [2, 326], "35": [2, 326], "36": [2, 326], "37": [2, 326], "38": [2, 326], "39": [2, 326], "41": [2, 326], "42": [2, 326], "44": [2, 326], "45": [2, 326], "47": [2, 326], "48": [2, 326], "49": [2, 326], "50": [2, 326], "90": [2, 326], "96": [2, 326], "97": [2, 326], "118": [2, 326], "131": [2, 326], "195": [2, 326], "196": [2, 326], "197": [2, 326] }, { "4": [2, 327], "5": [2, 327], "31": [2, 327], "32": [2, 327], "34": [2, 327], "35": [2, 327], "36": [2, 327], "37": [2, 327], "38": [2, 327], "39": [2, 327], "41": [2, 327], "42": [2, 327], "44": [2, 327], "45": [2, 327], "47": [2, 327], "48": [2, 327], "49": [2, 327], "50": [2, 327], "90": [2, 327], "96": [2, 327], "97": [2, 327], "118": [2, 327], "131": [2, 327], "195": [2, 327], "196": [2, 327], "197": [2, 327] }, { "4": [2, 328], "5": [2, 328], "31": [2, 328], "32": [2, 328], "34": [2, 328], "35": [2, 328], "36": [2, 328], "37": [2, 328], "38": [2, 328], "39": [2, 328], "41": [2, 328], "42": [2, 328], "44": [2, 328], "45": [2, 328], "47": [2, 328], "48": [2, 328], "49": [2, 328], "50": [2, 328], "90": [2, 328], "96": [2, 328], "97": [2, 328], "118": [2, 328], "131": [2, 328], "195": [2, 328], "196": [2, 328], "197": [2, 328] }, { "4": [2, 329], "5": [2, 329], "31": [2, 329], "32": [2, 329], "34": [2, 329], "35": [2, 329], "36": [2, 329], "37": [2, 329], "38": [2, 329], "39": [2, 329], "41": [2, 329], "42": [2, 329], "44": [2, 329], "45": [2, 329], "47": [2, 329], "48": [2, 329], "49": [2, 329], "50": [2, 329], "90": [2, 329], "96": [2, 329], "97": [2, 329], "118": [2, 329], "131": [2, 329], "195": [2, 329], "196": [2, 329], "197": [2, 329] }, { "4": [2, 330], "5": [2, 330], "31": [2, 330], "32": [2, 330], "34": [2, 330], "35": [2, 330], "36": [2, 330], "37": [2, 330], "38": [2, 330], "39": [2, 330], "41": [2, 330], "42": [2, 330], "44": [2, 330], "45": [2, 330], "47": [2, 330], "48": [2, 330], "49": [2, 330], "50": [2, 330], "90": [2, 330], "96": [2, 330], "97": [2, 330], "118": [2, 330], "131": [2, 330], "195": [2, 330], "196": [2, 330], "197": [2, 330] }, { "4": [2, 331], "5": [2, 331], "31": [2, 331], "32": [2, 331], "34": [2, 331], "35": [2, 331], "36": [2, 331], "37": [2, 331], "38": [2, 331], "39": [2, 331], "41": [2, 331], "42": [2, 331], "44": [2, 331], "45": [2, 331], "47": [2, 331], "48": [2, 331], "49": [2, 331], "50": [2, 331], "90": [2, 331], "96": [2, 331], "97": [2, 331], "118": [2, 331], "131": [2, 331], "195": [2, 331], "196": [2, 331], "197": [2, 331] }, { "4": [2, 332], "5": [2, 332], "31": [2, 332], "32": [2, 332], "34": [2, 332], "35": [2, 332], "36": [2, 332], "37": [2, 332], "38": [2, 332], "39": [2, 332], "41": [2, 332], "42": [2, 332], "44": [2, 332], "45": [2, 332], "47": [2, 332], "48": [2, 332], "49": [2, 332], "50": [2, 332], "90": [2, 332], "96": [2, 332], "97": [2, 332], "118": [2, 332], "131": [2, 332], "195": [2, 332], "196": [2, 332], "197": [2, 332] }, { "4": [2, 333], "5": [2, 333], "31": [2, 333], "32": [2, 333], "34": [2, 333], "35": [2, 333], "36": [2, 333], "37": [2, 333], "38": [2, 333], "39": [2, 333], "41": [2, 333], "42": [2, 333], "44": [2, 333], "45": [2, 333], "47": [2, 333], "48": [2, 333], "49": [2, 333], "50": [2, 333], "90": [2, 333], "96": [2, 333], "97": [2, 333], "118": [2, 333], "131": [2, 333], "195": [2, 333], "196": [2, 333], "197": [2, 333] }, { "4": [2, 334], "5": [2, 334], "31": [2, 334], "32": [2, 334], "34": [2, 334], "35": [2, 334], "36": [2, 334], "37": [2, 334], "38": [2, 334], "39": [2, 334], "41": [2, 334], "42": [2, 334], "44": [2, 334], "45": [2, 334], "47": [2, 334], "48": [2, 334], "49": [2, 334], "50": [2, 334], "90": [2, 334], "96": [2, 334], "97": [2, 334], "118": [2, 334], "131": [2, 334], "195": [2, 334], "196": [2, 334], "197": [2, 334] }, { "4": [2, 335], "5": [2, 335], "31": [2, 335], "32": [2, 335], "34": [2, 335], "35": [2, 335], "36": [2, 335], "37": [2, 335], "38": [2, 335], "39": [2, 335], "41": [2, 335], "42": [2, 335], "44": [2, 335], "45": [2, 335], "47": [2, 335], "48": [2, 335], "49": [2, 335], "50": [2, 335], "90": [2, 335], "96": [2, 335], "97": [2, 335], "118": [2, 335], "131": [2, 335], "195": [2, 335], "196": [2, 335], "197": [2, 335] }, { "4": [2, 336], "5": [2, 336], "31": [2, 336], "32": [2, 336], "34": [2, 336], "35": [2, 336], "36": [2, 336], "37": [2, 336], "38": [2, 336], "39": [2, 336], "41": [2, 336], "42": [2, 336], "44": [2, 336], "45": [2, 336], "47": [2, 336], "48": [2, 336], "49": [2, 336], "50": [2, 336], "90": [2, 336], "96": [2, 336], "97": [2, 336], "118": [2, 336], "131": [2, 336], "195": [2, 336], "196": [2, 336], "197": [2, 336] }, { "4": [2, 337], "5": [2, 337], "31": [2, 337], "32": [2, 337], "34": [2, 337], "35": [2, 337], "36": [2, 337], "37": [2, 337], "38": [2, 337], "39": [2, 337], "41": [2, 337], "42": [2, 337], "44": [2, 337], "45": [2, 337], "47": [2, 337], "48": [2, 337], "49": [2, 337], "50": [2, 337], "90": [2, 337], "96": [2, 337], "97": [2, 337], "118": [2, 337], "131": [2, 337], "195": [2, 337], "196": [2, 337], "197": [2, 337] }, { "4": [2, 338], "5": [2, 338], "31": [2, 338], "32": [2, 338], "34": [2, 338], "35": [2, 338], "36": [2, 338], "37": [2, 338], "38": [2, 338], "39": [2, 338], "41": [2, 338], "42": [2, 338], "44": [2, 338], "45": [2, 338], "47": [2, 338], "48": [2, 338], "49": [2, 338], "50": [2, 338], "90": [2, 338], "96": [2, 338], "97": [2, 338], "118": [2, 338], "131": [2, 338], "195": [2, 338], "196": [2, 338], "197": [2, 338] }, { "4": [2, 339], "5": [2, 339], "31": [2, 339], "32": [2, 339], "34": [2, 339], "35": [2, 339], "36": [2, 339], "37": [2, 339], "38": [2, 339], "39": [2, 339], "41": [2, 339], "42": [2, 339], "44": [2, 339], "45": [2, 339], "47": [2, 339], "48": [2, 339], "49": [2, 339], "50": [2, 339], "90": [2, 339], "96": [2, 339], "97": [2, 339], "118": [2, 339], "131": [2, 339], "195": [2, 339], "196": [2, 339], "197": [2, 339] }, { "4": [2, 340], "5": [2, 340], "31": [2, 340], "32": [2, 340], "34": [2, 340], "35": [2, 340], "36": [2, 340], "37": [2, 340], "38": [2, 340], "39": [2, 340], "41": [2, 340], "42": [2, 340], "44": [2, 340], "45": [2, 340], "47": [2, 340], "48": [2, 340], "49": [2, 340], "50": [2, 340], "90": [2, 340], "96": [2, 340], "97": [2, 340], "118": [2, 340], "131": [2, 340], "195": [2, 340], "196": [2, 340], "197": [2, 340] }, { "4": [2, 341], "5": [2, 341], "31": [2, 341], "32": [2, 341], "34": [2, 341], "35": [2, 341], "36": [2, 341], "37": [2, 341], "38": [2, 341], "39": [2, 341], "41": [2, 341], "42": [2, 341], "44": [2, 341], "45": [2, 341], "47": [2, 341], "48": [2, 341], "49": [2, 341], "50": [2, 341], "90": [2, 341], "96": [2, 341], "97": [2, 341], "118": [2, 341], "131": [2, 341], "195": [2, 341], "196": [2, 341], "197": [2, 341] }, { "4": [2, 342], "5": [2, 342], "31": [2, 342], "32": [2, 342], "34": [2, 342], "35": [2, 342], "36": [2, 342], "37": [2, 342], "38": [2, 342], "39": [2, 342], "41": [2, 342], "42": [2, 342], "44": [2, 342], "45": [2, 342], "47": [2, 342], "48": [2, 342], "49": [2, 342], "50": [2, 342], "90": [2, 342], "96": [2, 342], "97": [2, 342], "118": [2, 342], "131": [2, 342], "195": [2, 342], "196": [2, 342], "197": [2, 342] }, { "4": [2, 343], "5": [2, 343], "31": [2, 343], "32": [2, 343], "34": [2, 343], "35": [2, 343], "36": [2, 343], "37": [2, 343], "38": [2, 343], "39": [2, 343], "41": [2, 343], "42": [2, 343], "44": [2, 343], "45": [2, 343], "47": [2, 343], "48": [2, 343], "49": [2, 343], "50": [2, 343], "90": [2, 343], "96": [2, 343], "97": [2, 343], "118": [2, 343], "131": [2, 343], "195": [2, 343], "196": [2, 343], "197": [2, 343] }, { "52": [1, 305], "62": [1, 308], "71": [1, 309], "81": [1, 306], "86": [1, 307] }, { "4": [2, 388], "5": [2, 388], "31": [2, 388], "32": [2, 388], "34": [2, 388], "35": [2, 388], "36": [2, 388], "37": [2, 388], "38": [2, 388], "39": [2, 388], "41": [2, 388], "42": [2, 388], "44": [2, 388], "45": [2, 388], "47": [2, 388], "48": [2, 388], "49": [2, 388], "50": [2, 388], "90": [2, 388], "96": [2, 388], "97": [2, 388], "118": [2, 388], "131": [2, 388], "195": [2, 388], "196": [2, 388], "197": [2, 388] }, { "33": [1, 310], "51": [1, 378], "52": [1, 379], "53": [1, 380], "54": [1, 381], "57": [1, 382], "58": [1, 383], "62": [1, 384], "65": [1, 385], "66": [1, 386], "69": [1, 394], "71": [1, 387], "73": [1, 388], "75": [1, 389], "76": [1, 390], "81": [1, 391], "83": [1, 392], "86": [1, 393], "91": [1, 324], "92": [1, 57], "93": [1, 372], "94": [1, 325], "95": [1, 326], "96": [1, 327], "97": [1, 74], "98": [1, 79], "99": [1, 339], "100": [1, 66], "101": [1, 67], "102": [1, 373], "103": [1, 340], "104": [1, 341], "105": [1, 342], "106": [1, 313], "107": [1, 374], "108": [1, 53], "109": [1, 343], "110": [1, 107], "111": [1, 108], "112": [1, 344], "113": [1, 399], "114": [1, 328], "115": [1, 68], "116": [1, 345], "117": [1, 346], "118": [1, 75], "119": [1, 101], "120": [1, 347], "121": [1, 375], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 314], "127": [1, 52], "128": [1, 348], "129": [1, 315], "130": [1, 323], "131": [1, 329], "132": [1, 349], "133": [1, 350], "134": [1, 71], "135": [1, 59], "136": [1, 330], "137": [1, 351], "138": [1, 352], "139": [1, 316], "140": [1, 317], "141": [1, 110], "142": [1, 111], "143": [1, 353], "144": [1, 354], "145": [1, 72], "146": [1, 376], "147": [1, 60], "148": [1, 318], "149": [1, 61], "150": [1, 77], "151": [1, 76], "152": [1, 331], "153": [1, 332], "154": [1, 62], "155": [1, 54], "156": [1, 319], "157": [1, 333], "158": [1, 334], "159": [1, 320], "160": [1, 321], "161": [1, 377], "162": [1, 336], "163": [1, 335], "164": [1, 355], "165": [1, 105], "166": [1, 356], "167": [1, 63], "168": [1, 357], "169": [1, 358], "170": [1, 64], "171": [1, 337], "172": [1, 55], "173": [1, 359], "174": [1, 102], "175": [1, 338], "176": [1, 103], "177": [1, 360], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "182": [1, 361], "183": [1, 362], "184": [1, 363], "185": [1, 364], "186": [1, 322], "187": [1, 365], "188": [1, 366], "189": [1, 367], "190": [1, 368], "191": [1, 369], "192": [1, 370], "193": [1, 371], "198": [1, 109], "199": [1, 112] }, { "4": [2, 401], "5": [2, 401], "31": [2, 401], "32": [2, 401], "34": [2, 401], "35": [2, 401], "36": [2, 401], "37": [2, 401], "38": [2, 401], "39": [2, 401], "41": [2, 401], "42": [2, 401], "44": [2, 401], "45": [2, 401], "47": [2, 401], "48": [2, 401], "49": [2, 401], "50": [2, 401], "90": [2, 401], "96": [2, 401], "97": [2, 401], "118": [2, 401], "131": [2, 401], "195": [2, 401], "196": [2, 401], "197": [2, 401] }, { "4": [2, 402], "5": [2, 402], "31": [2, 402], "32": [2, 402], "34": [2, 402], "35": [2, 402], "36": [2, 402], "37": [2, 402], "38": [2, 402], "39": [2, 402], "40": [1, 400], "41": [2, 402], "42": [2, 402], "44": [2, 402], "45": [2, 402], "47": [2, 402], "48": [2, 402], "49": [2, 402], "50": [2, 402], "90": [2, 402], "96": [2, 402], "97": [2, 402], "118": [2, 402], "131": [2, 402], "195": [2, 402], "196": [2, 402], "197": [2, 402] }, { "4": [2, 397], "5": [2, 397], "31": [2, 397], "32": [2, 397], "34": [2, 397], "35": [2, 397], "36": [2, 397], "37": [2, 397], "38": [2, 397], "39": [2, 397], "41": [2, 397], "42": [2, 397], "44": [2, 397], "45": [2, 397], "47": [2, 397], "48": [2, 397], "49": [2, 397], "50": [2, 397], "90": [2, 397], "96": [2, 397], "97": [2, 397], "118": [2, 397], "131": [2, 397], "195": [2, 397], "196": [2, 397], "197": [2, 397] }, { "4": [2, 400], "5": [2, 400], "31": [2, 400], "32": [2, 400], "34": [2, 400], "35": [2, 400], "36": [2, 400], "37": [2, 400], "38": [2, 400], "39": [2, 400], "41": [2, 400], "42": [2, 400], "44": [2, 400], "45": [2, 400], "47": [2, 400], "48": [2, 400], "49": [2, 400], "50": [2, 400], "90": [2, 400], "96": [2, 400], "97": [2, 400], "118": [2, 400], "131": [2, 400], "195": [2, 400], "196": [2, 400], "197": [2, 400] }, { "4": [2, 399], "5": [2, 399], "31": [2, 399], "32": [2, 399], "34": [2, 399], "35": [2, 399], "36": [2, 399], "37": [2, 399], "38": [2, 399], "39": [2, 399], "41": [2, 399], "42": [2, 399], "44": [2, 399], "45": [2, 399], "47": [2, 399], "48": [2, 399], "49": [2, 399], "50": [2, 399], "90": [2, 399], "96": [2, 399], "97": [2, 399], "118": [2, 399], "131": [2, 399], "195": [2, 399], "196": [2, 399], "197": [2, 399] }, { "4": [2, 398], "5": [2, 398], "31": [2, 398], "32": [2, 398], "34": [2, 398], "35": [2, 398], "36": [2, 398], "37": [2, 398], "38": [2, 398], "39": [2, 398], "41": [2, 398], "42": [2, 398], "44": [2, 398], "45": [2, 398], "47": [2, 398], "48": [2, 398], "49": [2, 398], "50": [2, 398], "90": [2, 398], "96": [2, 398], "97": [2, 398], "118": [2, 398], "131": [2, 398], "195": [2, 398], "196": [2, 398], "197": [2, 398] }, { "4": [2, 396], "5": [2, 396], "31": [2, 396], "32": [2, 396], "34": [2, 396], "35": [2, 396], "36": [2, 396], "37": [2, 396], "38": [2, 396], "39": [2, 396], "41": [2, 396], "42": [2, 396], "44": [2, 396], "45": [2, 396], "47": [2, 396], "48": [2, 396], "49": [2, 396], "50": [2, 396], "90": [2, 396], "96": [2, 396], "97": [2, 396], "118": [2, 396], "131": [2, 396], "195": [2, 396], "196": [2, 396], "197": [2, 396] }, { "4": [2, 395], "5": [2, 395], "31": [2, 395], "32": [2, 395], "34": [2, 395], "35": [2, 395], "36": [2, 395], "37": [2, 395], "38": [2, 395], "39": [2, 395], "41": [2, 395], "42": [2, 395], "44": [2, 395], "45": [2, 395], "47": [2, 395], "48": [2, 395], "49": [2, 395], "50": [2, 395], "90": [2, 395], "96": [2, 395], "97": [2, 395], "118": [2, 395], "131": [2, 395], "195": [2, 395], "196": [2, 395], "197": [2, 395] }, { "64": [1, 401] }, { "4": [2, 349], "5": [2, 349], "31": [2, 349], "32": [2, 349], "34": [2, 349], "35": [2, 349], "36": [2, 349], "37": [2, 349], "38": [2, 349], "39": [2, 349], "41": [2, 349], "42": [2, 349], "44": [2, 349], "45": [2, 349], "47": [2, 349], "48": [2, 349], "49": [2, 349], "50": [2, 349], "64": [1, 402], "90": [2, 349], "96": [2, 349], "97": [2, 349], "118": [2, 349], "131": [2, 349], "195": [2, 349], "196": [2, 349], "197": [2, 349] }, { "4": [2, 394], "5": [2, 394], "31": [2, 394], "32": [2, 394], "34": [2, 394], "35": [2, 394], "36": [2, 394], "37": [2, 394], "38": [2, 394], "39": [2, 394], "41": [2, 394], "42": [2, 394], "44": [2, 394], "45": [2, 394], "47": [2, 394], "48": [2, 394], "49": [2, 394], "50": [2, 394], "90": [2, 394], "96": [2, 394], "97": [2, 394], "118": [2, 394], "131": [2, 394], "195": [2, 394], "196": [2, 394], "197": [2, 394] }, { "4": [2, 348], "5": [2, 348], "31": [2, 348], "32": [2, 348], "34": [2, 348], "35": [2, 348], "36": [2, 348], "37": [2, 348], "38": [2, 348], "39": [2, 348], "41": [2, 348], "42": [2, 348], "44": [2, 348], "45": [2, 348], "47": [2, 348], "48": [2, 348], "49": [2, 348], "50": [2, 348], "64": [2, 348], "90": [2, 348], "96": [2, 348], "97": [2, 348], "118": [2, 348], "131": [2, 348], "194": [2, 348], "195": [2, 348], "196": [2, 348], "197": [2, 348] }, { "4": [2, 368], "5": [2, 368], "31": [2, 368], "32": [2, 368], "34": [2, 368], "35": [2, 368], "36": [2, 368], "37": [2, 368], "38": [2, 368], "39": [2, 368], "41": [2, 368], "42": [2, 368], "44": [2, 368], "45": [2, 368], "47": [2, 368], "48": [2, 368], "49": [2, 368], "50": [2, 368], "90": [2, 368], "96": [2, 368], "97": [2, 368], "118": [2, 368], "131": [2, 368], "195": [2, 368], "196": [2, 368], "197": [2, 368] }, { "40": [1, 403] }, { "4": [2, 364], "5": [2, 364], "31": [2, 364], "32": [2, 364], "34": [2, 364], "35": [2, 364], "36": [2, 364], "37": [2, 364], "38": [2, 364], "39": [2, 364], "41": [2, 364], "42": [2, 364], "44": [2, 364], "45": [2, 364], "47": [2, 364], "48": [2, 364], "49": [2, 364], "50": [2, 364], "90": [2, 364], "96": [2, 364], "97": [2, 364], "118": [2, 364], "131": [2, 364], "195": [2, 364], "196": [2, 364], "197": [2, 364] }, { "4": [2, 370], "5": [2, 370], "31": [2, 370], "32": [2, 370], "34": [2, 370], "35": [2, 370], "36": [2, 370], "37": [2, 370], "38": [2, 370], "39": [2, 370], "41": [2, 370], "42": [2, 370], "44": [2, 370], "45": [2, 370], "47": [2, 370], "48": [2, 370], "49": [2, 370], "50": [2, 370], "90": [2, 370], "96": [2, 370], "97": [2, 370], "118": [2, 370], "131": [2, 370], "195": [2, 370], "196": [2, 370], "197": [2, 370] }, { "4": [2, 372], "5": [2, 372], "31": [2, 372], "32": [2, 372], "34": [2, 372], "35": [2, 372], "36": [2, 372], "37": [2, 372], "38": [2, 372], "39": [2, 372], "41": [2, 372], "42": [2, 372], "44": [2, 372], "45": [2, 372], "47": [2, 372], "48": [2, 372], "49": [2, 372], "50": [2, 372], "90": [2, 372], "96": [2, 372], "97": [2, 372], "118": [2, 372], "131": [2, 372], "195": [2, 372], "196": [2, 372], "197": [2, 372] }, { "4": [2, 377], "5": [2, 377], "31": [2, 377], "32": [2, 377], "34": [2, 377], "35": [2, 377], "36": [2, 377], "37": [2, 377], "38": [2, 377], "39": [2, 377], "41": [2, 377], "42": [2, 377], "44": [2, 377], "45": [2, 377], "47": [2, 377], "48": [2, 377], "49": [2, 377], "50": [2, 377], "90": [2, 377], "96": [2, 377], "97": [2, 377], "118": [2, 377], "131": [2, 377], "195": [2, 377], "196": [2, 377], "197": [2, 377] }, { "33": [1, 404] }, { "33": [1, 405] }, { "4": [2, 359], "5": [2, 359], "31": [2, 359], "32": [2, 359], "34": [2, 359], "35": [2, 359], "36": [2, 359], "37": [2, 359], "38": [2, 359], "39": [2, 359], "41": [2, 359], "42": [2, 359], "44": [2, 359], "45": [2, 359], "47": [2, 359], "48": [2, 359], "49": [2, 359], "50": [2, 359], "90": [2, 359], "96": [2, 359], "97": [2, 359], "118": [2, 359], "131": [2, 359], "195": [2, 359], "196": [2, 359], "197": [2, 359] }, { "40": [1, 406] }, { "4": [2, 357], "5": [2, 357], "31": [2, 357], "32": [2, 357], "34": [2, 357], "35": [2, 357], "36": [2, 357], "37": [2, 357], "38": [2, 357], "39": [2, 357], "41": [2, 357], "42": [2, 357], "44": [2, 357], "45": [2, 357], "47": [2, 357], "48": [2, 357], "49": [2, 357], "50": [2, 357], "90": [2, 357], "96": [2, 357], "97": [2, 357], "118": [2, 357], "131": [2, 357], "195": [2, 357], "196": [2, 357], "197": [2, 357] }, { "4": [2, 356], "5": [2, 356], "31": [2, 356], "32": [2, 356], "34": [2, 356], "35": [2, 356], "36": [2, 356], "37": [2, 356], "38": [2, 356], "39": [2, 356], "41": [2, 356], "42": [2, 356], "44": [2, 356], "45": [2, 356], "47": [2, 356], "48": [2, 356], "49": [2, 356], "50": [2, 356], "60": [1, 407], "69": [1, 409], "90": [2, 356], "96": [2, 356], "97": [2, 356], "118": [2, 356], "131": [2, 356], "140": [1, 411], "151": [1, 410], "195": [2, 356], "196": [2, 356], "197": [2, 356] }, { "4": [2, 351], "5": [2, 351], "31": [2, 351], "32": [2, 351], "34": [2, 351], "35": [2, 351], "36": [2, 351], "37": [2, 351], "38": [2, 351], "39": [2, 351], "41": [2, 351], "42": [2, 351], "44": [2, 351], "45": [2, 351], "47": [2, 351], "48": [2, 351], "49": [2, 351], "50": [2, 351], "60": [2, 351], "69": [2, 351], "74": [2, 351], "90": [2, 351], "96": [2, 351], "97": [2, 351], "118": [2, 351], "131": [2, 351], "140": [2, 351], "151": [2, 351], "195": [2, 351], "196": [2, 351], "197": [2, 351] }, { "74": [1, 204] }, { "4": [2, 350], "5": [2, 350], "31": [2, 350], "32": [2, 350], "34": [2, 350], "35": [2, 350], "36": [2, 350], "37": [2, 350], "38": [2, 350], "39": [2, 350], "41": [2, 350], "42": [2, 350], "44": [2, 350], "45": [2, 350], "47": [2, 350], "48": [2, 350], "49": [2, 350], "50": [2, 350], "60": [2, 350], "69": [2, 350], "74": [2, 350], "90": [2, 350], "96": [2, 350], "97": [2, 350], "118": [2, 350], "131": [2, 350], "140": [2, 350], "151": [2, 350], "195": [2, 350], "196": [2, 350], "197": [2, 350] }, { "4": [2, 353], "5": [2, 353], "31": [2, 353], "32": [2, 353], "34": [2, 353], "35": [2, 353], "36": [2, 353], "37": [2, 353], "38": [2, 353], "39": [2, 353], "41": [2, 353], "42": [2, 353], "44": [2, 353], "45": [2, 353], "47": [2, 353], "48": [2, 353], "49": [2, 353], "50": [2, 353], "90": [2, 353], "96": [2, 353], "97": [2, 353], "118": [2, 353], "131": [2, 353], "195": [2, 353], "196": [2, 353], "197": [2, 353] }, { "4": [2, 354], "5": [2, 354], "31": [2, 354], "32": [2, 354], "34": [2, 354], "35": [2, 354], "36": [2, 354], "37": [2, 354], "38": [2, 354], "39": [2, 354], "41": [2, 354], "42": [2, 354], "44": [2, 354], "45": [2, 354], "47": [2, 354], "48": [2, 354], "49": [2, 354], "50": [2, 354], "90": [2, 354], "96": [2, 354], "97": [2, 354], "118": [2, 354], "131": [2, 354], "195": [2, 354], "196": [2, 354], "197": [2, 354] }, { "4": [2, 200], "5": [2, 200], "32": [1, 413], "34": [2, 200], "38": [2, 200], "42": [2, 200], "44": [2, 200], "48": [2, 200], "90": [2, 200], "131": [2, 200], "195": [2, 200], "196": [2, 200] }, { "4": [2, 202], "5": [2, 202], "32": [2, 202], "34": [2, 202], "38": [2, 202], "42": [2, 202], "44": [2, 202], "48": [2, 202], "90": [2, 202], "131": [2, 202], "195": [2, 202], "196": [2, 202] }, { "4": [2, 204], "5": [2, 204], "32": [2, 204], "34": [2, 204], "38": [2, 204], "42": [2, 204], "44": [2, 204], "48": [2, 204], "90": [2, 204], "131": [2, 204], "195": [2, 204], "196": [2, 204] }, { "4": [2, 205], "5": [2, 205], "32": [2, 205], "34": [2, 205], "38": [2, 205], "42": [2, 205], "44": [2, 205], "48": [2, 205], "90": [2, 205], "131": [2, 205], "195": [2, 205], "196": [2, 205] }, { "33": [1, 415], "95": [1, 214], "110": [1, 107], "111": [1, 108], "124": [1, 104], "130": [1, 213], "141": [1, 110], "142": [1, 111], "165": [1, 105], "176": [1, 103], "180": [1, 106], "186": [1, 212], "198": [1, 109], "199": [1, 112] }, { "4": [2, 256], "5": [2, 256], "32": [2, 256], "33": [1, 417], "34": [2, 256], "38": [2, 256], "42": [2, 256], "44": [2, 256], "48": [2, 256], "90": [2, 256], "96": [2, 256], "97": [2, 256], "118": [2, 256], "131": [2, 256], "195": [2, 256], "196": [2, 256] }, { "4": [2, 257], "5": [2, 257], "32": [2, 257], "34": [2, 257], "38": [2, 257], "42": [2, 257], "44": [2, 257], "48": [2, 257], "90": [2, 257], "96": [2, 257], "97": [2, 257], "118": [2, 257], "131": [2, 257], "195": [2, 257], "196": [2, 257] }, { "33": [1, 418] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "34": [1, 422] }, { "4": [2, 157], "5": [2, 157], "34": [2, 157], "44": [2, 157], "48": [1, 423], "196": [2, 157] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "33": [1, 42], "55": [1, 49], "69": [1, 431], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 430], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "175": [1, 43], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "33": [1, 42], "55": [1, 49], "69": [1, 436], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 439], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "175": [1, 43], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "4": [2, 187], "5": [2, 187], "34": [2, 187], "38": [2, 187], "44": [2, 187], "48": [2, 187], "89": [2, 187], "96": [2, 187], "131": [2, 187], "196": [2, 187] }, { "34": [1, 440], "69": [1, 95], "92": [1, 57], "94": [1, 443], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "4": [2, 189], "5": [2, 189], "34": [2, 189], "38": [2, 189], "40": [1, 445], "44": [2, 189], "48": [2, 189], "89": [2, 189], "96": [2, 189], "131": [2, 189], "196": [2, 189] }, { "69": [1, 394], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 399], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "33": [1, 448] }, { "4": [1, 25], "5": [1, 24], "147": [1, 451], "194": [1, 452] }, { "4": [1, 25], "5": [1, 24], "129": [1, 454], "194": [1, 456] }, { "55": [2, 76] }, { "1": [2, 22], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "44": [1, 461] }, { "69": [1, 394], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 399], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "42": [1, 463], "195": [1, 270] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 130], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "4": [2, 462], "5": [2, 462], "31": [2, 462], "32": [2, 462], "33": [1, 294], "34": [2, 462], "35": [2, 462], "36": [2, 462], "37": [2, 462], "38": [2, 462], "39": [2, 462], "40": [2, 462], "41": [2, 462], "42": [2, 462], "44": [2, 462], "45": [2, 462], "46": [2, 462], "47": [2, 462], "48": [2, 462], "49": [2, 462], "50": [2, 462], "85": [2, 462], "90": [2, 462], "96": [2, 462], "97": [2, 462], "118": [2, 462], "129": [2, 462], "131": [2, 462], "160": [2, 462], "195": [2, 462], "196": [2, 462], "197": [2, 462] }, { "37": [1, 276], "39": [1, 277], "45": [1, 281], "47": [1, 279], "49": [1, 282], "50": [1, 280], "197": [1, 283] }, { "34": [1, 465] }, { "34": [1, 466] }, { "34": [2, 275], "38": [2, 275], "42": [2, 275], "90": [2, 275], "97": [2, 275], "118": [2, 275], "196": [2, 275] }, { "4": [2, 272], "5": [2, 272], "31": [2, 272], "32": [2, 272], "34": [2, 272], "35": [2, 272], "36": [2, 272], "37": [2, 272], "38": [2, 272], "39": [2, 272], "40": [1, 293], "41": [2, 272], "42": [2, 272], "44": [2, 272], "45": [2, 272], "47": [2, 272], "48": [2, 272], "49": [2, 272], "50": [2, 272], "90": [2, 272], "96": [2, 272], "97": [2, 272], "118": [2, 272], "131": [2, 272], "160": [1, 467], "195": [2, 272], "196": [2, 272], "197": [2, 272] }, { "34": [2, 277], "38": [2, 277], "42": [2, 277], "90": [2, 277], "97": [2, 277], "118": [2, 277], "196": [2, 277] }, { "34": [2, 278], "37": [1, 469], "38": [2, 278], "39": [1, 470], "42": [2, 278], "90": [2, 278], "97": [2, 278], "118": [2, 278], "131": [1, 468], "196": [2, 278] }, { "4": [2, 280], "5": [2, 280], "31": [1, 473], "32": [2, 280], "34": [2, 280], "35": [1, 471], "37": [2, 280], "38": [2, 280], "39": [2, 280], "41": [1, 472], "42": [2, 280], "44": [2, 280], "48": [2, 280], "90": [2, 280], "97": [2, 280], "118": [2, 280], "131": [2, 280], "195": [2, 280], "196": [2, 280] }, { "4": [2, 283], "5": [2, 283], "31": [2, 283], "32": [2, 283], "34": [2, 283], "35": [2, 283], "36": [1, 474], "37": [2, 283], "38": [2, 283], "39": [2, 283], "41": [2, 283], "42": [2, 283], "44": [2, 283], "48": [2, 283], "90": [2, 283], "97": [2, 283], "118": [2, 283], "131": [2, 283], "195": [2, 283], "196": [2, 283] }, { "4": [2, 287], "5": [2, 287], "31": [2, 287], "32": [2, 287], "34": [2, 287], "35": [2, 287], "36": [2, 287], "37": [2, 287], "38": [2, 287], "39": [2, 287], "41": [2, 287], "42": [2, 287], "44": [2, 287], "48": [2, 287], "90": [2, 287], "97": [2, 287], "118": [2, 287], "131": [2, 287], "195": [2, 287], "196": [2, 287] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "4": [2, 264], "5": [2, 264], "31": [2, 264], "32": [2, 264], "34": [2, 264], "35": [2, 264], "36": [2, 264], "37": [2, 264], "38": [2, 264], "39": [2, 264], "41": [2, 264], "42": [2, 264], "44": [2, 264], "48": [2, 264], "90": [2, 264], "96": [2, 264], "97": [2, 264], "118": [2, 264], "131": [2, 264], "195": [2, 264], "196": [2, 264] }, { "4": [2, 265], "5": [2, 265], "31": [2, 265], "32": [2, 265], "34": [2, 265], "35": [2, 265], "36": [2, 265], "37": [2, 265], "38": [2, 265], "39": [2, 265], "41": [2, 265], "42": [2, 265], "44": [2, 265], "48": [2, 265], "90": [2, 265], "96": [2, 265], "97": [2, 265], "118": [2, 265], "131": [2, 265], "195": [2, 265], "196": [2, 265] }, { "4": [2, 266], "5": [2, 266], "31": [2, 266], "32": [2, 266], "34": [2, 266], "35": [2, 266], "36": [2, 266], "37": [2, 266], "38": [2, 266], "39": [2, 266], "41": [2, 266], "42": [2, 266], "44": [2, 266], "48": [2, 266], "90": [2, 266], "96": [2, 266], "97": [2, 266], "118": [2, 266], "131": [2, 266], "195": [2, 266], "196": [2, 266] }, { "4": [2, 267], "5": [2, 267], "31": [2, 267], "32": [2, 267], "34": [2, 267], "35": [2, 267], "36": [2, 267], "37": [2, 267], "38": [2, 267], "39": [2, 267], "41": [2, 267], "42": [2, 267], "44": [2, 267], "48": [2, 267], "90": [2, 267], "96": [2, 267], "97": [2, 267], "118": [2, 267], "131": [2, 267], "195": [2, 267], "196": [2, 267] }, { "4": [2, 268], "5": [2, 268], "31": [2, 268], "32": [2, 268], "34": [2, 268], "35": [2, 268], "36": [2, 268], "37": [2, 268], "38": [2, 268], "39": [2, 268], "41": [2, 268], "42": [2, 268], "44": [2, 268], "48": [2, 268], "90": [2, 268], "96": [2, 268], "97": [2, 268], "118": [2, 268], "131": [2, 268], "195": [2, 268], "196": [2, 268] }, { "4": [2, 269], "5": [2, 269], "31": [2, 269], "32": [2, 269], "34": [2, 269], "35": [2, 269], "36": [2, 269], "37": [2, 269], "38": [2, 269], "39": [2, 269], "41": [2, 269], "42": [2, 269], "44": [2, 269], "48": [2, 269], "90": [2, 269], "96": [2, 269], "97": [2, 269], "118": [2, 269], "131": [2, 269], "195": [2, 269], "196": [2, 269] }, { "4": [2, 270], "5": [2, 270], "31": [2, 270], "32": [2, 270], "34": [2, 270], "35": [2, 270], "36": [2, 270], "37": [2, 270], "38": [2, 270], "39": [2, 270], "41": [2, 270], "42": [2, 270], "44": [2, 270], "48": [2, 270], "90": [2, 270], "96": [2, 270], "97": [2, 270], "118": [2, 270], "131": [2, 270], "195": [2, 270], "196": [2, 270] }, { "4": [2, 271], "5": [2, 271], "31": [2, 271], "32": [2, 271], "34": [2, 271], "35": [2, 271], "36": [2, 271], "37": [2, 271], "38": [2, 271], "39": [2, 271], "41": [2, 271], "42": [2, 271], "44": [2, 271], "48": [2, 271], "90": [2, 271], "96": [2, 271], "97": [2, 271], "118": [2, 271], "131": [2, 271], "195": [2, 271], "196": [2, 271] }, { "4": [2, 321], "5": [2, 321], "31": [2, 321], "32": [2, 321], "34": [2, 321], "35": [2, 321], "36": [2, 321], "37": [2, 321], "38": [2, 321], "39": [2, 321], "41": [2, 321], "42": [2, 321], "44": [2, 321], "48": [2, 321], "90": [2, 321], "96": [2, 321], "97": [2, 321], "118": [2, 321], "131": [2, 321], "195": [2, 321], "196": [2, 321] }, { "4": [2, 322], "5": [2, 322], "31": [2, 322], "32": [2, 322], "34": [2, 322], "35": [2, 322], "36": [2, 322], "37": [2, 322], "38": [2, 322], "39": [2, 322], "41": [2, 322], "42": [2, 322], "44": [2, 322], "48": [2, 322], "90": [2, 322], "96": [2, 322], "97": [2, 322], "118": [2, 322], "131": [2, 322], "195": [2, 322], "196": [2, 322] }, { "4": [2, 346], "5": [2, 346], "31": [2, 346], "32": [2, 346], "34": [2, 346], "35": [2, 346], "36": [2, 346], "37": [2, 346], "38": [2, 346], "39": [2, 346], "41": [2, 346], "42": [2, 346], "44": [2, 346], "48": [2, 346], "90": [2, 346], "96": [2, 346], "97": [2, 346], "118": [2, 346], "131": [2, 346], "195": [2, 346], "196": [2, 346] }, { "4": [2, 347], "5": [2, 347], "31": [2, 347], "32": [2, 347], "34": [2, 347], "35": [2, 347], "36": [2, 347], "37": [2, 347], "38": [2, 347], "39": [2, 347], "41": [2, 347], "42": [2, 347], "44": [2, 347], "48": [2, 347], "90": [2, 347], "96": [2, 347], "97": [2, 347], "118": [2, 347], "131": [2, 347], "195": [2, 347], "196": [2, 347] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [2, 156], "5": [2, 156], "44": [2, 156] }, { "4": [2, 159], "5": [2, 159], "44": [2, 159] }, { "4": [2, 161], "5": [2, 161], "34": [2, 161], "44": [2, 161], "48": [2, 161], "196": [2, 161] }, { "4": [2, 173], "5": [2, 173], "34": [2, 173], "38": [2, 173], "44": [2, 173], "48": [2, 173], "131": [2, 173], "195": [1, 270], "196": [2, 173] }, { "4": [2, 174], "5": [2, 174], "34": [2, 174], "38": [2, 174], "44": [2, 174], "48": [2, 174], "131": [2, 174], "195": [1, 270], "196": [2, 174] }, { "160": [1, 479] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "160": [1, 481] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 484], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 484], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "3": [1, 490], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 484], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 484], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 495], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [2, 235], "9": [2, 235], "12": [2, 235], "15": [2, 235], "18": [2, 235], "20": [2, 235], "21": [2, 235], "22": [2, 235], "23": [2, 235], "25": [2, 235], "27": [2, 235], "28": [2, 235], "33": [2, 235], "39": [2, 235], "55": [2, 235], "56": [2, 235], "59": [2, 235], "61": [2, 235], "63": [2, 235], "67": [2, 235], "68": [2, 235], "69": [2, 235], "70": [2, 235], "72": [2, 235], "74": [2, 235], "77": [2, 235], "78": [2, 235], "79": [2, 235], "80": [2, 235], "82": [2, 235], "84": [2, 235], "87": [2, 235], "88": [2, 235], "89": [2, 235], "92": [2, 235], "95": [2, 235], "97": [2, 235], "98": [2, 235], "100": [2, 235], "101": [2, 235], "108": [2, 235], "110": [2, 235], "111": [2, 235], "113": [2, 235], "115": [2, 235], "118": [2, 235], "119": [2, 235], "122": [2, 235], "123": [2, 235], "124": [2, 235], "125": [2, 235], "126": [2, 235], "127": [2, 235], "130": [2, 235], "134": [2, 235], "135": [2, 235], "141": [2, 235], "142": [2, 235], "145": [2, 235], "147": [2, 235], "149": [2, 235], "150": [2, 235], "151": [2, 235], "154": [2, 235], "155": [2, 235], "156": [2, 235], "159": [2, 235], "165": [2, 235], "167": [2, 235], "170": [2, 235], "172": [2, 235], "174": [2, 235], "176": [2, 235], "178": [2, 235], "179": [2, 235], "180": [2, 235], "181": [2, 235], "186": [2, 235], "194": [2, 235], "198": [2, 235], "199": [2, 235] }, { "6": [2, 236], "9": [2, 236], "12": [2, 236], "15": [2, 236], "18": [2, 236], "20": [2, 236], "21": [2, 236], "22": [2, 236], "23": [2, 236], "25": [2, 236], "27": [2, 236], "28": [2, 236], "33": [2, 236], "39": [2, 236], "55": [2, 236], "56": [2, 236], "59": [2, 236], "61": [2, 236], "63": [2, 236], "67": [2, 236], "68": [2, 236], "69": [2, 236], "70": [2, 236], "72": [2, 236], "74": [2, 236], "77": [2, 236], "78": [2, 236], "79": [2, 236], "80": [2, 236], "82": [2, 236], "84": [2, 236], "87": [2, 236], "88": [2, 236], "89": [2, 236], "92": [2, 236], "95": [2, 236], "97": [2, 236], "98": [2, 236], "100": [2, 236], "101": [2, 236], "108": [2, 236], "110": [2, 236], "111": [2, 236], "113": [2, 236], "115": [2, 236], "118": [2, 236], "119": [2, 236], "122": [2, 236], "123": [2, 236], "124": [2, 236], "125": [2, 236], "126": [2, 236], "127": [2, 236], "130": [2, 236], "134": [2, 236], "135": [2, 236], "141": [2, 236], "142": [2, 236], "145": [2, 236], "147": [2, 236], "149": [2, 236], "150": [2, 236], "151": [2, 236], "154": [2, 236], "155": [2, 236], "156": [2, 236], "159": [2, 236], "165": [2, 236], "167": [2, 236], "170": [2, 236], "172": [2, 236], "174": [2, 236], "176": [2, 236], "178": [2, 236], "179": [2, 236], "180": [2, 236], "181": [2, 236], "186": [2, 236], "194": [2, 236], "198": [2, 236], "199": [2, 236] }, { "6": [2, 237], "9": [2, 237], "12": [2, 237], "15": [2, 237], "18": [2, 237], "20": [2, 237], "21": [2, 237], "22": [2, 237], "23": [2, 237], "25": [2, 237], "27": [2, 237], "28": [2, 237], "33": [2, 237], "39": [2, 237], "55": [2, 237], "56": [2, 237], "59": [2, 237], "61": [2, 237], "63": [2, 237], "67": [2, 237], "68": [2, 237], "69": [2, 237], "70": [2, 237], "72": [2, 237], "74": [2, 237], "77": [2, 237], "78": [2, 237], "79": [2, 237], "80": [2, 237], "82": [2, 237], "84": [2, 237], "87": [2, 237], "88": [2, 237], "89": [2, 237], "92": [2, 237], "95": [2, 237], "97": [2, 237], "98": [2, 237], "100": [2, 237], "101": [2, 237], "108": [2, 237], "110": [2, 237], "111": [2, 237], "113": [2, 237], "115": [2, 237], "118": [2, 237], "119": [2, 237], "122": [2, 237], "123": [2, 237], "124": [2, 237], "125": [2, 237], "126": [2, 237], "127": [2, 237], "130": [2, 237], "134": [2, 237], "135": [2, 237], "141": [2, 237], "142": [2, 237], "145": [2, 237], "147": [2, 237], "149": [2, 237], "150": [2, 237], "151": [2, 237], "154": [2, 237], "155": [2, 237], "156": [2, 237], "159": [2, 237], "165": [2, 237], "167": [2, 237], "170": [2, 237], "172": [2, 237], "174": [2, 237], "176": [2, 237], "178": [2, 237], "179": [2, 237], "180": [2, 237], "181": [2, 237], "186": [2, 237], "194": [2, 237], "198": [2, 237], "199": [2, 237] }, { "6": [2, 238], "9": [2, 238], "12": [2, 238], "15": [2, 238], "18": [2, 238], "20": [2, 238], "21": [2, 238], "22": [2, 238], "23": [2, 238], "25": [2, 238], "27": [2, 238], "28": [2, 238], "33": [2, 238], "39": [2, 238], "55": [2, 238], "56": [2, 238], "59": [2, 238], "61": [2, 238], "63": [2, 238], "67": [2, 238], "68": [2, 238], "69": [2, 238], "70": [2, 238], "72": [2, 238], "74": [2, 238], "77": [2, 238], "78": [2, 238], "79": [2, 238], "80": [2, 238], "82": [2, 238], "84": [2, 238], "87": [2, 238], "88": [2, 238], "89": [2, 238], "92": [2, 238], "95": [2, 238], "97": [2, 238], "98": [2, 238], "100": [2, 238], "101": [2, 238], "108": [2, 238], "110": [2, 238], "111": [2, 238], "113": [2, 238], "115": [2, 238], "118": [2, 238], "119": [2, 238], "122": [2, 238], "123": [2, 238], "124": [2, 238], "125": [2, 238], "126": [2, 238], "127": [2, 238], "130": [2, 238], "134": [2, 238], "135": [2, 238], "141": [2, 238], "142": [2, 238], "145": [2, 238], "147": [2, 238], "149": [2, 238], "150": [2, 238], "151": [2, 238], "154": [2, 238], "155": [2, 238], "156": [2, 238], "159": [2, 238], "165": [2, 238], "167": [2, 238], "170": [2, 238], "172": [2, 238], "174": [2, 238], "176": [2, 238], "178": [2, 238], "179": [2, 238], "180": [2, 238], "181": [2, 238], "186": [2, 238], "194": [2, 238], "198": [2, 238], "199": [2, 238] }, { "6": [2, 239], "9": [2, 239], "12": [2, 239], "15": [2, 239], "18": [2, 239], "20": [2, 239], "21": [2, 239], "22": [2, 239], "23": [2, 239], "25": [2, 239], "27": [2, 239], "28": [2, 239], "33": [2, 239], "39": [2, 239], "55": [2, 239], "56": [2, 239], "59": [2, 239], "61": [2, 239], "63": [2, 239], "67": [2, 239], "68": [2, 239], "69": [2, 239], "70": [2, 239], "72": [2, 239], "74": [2, 239], "77": [2, 239], "78": [2, 239], "79": [2, 239], "80": [2, 239], "82": [2, 239], "84": [2, 239], "87": [2, 239], "88": [2, 239], "89": [2, 239], "92": [2, 239], "95": [2, 239], "97": [2, 239], "98": [2, 239], "100": [2, 239], "101": [2, 239], "108": [2, 239], "110": [2, 239], "111": [2, 239], "113": [2, 239], "115": [2, 239], "118": [2, 239], "119": [2, 239], "122": [2, 239], "123": [2, 239], "124": [2, 239], "125": [2, 239], "126": [2, 239], "127": [2, 239], "130": [2, 239], "134": [2, 239], "135": [2, 239], "141": [2, 239], "142": [2, 239], "145": [2, 239], "147": [2, 239], "149": [2, 239], "150": [2, 239], "151": [2, 239], "154": [2, 239], "155": [2, 239], "156": [2, 239], "159": [2, 239], "165": [2, 239], "167": [2, 239], "170": [2, 239], "172": [2, 239], "174": [2, 239], "176": [2, 239], "178": [2, 239], "179": [2, 239], "180": [2, 239], "181": [2, 239], "186": [2, 239], "194": [2, 239], "198": [2, 239], "199": [2, 239] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 484], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 484], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 484], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 484], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "34": [1, 500] }, { "34": [1, 501], "195": [1, 270] }, { "34": [2, 215], "37": [1, 276], "38": [2, 215], "39": [1, 277], "45": [1, 281], "47": [1, 279], "49": [1, 282], "50": [1, 280], "90": [2, 215], "96": [2, 215], "131": [1, 275], "197": [1, 283] }, { "4": [2, 272], "5": [2, 272], "31": [2, 272], "32": [2, 272], "34": [2, 272], "35": [2, 272], "36": [2, 272], "37": [2, 272], "38": [2, 272], "39": [2, 272], "40": [1, 293], "41": [2, 272], "42": [2, 272], "44": [2, 272], "45": [2, 272], "47": [2, 272], "48": [2, 272], "49": [2, 272], "50": [2, 272], "90": [2, 272], "96": [2, 272], "97": [2, 272], "118": [2, 272], "131": [2, 272], "160": [1, 502], "195": [2, 272], "196": [2, 272], "197": [2, 272] }, { "34": [2, 214] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "4": [2, 299], "5": [2, 299], "31": [2, 299], "32": [2, 299], "34": [2, 299], "35": [2, 299], "36": [2, 299], "37": [2, 299], "38": [2, 299], "39": [2, 299], "41": [2, 299], "42": [2, 299], "44": [2, 299], "45": [2, 299], "47": [2, 299], "48": [2, 299], "49": [2, 299], "50": [2, 299], "90": [2, 299], "96": [2, 299], "97": [2, 299], "118": [2, 299], "131": [2, 299], "195": [2, 299], "196": [2, 299], "197": [2, 299] }, { "38": [1, 509], "90": [1, 508] }, { "34": [2, 344], "38": [2, 344], "90": [2, 344] }, { "4": [2, 301], "5": [2, 301], "31": [2, 301], "32": [2, 301], "34": [2, 301], "35": [2, 301], "36": [2, 301], "37": [2, 301], "38": [2, 301], "39": [2, 301], "41": [2, 301], "42": [2, 301], "44": [2, 301], "45": [2, 301], "47": [2, 301], "48": [2, 301], "49": [2, 301], "50": [2, 301], "90": [2, 301], "96": [2, 301], "97": [2, 301], "118": [2, 301], "131": [2, 301], "195": [2, 301], "196": [2, 301], "197": [2, 301] }, { "38": [1, 511], "196": [1, 510] }, { "40": [1, 293], "46": [1, 512] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "33": [1, 514] }, { "33": [1, 515] }, { "33": [1, 516] }, { "33": [1, 517] }, { "33": [1, 518] }, { "51": [1, 378], "52": [1, 379], "53": [1, 380], "54": [1, 381], "57": [1, 382], "58": [1, 383], "62": [1, 384], "65": [1, 385], "66": [1, 386], "69": [1, 394], "71": [1, 387], "73": [1, 388], "75": [1, 389], "76": [1, 390], "81": [1, 391], "83": [1, 392], "86": [1, 393], "91": [1, 324], "92": [1, 57], "93": [1, 372], "94": [1, 325], "95": [1, 326], "96": [1, 327], "97": [1, 74], "98": [1, 79], "99": [1, 339], "100": [1, 66], "101": [1, 67], "102": [1, 373], "103": [1, 340], "104": [1, 341], "105": [1, 342], "106": [1, 313], "107": [1, 374], "108": [1, 53], "109": [1, 343], "110": [1, 107], "111": [1, 108], "112": [1, 344], "113": [1, 399], "114": [1, 328], "115": [1, 68], "116": [1, 345], "117": [1, 346], "118": [1, 75], "119": [1, 101], "120": [1, 347], "121": [1, 375], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 314], "127": [1, 52], "128": [1, 348], "129": [1, 315], "130": [1, 323], "131": [1, 329], "132": [1, 349], "133": [1, 350], "134": [1, 71], "135": [1, 59], "136": [1, 330], "137": [1, 351], "138": [1, 352], "139": [1, 316], "140": [1, 317], "141": [1, 110], "142": [1, 111], "143": [1, 353], "144": [1, 354], "145": [1, 72], "146": [1, 376], "147": [1, 60], "148": [1, 318], "149": [1, 61], "150": [1, 77], "151": [1, 76], "152": [1, 331], "153": [1, 332], "154": [1, 62], "155": [1, 54], "156": [1, 319], "157": [1, 333], "158": [1, 334], "159": [1, 320], "160": [1, 321], "161": [1, 377], "162": [1, 336], "163": [1, 335], "164": [1, 355], "165": [1, 105], "166": [1, 356], "167": [1, 63], "168": [1, 357], "169": [1, 358], "170": [1, 64], "171": [1, 337], "172": [1, 55], "173": [1, 359], "174": [1, 102], "175": [1, 338], "176": [1, 103], "177": [1, 360], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "182": [1, 361], "183": [1, 362], "184": [1, 363], "185": [1, 364], "186": [1, 322], "187": [1, 365], "188": [1, 366], "189": [1, 367], "190": [1, 368], "191": [1, 369], "192": [1, 370], "193": [1, 371], "198": [1, 109], "199": [1, 112] }, { "4": [2, 390], "5": [2, 390], "31": [2, 390], "32": [2, 390], "34": [2, 390], "35": [2, 390], "36": [2, 390], "37": [2, 390], "38": [2, 390], "39": [2, 390], "41": [2, 390], "42": [2, 390], "44": [2, 390], "45": [2, 390], "47": [2, 390], "48": [2, 390], "49": [2, 390], "50": [2, 390], "90": [2, 390], "96": [2, 390], "97": [2, 390], "118": [2, 390], "131": [2, 390], "195": [2, 390], "196": [2, 390], "197": [2, 390] }, { "4": [2, 484], "5": [2, 484], "31": [2, 484], "32": [2, 484], "34": [2, 484], "35": [2, 484], "36": [2, 484], "37": [2, 484], "38": [2, 484], "39": [2, 484], "41": [2, 484], "42": [2, 484], "44": [2, 484], "45": [2, 484], "47": [2, 484], "48": [2, 484], "49": [2, 484], "50": [2, 484], "90": [2, 484], "96": [2, 484], "97": [2, 484], "118": [2, 484], "131": [2, 484], "195": [2, 484], "196": [2, 484], "197": [2, 484] }, { "4": [2, 485], "5": [2, 485], "31": [2, 485], "32": [2, 485], "34": [2, 485], "35": [2, 485], "36": [2, 485], "37": [2, 485], "38": [2, 485], "39": [2, 485], "41": [2, 485], "42": [2, 485], "44": [2, 485], "45": [2, 485], "47": [2, 485], "48": [2, 485], "49": [2, 485], "50": [2, 485], "90": [2, 485], "96": [2, 485], "97": [2, 485], "118": [2, 485], "131": [2, 485], "195": [2, 485], "196": [2, 485], "197": [2, 485] }, { "4": [2, 486], "5": [2, 486], "31": [2, 486], "32": [2, 486], "34": [2, 486], "35": [2, 486], "36": [2, 486], "37": [2, 486], "38": [2, 486], "39": [2, 486], "41": [2, 486], "42": [2, 486], "44": [2, 486], "45": [2, 486], "47": [2, 486], "48": [2, 486], "49": [2, 486], "50": [2, 486], "90": [2, 486], "96": [2, 486], "97": [2, 486], "118": [2, 486], "131": [2, 486], "195": [2, 486], "196": [2, 486], "197": [2, 486] }, { "4": [2, 487], "5": [2, 487], "31": [2, 487], "32": [2, 487], "34": [2, 487], "35": [2, 487], "36": [2, 487], "37": [2, 487], "38": [2, 487], "39": [2, 487], "41": [2, 487], "42": [2, 487], "44": [2, 487], "45": [2, 487], "47": [2, 487], "48": [2, 487], "49": [2, 487], "50": [2, 487], "90": [2, 487], "96": [2, 487], "97": [2, 487], "118": [2, 487], "131": [2, 487], "195": [2, 487], "196": [2, 487], "197": [2, 487] }, { "4": [2, 488], "5": [2, 488], "31": [2, 488], "32": [2, 488], "34": [2, 488], "35": [2, 488], "36": [2, 488], "37": [2, 488], "38": [2, 488], "39": [2, 488], "41": [2, 488], "42": [2, 488], "44": [2, 488], "45": [2, 488], "47": [2, 488], "48": [2, 488], "49": [2, 488], "50": [2, 488], "90": [2, 488], "96": [2, 488], "97": [2, 488], "118": [2, 488], "131": [2, 488], "195": [2, 488], "196": [2, 488], "197": [2, 488] }, { "4": [2, 489], "5": [2, 489], "31": [2, 489], "32": [2, 489], "34": [2, 489], "35": [2, 489], "36": [2, 489], "37": [2, 489], "38": [2, 489], "39": [2, 489], "41": [2, 489], "42": [2, 489], "44": [2, 489], "45": [2, 489], "47": [2, 489], "48": [2, 489], "49": [2, 489], "50": [2, 489], "90": [2, 489], "96": [2, 489], "97": [2, 489], "118": [2, 489], "131": [2, 489], "195": [2, 489], "196": [2, 489], "197": [2, 489] }, { "4": [2, 490], "5": [2, 490], "31": [2, 490], "32": [2, 490], "34": [2, 490], "35": [2, 490], "36": [2, 490], "37": [2, 490], "38": [2, 490], "39": [2, 490], "41": [2, 490], "42": [2, 490], "44": [2, 490], "45": [2, 490], "47": [2, 490], "48": [2, 490], "49": [2, 490], "50": [2, 490], "90": [2, 490], "96": [2, 490], "97": [2, 490], "118": [2, 490], "131": [2, 490], "195": [2, 490], "196": [2, 490], "197": [2, 490] }, { "4": [2, 491], "5": [2, 491], "31": [2, 491], "32": [2, 491], "34": [2, 491], "35": [2, 491], "36": [2, 491], "37": [2, 491], "38": [2, 491], "39": [2, 491], "41": [2, 491], "42": [2, 491], "44": [2, 491], "45": [2, 491], "47": [2, 491], "48": [2, 491], "49": [2, 491], "50": [2, 491], "90": [2, 491], "96": [2, 491], "97": [2, 491], "118": [2, 491], "131": [2, 491], "195": [2, 491], "196": [2, 491], "197": [2, 491] }, { "4": [2, 492], "5": [2, 492], "31": [2, 492], "32": [2, 492], "34": [2, 492], "35": [2, 492], "36": [2, 492], "37": [2, 492], "38": [2, 492], "39": [2, 492], "41": [2, 492], "42": [2, 492], "44": [2, 492], "45": [2, 492], "47": [2, 492], "48": [2, 492], "49": [2, 492], "50": [2, 492], "90": [2, 492], "96": [2, 492], "97": [2, 492], "118": [2, 492], "131": [2, 492], "195": [2, 492], "196": [2, 492], "197": [2, 492] }, { "4": [2, 493], "5": [2, 493], "31": [2, 493], "32": [2, 493], "34": [2, 493], "35": [2, 493], "36": [2, 493], "37": [2, 493], "38": [2, 493], "39": [2, 493], "41": [2, 493], "42": [2, 493], "44": [2, 493], "45": [2, 493], "47": [2, 493], "48": [2, 493], "49": [2, 493], "50": [2, 493], "90": [2, 493], "96": [2, 493], "97": [2, 493], "118": [2, 493], "131": [2, 493], "195": [2, 493], "196": [2, 493], "197": [2, 493] }, { "4": [2, 494], "5": [2, 494], "31": [2, 494], "32": [2, 494], "34": [2, 494], "35": [2, 494], "36": [2, 494], "37": [2, 494], "38": [2, 494], "39": [2, 494], "41": [2, 494], "42": [2, 494], "44": [2, 494], "45": [2, 494], "47": [2, 494], "48": [2, 494], "49": [2, 494], "50": [2, 494], "90": [2, 494], "96": [2, 494], "97": [2, 494], "118": [2, 494], "131": [2, 494], "195": [2, 494], "196": [2, 494], "197": [2, 494] }, { "4": [2, 495], "5": [2, 495], "31": [2, 495], "32": [2, 495], "34": [2, 495], "35": [2, 495], "36": [2, 495], "37": [2, 495], "38": [2, 495], "39": [2, 495], "41": [2, 495], "42": [2, 495], "44": [2, 495], "45": [2, 495], "47": [2, 495], "48": [2, 495], "49": [2, 495], "50": [2, 495], "90": [2, 495], "96": [2, 495], "97": [2, 495], "118": [2, 495], "131": [2, 495], "195": [2, 495], "196": [2, 495], "197": [2, 495] }, { "4": [2, 496], "5": [2, 496], "31": [2, 496], "32": [2, 496], "34": [2, 496], "35": [2, 496], "36": [2, 496], "37": [2, 496], "38": [2, 496], "39": [2, 496], "41": [2, 496], "42": [2, 496], "44": [2, 496], "45": [2, 496], "47": [2, 496], "48": [2, 496], "49": [2, 496], "50": [2, 496], "90": [2, 496], "96": [2, 496], "97": [2, 496], "118": [2, 496], "131": [2, 496], "195": [2, 496], "196": [2, 496], "197": [2, 496] }, { "4": [2, 497], "5": [2, 497], "31": [2, 497], "32": [2, 497], "34": [2, 497], "35": [2, 497], "36": [2, 497], "37": [2, 497], "38": [2, 497], "39": [2, 497], "41": [2, 497], "42": [2, 497], "44": [2, 497], "45": [2, 497], "47": [2, 497], "48": [2, 497], "49": [2, 497], "50": [2, 497], "90": [2, 497], "96": [2, 497], "97": [2, 497], "118": [2, 497], "131": [2, 497], "195": [2, 497], "196": [2, 497], "197": [2, 497] }, { "4": [2, 498], "5": [2, 498], "31": [2, 498], "32": [2, 498], "34": [2, 498], "35": [2, 498], "36": [2, 498], "37": [2, 498], "38": [2, 498], "39": [2, 498], "41": [2, 498], "42": [2, 498], "44": [2, 498], "45": [2, 498], "47": [2, 498], "48": [2, 498], "49": [2, 498], "50": [2, 498], "90": [2, 498], "96": [2, 498], "97": [2, 498], "118": [2, 498], "131": [2, 498], "195": [2, 498], "196": [2, 498], "197": [2, 498] }, { "4": [2, 499], "5": [2, 499], "31": [2, 499], "32": [2, 499], "34": [2, 499], "35": [2, 499], "36": [2, 499], "37": [2, 499], "38": [2, 499], "39": [2, 499], "41": [2, 499], "42": [2, 499], "44": [2, 499], "45": [2, 499], "47": [2, 499], "48": [2, 499], "49": [2, 499], "50": [2, 499], "90": [2, 499], "96": [2, 499], "97": [2, 499], "118": [2, 499], "131": [2, 499], "195": [2, 499], "196": [2, 499], "197": [2, 499] }, { "4": [2, 500], "5": [2, 500], "31": [2, 500], "32": [2, 500], "34": [2, 500], "35": [2, 500], "36": [2, 500], "37": [2, 500], "38": [2, 500], "39": [2, 500], "41": [2, 500], "42": [2, 500], "44": [2, 500], "45": [2, 500], "47": [2, 500], "48": [2, 500], "49": [2, 500], "50": [2, 500], "90": [2, 500], "96": [2, 500], "97": [2, 500], "118": [2, 500], "131": [2, 500], "195": [2, 500], "196": [2, 500], "197": [2, 500] }, { "4": [2, 501], "5": [2, 501], "31": [2, 501], "32": [2, 501], "34": [2, 501], "35": [2, 501], "36": [2, 501], "37": [2, 501], "38": [2, 501], "39": [2, 501], "41": [2, 501], "42": [2, 501], "44": [2, 501], "45": [2, 501], "47": [2, 501], "48": [2, 501], "49": [2, 501], "50": [2, 501], "90": [2, 501], "96": [2, 501], "97": [2, 501], "118": [2, 501], "131": [2, 501], "195": [2, 501], "196": [2, 501], "197": [2, 501] }, { "4": [2, 502], "5": [2, 502], "31": [2, 502], "32": [2, 502], "34": [2, 502], "35": [2, 502], "36": [2, 502], "37": [2, 502], "38": [2, 502], "39": [2, 502], "41": [2, 502], "42": [2, 502], "44": [2, 502], "45": [2, 502], "47": [2, 502], "48": [2, 502], "49": [2, 502], "50": [2, 502], "90": [2, 502], "96": [2, 502], "97": [2, 502], "118": [2, 502], "131": [2, 502], "195": [2, 502], "196": [2, 502], "197": [2, 502] }, { "4": [2, 503], "5": [2, 503], "31": [2, 503], "32": [2, 503], "34": [2, 503], "35": [2, 503], "36": [2, 503], "37": [2, 503], "38": [2, 503], "39": [2, 503], "41": [2, 503], "42": [2, 503], "44": [2, 503], "45": [2, 503], "47": [2, 503], "48": [2, 503], "49": [2, 503], "50": [2, 503], "90": [2, 503], "96": [2, 503], "97": [2, 503], "118": [2, 503], "131": [2, 503], "195": [2, 503], "196": [2, 503], "197": [2, 503] }, { "4": [2, 504], "5": [2, 504], "31": [2, 504], "32": [2, 504], "34": [2, 504], "35": [2, 504], "36": [2, 504], "37": [2, 504], "38": [2, 504], "39": [2, 504], "41": [2, 504], "42": [2, 504], "44": [2, 504], "45": [2, 504], "47": [2, 504], "48": [2, 504], "49": [2, 504], "50": [2, 504], "90": [2, 504], "96": [2, 504], "97": [2, 504], "118": [2, 504], "131": [2, 504], "195": [2, 504], "196": [2, 504], "197": [2, 504] }, { "4": [2, 505], "5": [2, 505], "31": [2, 505], "32": [2, 505], "34": [2, 505], "35": [2, 505], "36": [2, 505], "37": [2, 505], "38": [2, 505], "39": [2, 505], "41": [2, 505], "42": [2, 505], "44": [2, 505], "45": [2, 505], "47": [2, 505], "48": [2, 505], "49": [2, 505], "50": [2, 505], "90": [2, 505], "96": [2, 505], "97": [2, 505], "118": [2, 505], "131": [2, 505], "195": [2, 505], "196": [2, 505], "197": [2, 505] }, { "4": [2, 506], "5": [2, 506], "31": [2, 506], "32": [2, 506], "34": [2, 506], "35": [2, 506], "36": [2, 506], "37": [2, 506], "38": [2, 506], "39": [2, 506], "41": [2, 506], "42": [2, 506], "44": [2, 506], "45": [2, 506], "47": [2, 506], "48": [2, 506], "49": [2, 506], "50": [2, 506], "90": [2, 506], "96": [2, 506], "97": [2, 506], "118": [2, 506], "131": [2, 506], "195": [2, 506], "196": [2, 506], "197": [2, 506] }, { "4": [2, 507], "5": [2, 507], "31": [2, 507], "32": [2, 507], "34": [2, 507], "35": [2, 507], "36": [2, 507], "37": [2, 507], "38": [2, 507], "39": [2, 507], "41": [2, 507], "42": [2, 507], "44": [2, 507], "45": [2, 507], "47": [2, 507], "48": [2, 507], "49": [2, 507], "50": [2, 507], "90": [2, 507], "96": [2, 507], "97": [2, 507], "118": [2, 507], "131": [2, 507], "195": [2, 507], "196": [2, 507], "197": [2, 507] }, { "4": [2, 508], "5": [2, 508], "31": [2, 508], "32": [2, 508], "34": [2, 508], "35": [2, 508], "36": [2, 508], "37": [2, 508], "38": [2, 508], "39": [2, 508], "41": [2, 508], "42": [2, 508], "44": [2, 508], "45": [2, 508], "47": [2, 508], "48": [2, 508], "49": [2, 508], "50": [2, 508], "90": [2, 508], "96": [2, 508], "97": [2, 508], "118": [2, 508], "131": [2, 508], "195": [2, 508], "196": [2, 508], "197": [2, 508] }, { "4": [2, 509], "5": [2, 509], "31": [2, 509], "32": [2, 509], "34": [2, 509], "35": [2, 509], "36": [2, 509], "37": [2, 509], "38": [2, 509], "39": [2, 509], "41": [2, 509], "42": [2, 509], "44": [2, 509], "45": [2, 509], "47": [2, 509], "48": [2, 509], "49": [2, 509], "50": [2, 509], "90": [2, 509], "96": [2, 509], "97": [2, 509], "118": [2, 509], "131": [2, 509], "195": [2, 509], "196": [2, 509], "197": [2, 509] }, { "4": [2, 510], "5": [2, 510], "31": [2, 510], "32": [2, 510], "34": [2, 510], "35": [2, 510], "36": [2, 510], "37": [2, 510], "38": [2, 510], "39": [2, 510], "41": [2, 510], "42": [2, 510], "44": [2, 510], "45": [2, 510], "47": [2, 510], "48": [2, 510], "49": [2, 510], "50": [2, 510], "90": [2, 510], "96": [2, 510], "97": [2, 510], "118": [2, 510], "131": [2, 510], "195": [2, 510], "196": [2, 510], "197": [2, 510] }, { "4": [2, 511], "5": [2, 511], "31": [2, 511], "32": [2, 511], "34": [2, 511], "35": [2, 511], "36": [2, 511], "37": [2, 511], "38": [2, 511], "39": [2, 511], "41": [2, 511], "42": [2, 511], "44": [2, 511], "45": [2, 511], "47": [2, 511], "48": [2, 511], "49": [2, 511], "50": [2, 511], "90": [2, 511], "96": [2, 511], "97": [2, 511], "118": [2, 511], "131": [2, 511], "195": [2, 511], "196": [2, 511], "197": [2, 511] }, { "4": [2, 512], "5": [2, 512], "31": [2, 512], "32": [2, 512], "34": [2, 512], "35": [2, 512], "36": [2, 512], "37": [2, 512], "38": [2, 512], "39": [2, 512], "41": [2, 512], "42": [2, 512], "44": [2, 512], "45": [2, 512], "47": [2, 512], "48": [2, 512], "49": [2, 512], "50": [2, 512], "90": [2, 512], "96": [2, 512], "97": [2, 512], "118": [2, 512], "131": [2, 512], "195": [2, 512], "196": [2, 512], "197": [2, 512] }, { "4": [2, 513], "5": [2, 513], "31": [2, 513], "32": [2, 513], "34": [2, 513], "35": [2, 513], "36": [2, 513], "37": [2, 513], "38": [2, 513], "39": [2, 513], "41": [2, 513], "42": [2, 513], "44": [2, 513], "45": [2, 513], "47": [2, 513], "48": [2, 513], "49": [2, 513], "50": [2, 513], "90": [2, 513], "96": [2, 513], "97": [2, 513], "118": [2, 513], "131": [2, 513], "195": [2, 513], "196": [2, 513], "197": [2, 513] }, { "4": [2, 514], "5": [2, 514], "31": [2, 514], "32": [2, 514], "34": [2, 514], "35": [2, 514], "36": [2, 514], "37": [2, 514], "38": [2, 514], "39": [2, 514], "41": [2, 514], "42": [2, 514], "44": [2, 514], "45": [2, 514], "47": [2, 514], "48": [2, 514], "49": [2, 514], "50": [2, 514], "90": [2, 514], "96": [2, 514], "97": [2, 514], "118": [2, 514], "131": [2, 514], "195": [2, 514], "196": [2, 514], "197": [2, 514] }, { "4": [2, 515], "5": [2, 515], "31": [2, 515], "32": [2, 515], "34": [2, 515], "35": [2, 515], "36": [2, 515], "37": [2, 515], "38": [2, 515], "39": [2, 515], "41": [2, 515], "42": [2, 515], "44": [2, 515], "45": [2, 515], "47": [2, 515], "48": [2, 515], "49": [2, 515], "50": [2, 515], "90": [2, 515], "96": [2, 515], "97": [2, 515], "118": [2, 515], "131": [2, 515], "195": [2, 515], "196": [2, 515], "197": [2, 515] }, { "4": [2, 516], "5": [2, 516], "31": [2, 516], "32": [2, 516], "34": [2, 516], "35": [2, 516], "36": [2, 516], "37": [2, 516], "38": [2, 516], "39": [2, 516], "41": [2, 516], "42": [2, 516], "44": [2, 516], "45": [2, 516], "47": [2, 516], "48": [2, 516], "49": [2, 516], "50": [2, 516], "90": [2, 516], "96": [2, 516], "97": [2, 516], "118": [2, 516], "131": [2, 516], "195": [2, 516], "196": [2, 516], "197": [2, 516] }, { "4": [2, 517], "5": [2, 517], "31": [2, 517], "32": [2, 517], "34": [2, 517], "35": [2, 517], "36": [2, 517], "37": [2, 517], "38": [2, 517], "39": [2, 517], "41": [2, 517], "42": [2, 517], "44": [2, 517], "45": [2, 517], "47": [2, 517], "48": [2, 517], "49": [2, 517], "50": [2, 517], "90": [2, 517], "96": [2, 517], "97": [2, 517], "118": [2, 517], "131": [2, 517], "195": [2, 517], "196": [2, 517], "197": [2, 517] }, { "4": [2, 518], "5": [2, 518], "31": [2, 518], "32": [2, 518], "34": [2, 518], "35": [2, 518], "36": [2, 518], "37": [2, 518], "38": [2, 518], "39": [2, 518], "41": [2, 518], "42": [2, 518], "44": [2, 518], "45": [2, 518], "47": [2, 518], "48": [2, 518], "49": [2, 518], "50": [2, 518], "90": [2, 518], "96": [2, 518], "97": [2, 518], "118": [2, 518], "131": [2, 518], "195": [2, 518], "196": [2, 518], "197": [2, 518] }, { "4": [2, 519], "5": [2, 519], "31": [2, 519], "32": [2, 519], "34": [2, 519], "35": [2, 519], "36": [2, 519], "37": [2, 519], "38": [2, 519], "39": [2, 519], "41": [2, 519], "42": [2, 519], "44": [2, 519], "45": [2, 519], "47": [2, 519], "48": [2, 519], "49": [2, 519], "50": [2, 519], "90": [2, 519], "96": [2, 519], "97": [2, 519], "118": [2, 519], "131": [2, 519], "195": [2, 519], "196": [2, 519], "197": [2, 519] }, { "4": [2, 520], "5": [2, 520], "31": [2, 520], "32": [2, 520], "34": [2, 520], "35": [2, 520], "36": [2, 520], "37": [2, 520], "38": [2, 520], "39": [2, 520], "41": [2, 520], "42": [2, 520], "44": [2, 520], "45": [2, 520], "47": [2, 520], "48": [2, 520], "49": [2, 520], "50": [2, 520], "90": [2, 520], "96": [2, 520], "97": [2, 520], "118": [2, 520], "131": [2, 520], "195": [2, 520], "196": [2, 520], "197": [2, 520] }, { "4": [2, 521], "5": [2, 521], "31": [2, 521], "32": [2, 521], "34": [2, 521], "35": [2, 521], "36": [2, 521], "37": [2, 521], "38": [2, 521], "39": [2, 521], "41": [2, 521], "42": [2, 521], "44": [2, 521], "45": [2, 521], "47": [2, 521], "48": [2, 521], "49": [2, 521], "50": [2, 521], "90": [2, 521], "96": [2, 521], "97": [2, 521], "118": [2, 521], "131": [2, 521], "195": [2, 521], "196": [2, 521], "197": [2, 521] }, { "4": [2, 522], "5": [2, 522], "31": [2, 522], "32": [2, 522], "34": [2, 522], "35": [2, 522], "36": [2, 522], "37": [2, 522], "38": [2, 522], "39": [2, 522], "41": [2, 522], "42": [2, 522], "44": [2, 522], "45": [2, 522], "47": [2, 522], "48": [2, 522], "49": [2, 522], "50": [2, 522], "90": [2, 522], "96": [2, 522], "97": [2, 522], "118": [2, 522], "131": [2, 522], "195": [2, 522], "196": [2, 522], "197": [2, 522] }, { "4": [2, 523], "5": [2, 523], "31": [2, 523], "32": [2, 523], "34": [2, 523], "35": [2, 523], "36": [2, 523], "37": [2, 523], "38": [2, 523], "39": [2, 523], "41": [2, 523], "42": [2, 523], "44": [2, 523], "45": [2, 523], "47": [2, 523], "48": [2, 523], "49": [2, 523], "50": [2, 523], "90": [2, 523], "96": [2, 523], "97": [2, 523], "118": [2, 523], "131": [2, 523], "195": [2, 523], "196": [2, 523], "197": [2, 523] }, { "4": [2, 524], "5": [2, 524], "31": [2, 524], "32": [2, 524], "34": [2, 524], "35": [2, 524], "36": [2, 524], "37": [2, 524], "38": [2, 524], "39": [2, 524], "41": [2, 524], "42": [2, 524], "44": [2, 524], "45": [2, 524], "47": [2, 524], "48": [2, 524], "49": [2, 524], "50": [2, 524], "90": [2, 524], "96": [2, 524], "97": [2, 524], "118": [2, 524], "131": [2, 524], "195": [2, 524], "196": [2, 524], "197": [2, 524] }, { "4": [2, 525], "5": [2, 525], "31": [2, 525], "32": [2, 525], "34": [2, 525], "35": [2, 525], "36": [2, 525], "37": [2, 525], "38": [2, 525], "39": [2, 525], "41": [2, 525], "42": [2, 525], "44": [2, 525], "45": [2, 525], "47": [2, 525], "48": [2, 525], "49": [2, 525], "50": [2, 525], "90": [2, 525], "96": [2, 525], "97": [2, 525], "118": [2, 525], "131": [2, 525], "195": [2, 525], "196": [2, 525], "197": [2, 525] }, { "4": [2, 526], "5": [2, 526], "31": [2, 526], "32": [2, 526], "34": [2, 526], "35": [2, 526], "36": [2, 526], "37": [2, 526], "38": [2, 526], "39": [2, 526], "41": [2, 526], "42": [2, 526], "44": [2, 526], "45": [2, 526], "47": [2, 526], "48": [2, 526], "49": [2, 526], "50": [2, 526], "90": [2, 526], "96": [2, 526], "97": [2, 526], "118": [2, 526], "131": [2, 526], "195": [2, 526], "196": [2, 526], "197": [2, 526] }, { "4": [2, 527], "5": [2, 527], "31": [2, 527], "32": [2, 527], "34": [2, 527], "35": [2, 527], "36": [2, 527], "37": [2, 527], "38": [2, 527], "39": [2, 527], "41": [2, 527], "42": [2, 527], "44": [2, 527], "45": [2, 527], "47": [2, 527], "48": [2, 527], "49": [2, 527], "50": [2, 527], "90": [2, 527], "96": [2, 527], "97": [2, 527], "118": [2, 527], "131": [2, 527], "195": [2, 527], "196": [2, 527], "197": [2, 527] }, { "4": [2, 528], "5": [2, 528], "31": [2, 528], "32": [2, 528], "34": [2, 528], "35": [2, 528], "36": [2, 528], "37": [2, 528], "38": [2, 528], "39": [2, 528], "41": [2, 528], "42": [2, 528], "44": [2, 528], "45": [2, 528], "47": [2, 528], "48": [2, 528], "49": [2, 528], "50": [2, 528], "90": [2, 528], "96": [2, 528], "97": [2, 528], "118": [2, 528], "131": [2, 528], "195": [2, 528], "196": [2, 528], "197": [2, 528] }, { "4": [2, 529], "5": [2, 529], "31": [2, 529], "32": [2, 529], "34": [2, 529], "35": [2, 529], "36": [2, 529], "37": [2, 529], "38": [2, 529], "39": [2, 529], "41": [2, 529], "42": [2, 529], "44": [2, 529], "45": [2, 529], "47": [2, 529], "48": [2, 529], "49": [2, 529], "50": [2, 529], "90": [2, 529], "96": [2, 529], "97": [2, 529], "118": [2, 529], "131": [2, 529], "195": [2, 529], "196": [2, 529], "197": [2, 529] }, { "4": [2, 530], "5": [2, 530], "31": [2, 530], "32": [2, 530], "34": [2, 530], "35": [2, 530], "36": [2, 530], "37": [2, 530], "38": [2, 530], "39": [2, 530], "41": [2, 530], "42": [2, 530], "44": [2, 530], "45": [2, 530], "47": [2, 530], "48": [2, 530], "49": [2, 530], "50": [2, 530], "90": [2, 530], "96": [2, 530], "97": [2, 530], "118": [2, 530], "131": [2, 530], "195": [2, 530], "196": [2, 530], "197": [2, 530] }, { "4": [2, 531], "5": [2, 531], "31": [2, 531], "32": [2, 531], "34": [2, 531], "35": [2, 531], "36": [2, 531], "37": [2, 531], "38": [2, 531], "39": [2, 531], "41": [2, 531], "42": [2, 531], "44": [2, 531], "45": [2, 531], "47": [2, 531], "48": [2, 531], "49": [2, 531], "50": [2, 531], "90": [2, 531], "96": [2, 531], "97": [2, 531], "118": [2, 531], "131": [2, 531], "195": [2, 531], "196": [2, 531], "197": [2, 531] }, { "4": [2, 532], "5": [2, 532], "31": [2, 532], "32": [2, 532], "34": [2, 532], "35": [2, 532], "36": [2, 532], "37": [2, 532], "38": [2, 532], "39": [2, 532], "41": [2, 532], "42": [2, 532], "44": [2, 532], "45": [2, 532], "47": [2, 532], "48": [2, 532], "49": [2, 532], "50": [2, 532], "90": [2, 532], "96": [2, 532], "97": [2, 532], "118": [2, 532], "131": [2, 532], "195": [2, 532], "196": [2, 532], "197": [2, 532] }, { "4": [2, 533], "5": [2, 533], "31": [2, 533], "32": [2, 533], "34": [2, 533], "35": [2, 533], "36": [2, 533], "37": [2, 533], "38": [2, 533], "39": [2, 533], "41": [2, 533], "42": [2, 533], "44": [2, 533], "45": [2, 533], "47": [2, 533], "48": [2, 533], "49": [2, 533], "50": [2, 533], "90": [2, 533], "96": [2, 533], "97": [2, 533], "118": [2, 533], "131": [2, 533], "195": [2, 533], "196": [2, 533], "197": [2, 533] }, { "4": [2, 534], "5": [2, 534], "31": [2, 534], "32": [2, 534], "34": [2, 534], "35": [2, 534], "36": [2, 534], "37": [2, 534], "38": [2, 534], "39": [2, 534], "41": [2, 534], "42": [2, 534], "44": [2, 534], "45": [2, 534], "47": [2, 534], "48": [2, 534], "49": [2, 534], "50": [2, 534], "90": [2, 534], "96": [2, 534], "97": [2, 534], "118": [2, 534], "131": [2, 534], "195": [2, 534], "196": [2, 534], "197": [2, 534] }, { "4": [2, 535], "5": [2, 535], "31": [2, 535], "32": [2, 535], "34": [2, 535], "35": [2, 535], "36": [2, 535], "37": [2, 535], "38": [2, 535], "39": [2, 535], "41": [2, 535], "42": [2, 535], "44": [2, 535], "45": [2, 535], "47": [2, 535], "48": [2, 535], "49": [2, 535], "50": [2, 535], "90": [2, 535], "96": [2, 535], "97": [2, 535], "118": [2, 535], "131": [2, 535], "195": [2, 535], "196": [2, 535], "197": [2, 535] }, { "4": [2, 536], "5": [2, 536], "31": [2, 536], "32": [2, 536], "34": [2, 536], "35": [2, 536], "36": [2, 536], "37": [2, 536], "38": [2, 536], "39": [2, 536], "41": [2, 536], "42": [2, 536], "44": [2, 536], "45": [2, 536], "47": [2, 536], "48": [2, 536], "49": [2, 536], "50": [2, 536], "90": [2, 536], "96": [2, 536], "97": [2, 536], "118": [2, 536], "131": [2, 536], "195": [2, 536], "196": [2, 536], "197": [2, 536] }, { "4": [2, 537], "5": [2, 537], "31": [2, 537], "32": [2, 537], "34": [2, 537], "35": [2, 537], "36": [2, 537], "37": [2, 537], "38": [2, 537], "39": [2, 537], "41": [2, 537], "42": [2, 537], "44": [2, 537], "45": [2, 537], "47": [2, 537], "48": [2, 537], "49": [2, 537], "50": [2, 537], "90": [2, 537], "96": [2, 537], "97": [2, 537], "118": [2, 537], "131": [2, 537], "195": [2, 537], "196": [2, 537], "197": [2, 537] }, { "4": [2, 538], "5": [2, 538], "31": [2, 538], "32": [2, 538], "34": [2, 538], "35": [2, 538], "36": [2, 538], "37": [2, 538], "38": [2, 538], "39": [2, 538], "41": [2, 538], "42": [2, 538], "44": [2, 538], "45": [2, 538], "47": [2, 538], "48": [2, 538], "49": [2, 538], "50": [2, 538], "90": [2, 538], "96": [2, 538], "97": [2, 538], "118": [2, 538], "131": [2, 538], "195": [2, 538], "196": [2, 538], "197": [2, 538] }, { "4": [2, 539], "5": [2, 539], "31": [2, 539], "32": [2, 539], "34": [2, 539], "35": [2, 539], "36": [2, 539], "37": [2, 539], "38": [2, 539], "39": [2, 539], "41": [2, 539], "42": [2, 539], "44": [2, 539], "45": [2, 539], "47": [2, 539], "48": [2, 539], "49": [2, 539], "50": [2, 539], "90": [2, 539], "96": [2, 539], "97": [2, 539], "118": [2, 539], "131": [2, 539], "195": [2, 539], "196": [2, 539], "197": [2, 539] }, { "4": [2, 540], "5": [2, 540], "31": [2, 540], "32": [2, 540], "34": [2, 540], "35": [2, 540], "36": [2, 540], "37": [2, 540], "38": [2, 540], "39": [2, 540], "41": [2, 540], "42": [2, 540], "44": [2, 540], "45": [2, 540], "47": [2, 540], "48": [2, 540], "49": [2, 540], "50": [2, 540], "90": [2, 540], "96": [2, 540], "97": [2, 540], "118": [2, 540], "131": [2, 540], "195": [2, 540], "196": [2, 540], "197": [2, 540] }, { "4": [2, 541], "5": [2, 541], "31": [2, 541], "32": [2, 541], "34": [2, 541], "35": [2, 541], "36": [2, 541], "37": [2, 541], "38": [2, 541], "39": [2, 541], "41": [2, 541], "42": [2, 541], "44": [2, 541], "45": [2, 541], "47": [2, 541], "48": [2, 541], "49": [2, 541], "50": [2, 541], "90": [2, 541], "96": [2, 541], "97": [2, 541], "118": [2, 541], "131": [2, 541], "195": [2, 541], "196": [2, 541], "197": [2, 541] }, { "4": [2, 542], "5": [2, 542], "31": [2, 542], "32": [2, 542], "34": [2, 542], "35": [2, 542], "36": [2, 542], "37": [2, 542], "38": [2, 542], "39": [2, 542], "41": [2, 542], "42": [2, 542], "44": [2, 542], "45": [2, 542], "47": [2, 542], "48": [2, 542], "49": [2, 542], "50": [2, 542], "90": [2, 542], "96": [2, 542], "97": [2, 542], "118": [2, 542], "131": [2, 542], "195": [2, 542], "196": [2, 542], "197": [2, 542] }, { "4": [2, 543], "5": [2, 543], "31": [2, 543], "32": [2, 543], "34": [2, 543], "35": [2, 543], "36": [2, 543], "37": [2, 543], "38": [2, 543], "39": [2, 543], "41": [2, 543], "42": [2, 543], "44": [2, 543], "45": [2, 543], "47": [2, 543], "48": [2, 543], "49": [2, 543], "50": [2, 543], "90": [2, 543], "96": [2, 543], "97": [2, 543], "118": [2, 543], "131": [2, 543], "195": [2, 543], "196": [2, 543], "197": [2, 543] }, { "4": [2, 544], "5": [2, 544], "31": [2, 544], "32": [2, 544], "34": [2, 544], "35": [2, 544], "36": [2, 544], "37": [2, 544], "38": [2, 544], "39": [2, 544], "41": [2, 544], "42": [2, 544], "44": [2, 544], "45": [2, 544], "47": [2, 544], "48": [2, 544], "49": [2, 544], "50": [2, 544], "90": [2, 544], "96": [2, 544], "97": [2, 544], "118": [2, 544], "131": [2, 544], "195": [2, 544], "196": [2, 544], "197": [2, 544] }, { "4": [2, 545], "5": [2, 545], "31": [2, 545], "32": [2, 545], "34": [2, 545], "35": [2, 545], "36": [2, 545], "37": [2, 545], "38": [2, 545], "39": [2, 545], "41": [2, 545], "42": [2, 545], "44": [2, 545], "45": [2, 545], "47": [2, 545], "48": [2, 545], "49": [2, 545], "50": [2, 545], "90": [2, 545], "96": [2, 545], "97": [2, 545], "118": [2, 545], "131": [2, 545], "195": [2, 545], "196": [2, 545], "197": [2, 545] }, { "4": [2, 546], "5": [2, 546], "31": [2, 546], "32": [2, 546], "34": [2, 546], "35": [2, 546], "36": [2, 546], "37": [2, 546], "38": [2, 546], "39": [2, 546], "41": [2, 546], "42": [2, 546], "44": [2, 546], "45": [2, 546], "47": [2, 546], "48": [2, 546], "49": [2, 546], "50": [2, 546], "90": [2, 546], "96": [2, 546], "97": [2, 546], "118": [2, 546], "131": [2, 546], "195": [2, 546], "196": [2, 546], "197": [2, 546] }, { "4": [2, 547], "5": [2, 547], "31": [2, 547], "32": [2, 547], "34": [2, 547], "35": [2, 547], "36": [2, 547], "37": [2, 547], "38": [2, 547], "39": [2, 547], "41": [2, 547], "42": [2, 547], "44": [2, 547], "45": [2, 547], "47": [2, 547], "48": [2, 547], "49": [2, 547], "50": [2, 547], "90": [2, 547], "96": [2, 547], "97": [2, 547], "118": [2, 547], "131": [2, 547], "195": [2, 547], "196": [2, 547], "197": [2, 547] }, { "4": [2, 548], "5": [2, 548], "31": [2, 548], "32": [2, 548], "34": [2, 548], "35": [2, 548], "36": [2, 548], "37": [2, 548], "38": [2, 548], "39": [2, 548], "41": [2, 548], "42": [2, 548], "44": [2, 548], "45": [2, 548], "47": [2, 548], "48": [2, 548], "49": [2, 548], "50": [2, 548], "90": [2, 548], "96": [2, 548], "97": [2, 548], "118": [2, 548], "131": [2, 548], "195": [2, 548], "196": [2, 548], "197": [2, 548] }, { "4": [2, 549], "5": [2, 549], "31": [2, 549], "32": [2, 549], "34": [2, 549], "35": [2, 549], "36": [2, 549], "37": [2, 549], "38": [2, 549], "39": [2, 549], "41": [2, 549], "42": [2, 549], "44": [2, 549], "45": [2, 549], "47": [2, 549], "48": [2, 549], "49": [2, 549], "50": [2, 549], "90": [2, 549], "96": [2, 549], "97": [2, 549], "118": [2, 549], "131": [2, 549], "195": [2, 549], "196": [2, 549], "197": [2, 549] }, { "4": [2, 550], "5": [2, 550], "31": [2, 550], "32": [2, 550], "34": [2, 550], "35": [2, 550], "36": [2, 550], "37": [2, 550], "38": [2, 550], "39": [2, 550], "41": [2, 550], "42": [2, 550], "44": [2, 550], "45": [2, 550], "47": [2, 550], "48": [2, 550], "49": [2, 550], "50": [2, 550], "90": [2, 550], "96": [2, 550], "97": [2, 550], "118": [2, 550], "131": [2, 550], "195": [2, 550], "196": [2, 550], "197": [2, 550] }, { "4": [2, 551], "5": [2, 551], "31": [2, 551], "32": [2, 551], "34": [2, 551], "35": [2, 551], "36": [2, 551], "37": [2, 551], "38": [2, 551], "39": [2, 551], "41": [2, 551], "42": [2, 551], "44": [2, 551], "45": [2, 551], "47": [2, 551], "48": [2, 551], "49": [2, 551], "50": [2, 551], "90": [2, 551], "96": [2, 551], "97": [2, 551], "118": [2, 551], "131": [2, 551], "195": [2, 551], "196": [2, 551], "197": [2, 551] }, { "4": [2, 552], "5": [2, 552], "31": [2, 552], "32": [2, 552], "34": [2, 552], "35": [2, 552], "36": [2, 552], "37": [2, 552], "38": [2, 552], "39": [2, 552], "41": [2, 552], "42": [2, 552], "44": [2, 552], "45": [2, 552], "47": [2, 552], "48": [2, 552], "49": [2, 552], "50": [2, 552], "90": [2, 552], "96": [2, 552], "97": [2, 552], "118": [2, 552], "131": [2, 552], "195": [2, 552], "196": [2, 552], "197": [2, 552] }, { "4": [2, 553], "5": [2, 553], "31": [2, 553], "32": [2, 553], "34": [2, 553], "35": [2, 553], "36": [2, 553], "37": [2, 553], "38": [2, 553], "39": [2, 553], "41": [2, 553], "42": [2, 553], "44": [2, 553], "45": [2, 553], "47": [2, 553], "48": [2, 553], "49": [2, 553], "50": [2, 553], "90": [2, 553], "96": [2, 553], "97": [2, 553], "118": [2, 553], "131": [2, 553], "195": [2, 553], "196": [2, 553], "197": [2, 553] }, { "4": [2, 554], "5": [2, 554], "31": [2, 554], "32": [2, 554], "34": [2, 554], "35": [2, 554], "36": [2, 554], "37": [2, 554], "38": [2, 554], "39": [2, 554], "41": [2, 554], "42": [2, 554], "44": [2, 554], "45": [2, 554], "47": [2, 554], "48": [2, 554], "49": [2, 554], "50": [2, 554], "90": [2, 554], "96": [2, 554], "97": [2, 554], "118": [2, 554], "131": [2, 554], "195": [2, 554], "196": [2, 554], "197": [2, 554] }, { "4": [2, 555], "5": [2, 555], "31": [2, 555], "32": [2, 555], "34": [2, 555], "35": [2, 555], "36": [2, 555], "37": [2, 555], "38": [2, 555], "39": [2, 555], "41": [2, 555], "42": [2, 555], "44": [2, 555], "45": [2, 555], "47": [2, 555], "48": [2, 555], "49": [2, 555], "50": [2, 555], "90": [2, 555], "96": [2, 555], "97": [2, 555], "118": [2, 555], "131": [2, 555], "195": [2, 555], "196": [2, 555], "197": [2, 555] }, { "4": [2, 556], "5": [2, 556], "31": [2, 556], "32": [2, 556], "34": [2, 556], "35": [2, 556], "36": [2, 556], "37": [2, 556], "38": [2, 556], "39": [2, 556], "41": [2, 556], "42": [2, 556], "44": [2, 556], "45": [2, 556], "47": [2, 556], "48": [2, 556], "49": [2, 556], "50": [2, 556], "90": [2, 556], "96": [2, 556], "97": [2, 556], "118": [2, 556], "131": [2, 556], "195": [2, 556], "196": [2, 556], "197": [2, 556] }, { "4": [2, 557], "5": [2, 557], "31": [2, 557], "32": [2, 557], "34": [2, 557], "35": [2, 557], "36": [2, 557], "37": [2, 557], "38": [2, 557], "39": [2, 557], "41": [2, 557], "42": [2, 557], "44": [2, 557], "45": [2, 557], "47": [2, 557], "48": [2, 557], "49": [2, 557], "50": [2, 557], "90": [2, 557], "96": [2, 557], "97": [2, 557], "118": [2, 557], "131": [2, 557], "195": [2, 557], "196": [2, 557], "197": [2, 557] }, { "4": [2, 558], "5": [2, 558], "31": [2, 558], "32": [2, 558], "34": [2, 558], "35": [2, 558], "36": [2, 558], "37": [2, 558], "38": [2, 558], "39": [2, 558], "41": [2, 558], "42": [2, 558], "44": [2, 558], "45": [2, 558], "47": [2, 558], "48": [2, 558], "49": [2, 558], "50": [2, 558], "90": [2, 558], "96": [2, 558], "97": [2, 558], "118": [2, 558], "131": [2, 558], "195": [2, 558], "196": [2, 558], "197": [2, 558] }, { "4": [2, 559], "5": [2, 559], "31": [2, 559], "32": [2, 559], "34": [2, 559], "35": [2, 559], "36": [2, 559], "37": [2, 559], "38": [2, 559], "39": [2, 559], "41": [2, 559], "42": [2, 559], "44": [2, 559], "45": [2, 559], "47": [2, 559], "48": [2, 559], "49": [2, 559], "50": [2, 559], "90": [2, 559], "96": [2, 559], "97": [2, 559], "118": [2, 559], "131": [2, 559], "195": [2, 559], "196": [2, 559], "197": [2, 559] }, { "4": [2, 560], "5": [2, 560], "31": [2, 560], "32": [2, 560], "34": [2, 560], "35": [2, 560], "36": [2, 560], "37": [2, 560], "38": [2, 560], "39": [2, 560], "41": [2, 560], "42": [2, 560], "44": [2, 560], "45": [2, 560], "47": [2, 560], "48": [2, 560], "49": [2, 560], "50": [2, 560], "90": [2, 560], "96": [2, 560], "97": [2, 560], "118": [2, 560], "131": [2, 560], "195": [2, 560], "196": [2, 560], "197": [2, 560] }, { "4": [2, 561], "5": [2, 561], "31": [2, 561], "32": [2, 561], "34": [2, 561], "35": [2, 561], "36": [2, 561], "37": [2, 561], "38": [2, 561], "39": [2, 561], "41": [2, 561], "42": [2, 561], "44": [2, 561], "45": [2, 561], "47": [2, 561], "48": [2, 561], "49": [2, 561], "50": [2, 561], "90": [2, 561], "96": [2, 561], "97": [2, 561], "118": [2, 561], "131": [2, 561], "195": [2, 561], "196": [2, 561], "197": [2, 561] }, { "4": [2, 562], "5": [2, 562], "31": [2, 562], "32": [2, 562], "34": [2, 562], "35": [2, 562], "36": [2, 562], "37": [2, 562], "38": [2, 562], "39": [2, 562], "41": [2, 562], "42": [2, 562], "44": [2, 562], "45": [2, 562], "47": [2, 562], "48": [2, 562], "49": [2, 562], "50": [2, 562], "90": [2, 562], "96": [2, 562], "97": [2, 562], "118": [2, 562], "131": [2, 562], "195": [2, 562], "196": [2, 562], "197": [2, 562] }, { "4": [2, 563], "5": [2, 563], "31": [2, 563], "32": [2, 563], "34": [2, 563], "35": [2, 563], "36": [2, 563], "37": [2, 563], "38": [2, 563], "39": [2, 563], "41": [2, 563], "42": [2, 563], "44": [2, 563], "45": [2, 563], "47": [2, 563], "48": [2, 563], "49": [2, 563], "50": [2, 563], "90": [2, 563], "96": [2, 563], "97": [2, 563], "118": [2, 563], "131": [2, 563], "195": [2, 563], "196": [2, 563], "197": [2, 563] }, { "4": [2, 564], "5": [2, 564], "31": [2, 564], "32": [2, 564], "34": [2, 564], "35": [2, 564], "36": [2, 564], "37": [2, 564], "38": [2, 564], "39": [2, 564], "41": [2, 564], "42": [2, 564], "44": [2, 564], "45": [2, 564], "47": [2, 564], "48": [2, 564], "49": [2, 564], "50": [2, 564], "90": [2, 564], "96": [2, 564], "97": [2, 564], "118": [2, 564], "131": [2, 564], "195": [2, 564], "196": [2, 564], "197": [2, 564] }, { "4": [2, 565], "5": [2, 565], "31": [2, 565], "32": [2, 565], "34": [2, 565], "35": [2, 565], "36": [2, 565], "37": [2, 565], "38": [2, 565], "39": [2, 565], "41": [2, 565], "42": [2, 565], "44": [2, 565], "45": [2, 565], "47": [2, 565], "48": [2, 565], "49": [2, 565], "50": [2, 565], "90": [2, 565], "96": [2, 565], "97": [2, 565], "118": [2, 565], "131": [2, 565], "195": [2, 565], "196": [2, 565], "197": [2, 565] }, { "4": [2, 471], "5": [2, 471], "31": [2, 471], "32": [2, 471], "33": [2, 471], "34": [2, 471], "35": [2, 471], "36": [2, 471], "37": [2, 471], "38": [2, 471], "39": [2, 471], "41": [2, 471], "42": [2, 471], "44": [2, 471], "45": [2, 471], "47": [2, 471], "48": [2, 471], "49": [2, 471], "50": [2, 471], "90": [2, 471], "96": [2, 471], "97": [2, 471], "118": [2, 471], "129": [2, 471], "131": [2, 471], "135": [2, 471], "195": [2, 471], "196": [2, 471], "197": [2, 471] }, { "4": [2, 472], "5": [2, 472], "31": [2, 472], "32": [2, 472], "33": [2, 472], "34": [2, 472], "35": [2, 472], "36": [2, 472], "37": [2, 472], "38": [2, 472], "39": [2, 472], "41": [2, 472], "42": [2, 472], "44": [2, 472], "45": [2, 472], "47": [2, 472], "48": [2, 472], "49": [2, 472], "50": [2, 472], "90": [2, 472], "96": [2, 472], "97": [2, 472], "118": [2, 472], "129": [2, 472], "131": [2, 472], "135": [2, 472], "195": [2, 472], "196": [2, 472], "197": [2, 472] }, { "4": [2, 473], "5": [2, 473], "31": [2, 473], "32": [2, 473], "33": [2, 473], "34": [2, 473], "35": [2, 473], "36": [2, 473], "37": [2, 473], "38": [2, 473], "39": [2, 473], "41": [2, 473], "42": [2, 473], "44": [2, 473], "45": [2, 473], "47": [2, 473], "48": [2, 473], "49": [2, 473], "50": [2, 473], "90": [2, 473], "96": [2, 473], "97": [2, 473], "118": [2, 473], "129": [2, 473], "131": [2, 473], "135": [2, 473], "195": [2, 473], "196": [2, 473], "197": [2, 473] }, { "4": [2, 474], "5": [2, 474], "31": [2, 474], "32": [2, 474], "33": [2, 474], "34": [2, 474], "35": [2, 474], "36": [2, 474], "37": [2, 474], "38": [2, 474], "39": [2, 474], "41": [2, 474], "42": [2, 474], "44": [2, 474], "45": [2, 474], "47": [2, 474], "48": [2, 474], "49": [2, 474], "50": [2, 474], "90": [2, 474], "96": [2, 474], "97": [2, 474], "118": [2, 474], "129": [2, 474], "131": [2, 474], "135": [2, 474], "195": [2, 474], "196": [2, 474], "197": [2, 474] }, { "4": [2, 475], "5": [2, 475], "31": [2, 475], "32": [2, 475], "33": [2, 475], "34": [2, 475], "35": [2, 475], "36": [2, 475], "37": [2, 475], "38": [2, 475], "39": [2, 475], "41": [2, 475], "42": [2, 475], "44": [2, 475], "45": [2, 475], "47": [2, 475], "48": [2, 475], "49": [2, 475], "50": [2, 475], "90": [2, 475], "96": [2, 475], "97": [2, 475], "118": [2, 475], "129": [2, 475], "131": [2, 475], "135": [2, 475], "195": [2, 475], "196": [2, 475], "197": [2, 475] }, { "4": [2, 476], "5": [2, 476], "31": [2, 476], "32": [2, 476], "33": [2, 476], "34": [2, 476], "35": [2, 476], "36": [2, 476], "37": [2, 476], "38": [2, 476], "39": [2, 476], "41": [2, 476], "42": [2, 476], "44": [2, 476], "45": [2, 476], "47": [2, 476], "48": [2, 476], "49": [2, 476], "50": [2, 476], "90": [2, 476], "96": [2, 476], "97": [2, 476], "118": [2, 476], "129": [2, 476], "131": [2, 476], "135": [2, 476], "195": [2, 476], "196": [2, 476], "197": [2, 476] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "33": [1, 521] }, { "4": [2, 393], "5": [2, 393], "31": [2, 393], "32": [2, 393], "33": [1, 522], "34": [2, 393], "35": [2, 393], "36": [2, 393], "37": [2, 393], "38": [2, 393], "39": [2, 393], "41": [2, 393], "42": [2, 393], "44": [2, 393], "45": [2, 393], "47": [2, 393], "48": [2, 393], "49": [2, 393], "50": [2, 393], "90": [2, 393], "96": [2, 393], "97": [2, 393], "118": [2, 393], "131": [2, 393], "195": [2, 393], "196": [2, 393], "197": [2, 393] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "69": [1, 409], "140": [1, 411], "151": [1, 410] }, { "69": [1, 409], "140": [1, 411], "151": [1, 410] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "4": [2, 358], "5": [2, 358], "31": [2, 358], "32": [2, 358], "34": [2, 358], "35": [2, 358], "36": [2, 358], "37": [2, 358], "38": [2, 358], "39": [2, 358], "41": [2, 358], "42": [2, 358], "44": [2, 358], "45": [2, 358], "47": [2, 358], "48": [2, 358], "49": [2, 358], "50": [2, 358], "90": [2, 358], "96": [2, 358], "97": [2, 358], "118": [2, 358], "131": [2, 358], "195": [2, 358], "196": [2, 358], "197": [2, 358] }, { "4": [2, 355], "5": [2, 355], "31": [2, 355], "32": [2, 355], "34": [2, 355], "35": [2, 355], "36": [2, 355], "37": [2, 355], "38": [2, 355], "39": [2, 355], "41": [2, 355], "42": [2, 355], "44": [2, 355], "45": [2, 355], "47": [2, 355], "48": [2, 355], "49": [2, 355], "50": [2, 355], "90": [2, 355], "96": [2, 355], "97": [2, 355], "118": [2, 355], "131": [2, 355], "195": [2, 355], "196": [2, 355], "197": [2, 355] }, { "4": [2, 481], "5": [2, 481], "31": [2, 481], "32": [2, 481], "34": [2, 481], "35": [2, 481], "36": [2, 481], "37": [2, 481], "38": [2, 481], "39": [2, 481], "41": [2, 481], "42": [2, 481], "44": [2, 481], "45": [2, 481], "47": [2, 481], "48": [2, 481], "49": [2, 481], "50": [2, 481], "90": [2, 481], "96": [2, 481], "97": [2, 481], "118": [2, 481], "131": [2, 481], "195": [2, 481], "196": [2, 481], "197": [2, 481] }, { "4": [2, 482], "5": [2, 482], "31": [2, 482], "32": [2, 482], "34": [2, 482], "35": [2, 482], "36": [2, 482], "37": [2, 482], "38": [2, 482], "39": [2, 482], "41": [2, 482], "42": [2, 482], "44": [2, 482], "45": [2, 482], "47": [2, 482], "48": [2, 482], "49": [2, 482], "50": [2, 482], "90": [2, 482], "96": [2, 482], "97": [2, 482], "118": [2, 482], "131": [2, 482], "195": [2, 482], "196": [2, 482], "197": [2, 482] }, { "4": [2, 483], "5": [2, 483], "31": [2, 483], "32": [2, 483], "34": [2, 483], "35": [2, 483], "36": [2, 483], "37": [2, 483], "38": [2, 483], "39": [2, 483], "41": [2, 483], "42": [2, 483], "44": [2, 483], "45": [2, 483], "47": [2, 483], "48": [2, 483], "49": [2, 483], "50": [2, 483], "90": [2, 483], "96": [2, 483], "97": [2, 483], "118": [2, 483], "131": [2, 483], "195": [2, 483], "196": [2, 483], "197": [2, 483] }, { "4": [2, 352], "5": [2, 352], "31": [2, 352], "32": [2, 352], "34": [2, 352], "35": [2, 352], "36": [2, 352], "37": [2, 352], "38": [2, 352], "39": [2, 352], "41": [2, 352], "42": [2, 352], "44": [2, 352], "45": [2, 352], "47": [2, 352], "48": [2, 352], "49": [2, 352], "50": [2, 352], "60": [2, 352], "69": [2, 352], "74": [2, 352], "90": [2, 352], "96": [2, 352], "97": [2, 352], "118": [2, 352], "131": [2, 352], "140": [2, 352], "151": [2, 352], "195": [2, 352], "196": [2, 352], "197": [2, 352] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "4": [2, 206], "5": [2, 206], "32": [2, 206], "34": [2, 206], "38": [2, 206], "42": [2, 206], "44": [2, 206], "48": [2, 206], "90": [2, 206], "131": [2, 206], "195": [2, 206], "196": [2, 206] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "33": [1, 294] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [2, 176], "5": [2, 176], "34": [2, 176], "38": [2, 176], "44": [2, 176], "48": [2, 176], "96": [2, 176], "131": [2, 176], "196": [2, 176] }, { "34": [2, 344], "38": [2, 344], "42": [1, 530], "90": [2, 344] }, { "38": [1, 509], "90": [1, 531] }, { "4": [2, 180], "5": [2, 180], "34": [2, 180], "38": [2, 180], "44": [2, 180], "48": [2, 180], "89": [2, 180], "96": [2, 180], "131": [2, 180], "196": [2, 180] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "160": [1, 532] }, { "97": [1, 533], "118": [1, 534] }, { "40": [1, 293], "160": [1, 535] }, { "40": [1, 293], "160": [1, 536] }, { "34": [1, 537] }, { "4": [2, 463], "5": [2, 463], "31": [2, 463], "32": [2, 463], "33": [1, 220], "34": [2, 463], "35": [2, 463], "36": [2, 463], "37": [2, 463], "38": [2, 463], "39": [2, 463], "40": [2, 463], "41": [2, 463], "42": [2, 463], "44": [2, 463], "45": [2, 463], "46": [2, 463], "47": [2, 463], "48": [2, 463], "49": [2, 463], "50": [2, 463], "85": [2, 463], "90": [2, 463], "96": [2, 463], "97": [2, 463], "118": [2, 463], "129": [2, 463], "131": [2, 463], "160": [2, 463], "195": [2, 463], "196": [2, 463], "197": [2, 463] }, { "4": [2, 465], "5": [2, 465], "31": [2, 465], "32": [2, 465], "33": [1, 221], "34": [2, 465], "35": [2, 465], "36": [2, 465], "37": [2, 465], "38": [2, 465], "39": [2, 465], "40": [2, 465], "41": [2, 465], "42": [2, 465], "44": [2, 465], "45": [2, 465], "46": [2, 465], "47": [2, 465], "48": [2, 465], "49": [2, 465], "50": [2, 465], "85": [2, 465], "90": [2, 465], "96": [2, 465], "97": [2, 465], "118": [2, 465], "129": [2, 465], "131": [2, 465], "160": [2, 465], "195": [2, 465], "196": [2, 465], "197": [2, 465] }, { "4": [2, 460], "5": [2, 460], "31": [2, 460], "32": [2, 460], "33": [2, 477], "34": [2, 460], "35": [2, 460], "36": [2, 460], "37": [2, 460], "38": [2, 460], "39": [2, 460], "40": [2, 460], "41": [2, 460], "42": [2, 460], "44": [2, 460], "45": [2, 460], "46": [2, 460], "47": [2, 460], "48": [2, 460], "49": [2, 460], "50": [2, 460], "85": [2, 460], "90": [2, 460], "96": [2, 460], "97": [2, 460], "118": [2, 460], "129": [2, 460], "131": [2, 460], "160": [2, 460], "195": [2, 460], "196": [2, 460], "197": [2, 460] }, { "4": [2, 461], "5": [2, 461], "31": [2, 461], "32": [2, 461], "33": [2, 478], "34": [2, 461], "35": [2, 461], "36": [2, 461], "37": [2, 461], "38": [2, 461], "39": [2, 461], "40": [2, 461], "41": [2, 461], "42": [2, 461], "44": [2, 461], "45": [2, 461], "46": [2, 461], "47": [2, 461], "48": [2, 461], "49": [2, 461], "50": [2, 461], "85": [2, 461], "90": [2, 461], "96": [2, 461], "97": [2, 461], "118": [2, 461], "129": [2, 461], "131": [2, 461], "160": [2, 461], "195": [2, 461], "196": [2, 461], "197": [2, 461] }, { "34": [1, 538] }, { "38": [1, 540], "160": [1, 539] }, { "38": [2, 479], "40": [1, 293], "160": [2, 479] }, { "4": [2, 460], "5": [2, 460], "31": [2, 460], "32": [2, 460], "33": [2, 477], "34": [2, 460], "35": [2, 460], "36": [2, 460], "37": [2, 460], "38": [2, 460], "39": [2, 460], "40": [2, 460], "41": [2, 460], "42": [2, 460], "44": [2, 460], "45": [2, 460], "46": [2, 460], "47": [2, 460], "48": [2, 460], "49": [2, 460], "50": [2, 460], "85": [2, 460], "90": [2, 460], "96": [2, 460], "97": [2, 460], "118": [2, 460], "129": [2, 460], "131": [2, 460], "160": [2, 460], "195": [2, 460], "196": [2, 460], "197": [2, 460] }, { "4": [2, 461], "5": [2, 461], "31": [2, 461], "32": [2, 461], "33": [2, 478], "34": [2, 461], "35": [2, 461], "36": [2, 461], "37": [2, 461], "38": [2, 461], "39": [2, 461], "40": [2, 461], "41": [2, 461], "42": [2, 461], "44": [2, 461], "45": [2, 461], "46": [2, 461], "47": [2, 461], "48": [2, 461], "49": [2, 461], "50": [2, 461], "85": [2, 461], "90": [2, 461], "96": [2, 461], "97": [2, 461], "118": [2, 461], "129": [2, 461], "131": [2, 461], "160": [2, 461], "195": [2, 461], "196": [2, 461], "197": [2, 461] }, { "4": [2, 463], "5": [2, 463], "31": [2, 463], "32": [2, 463], "33": [1, 220], "34": [2, 463], "35": [2, 463], "36": [2, 463], "37": [2, 463], "38": [2, 463], "39": [2, 463], "40": [2, 463], "41": [2, 463], "42": [2, 463], "44": [2, 463], "45": [2, 463], "46": [2, 463], "47": [2, 463], "48": [2, 463], "49": [2, 463], "50": [2, 463], "85": [2, 463], "90": [2, 463], "96": [2, 463], "97": [2, 463], "118": [2, 463], "129": [2, 463], "131": [2, 463], "160": [2, 463], "195": [2, 463], "196": [2, 463], "197": [2, 463] }, { "4": [2, 465], "5": [2, 465], "31": [2, 465], "32": [2, 465], "33": [1, 221], "34": [2, 465], "35": [2, 465], "36": [2, 465], "37": [2, 465], "38": [2, 465], "39": [2, 465], "40": [2, 465], "41": [2, 465], "42": [2, 465], "44": [2, 465], "45": [2, 465], "46": [2, 465], "47": [2, 465], "48": [2, 465], "49": [2, 465], "50": [2, 465], "85": [2, 465], "90": [2, 465], "96": [2, 465], "97": [2, 465], "118": [2, 465], "129": [2, 465], "131": [2, 465], "160": [2, 465], "195": [2, 465], "196": [2, 465], "197": [2, 465] }, { "4": [2, 192], "5": [2, 192], "34": [2, 192], "38": [2, 192], "40": [2, 192], "44": [2, 192], "48": [2, 192], "89": [2, 192], "96": [2, 192], "131": [2, 192], "196": [2, 192] }, { "34": [1, 541], "38": [1, 542] }, { "34": [2, 194], "38": [2, 194] }, { "46": [1, 543] }, { "46": [1, 544] }, { "69": [1, 394], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 399], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "33": [1, 224] }, { "4": [1, 25], "5": [1, 24], "194": [1, 547] }, { "34": [1, 549], "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "1": [2, 123], "33": [2, 123], "55": [2, 123], "69": [2, 123], "89": [2, 123], "91": [2, 123], "92": [2, 123], "97": [2, 123], "98": [2, 123], "100": [2, 123], "101": [2, 123], "106": [2, 123], "108": [2, 123], "113": [2, 123], "114": [2, 123], "115": [2, 123], "118": [2, 123], "122": [2, 123], "123": [2, 123], "125": [2, 123], "127": [2, 123], "134": [2, 123], "135": [2, 123], "136": [2, 123], "145": [2, 123], "147": [2, 123], "148": [2, 123], "149": [2, 123], "150": [2, 123], "151": [2, 123], "153": [2, 123], "154": [2, 123], "155": [2, 123], "158": [2, 123], "167": [2, 123], "170": [2, 123], "172": [2, 123], "175": [2, 123], "178": [2, 123], "179": [2, 123], "181": [2, 123] }, { "194": [1, 452] }, { "80": [1, 189] }, { "92": [1, 560], "167": [1, 561], "170": [1, 559], "178": [1, 558], "196": [1, 555] }, { "1": [2, 77], "33": [2, 77], "55": [2, 77], "69": [2, 77], "89": [2, 77], "91": [2, 77], "92": [2, 77], "97": [2, 77], "98": [2, 77], "100": [2, 77], "101": [2, 77], "106": [2, 77], "108": [2, 77], "113": [2, 77], "114": [2, 77], "115": [2, 77], "118": [2, 77], "122": [2, 77], "123": [2, 77], "125": [2, 77], "127": [2, 77], "134": [2, 77], "135": [2, 77], "136": [2, 77], "145": [2, 77], "147": [2, 77], "148": [2, 77], "149": [2, 77], "150": [2, 77], "151": [2, 77], "153": [2, 77], "154": [2, 77], "155": [2, 77], "158": [2, 77], "167": [2, 77], "170": [2, 77], "172": [2, 77], "175": [2, 77], "178": [2, 77], "179": [2, 77], "181": [2, 77] }, { "55": [1, 563] }, { "194": [1, 456] }, { "92": [1, 574], "125": [1, 569], "139": [1, 568], "149": [1, 573], "154": [1, 572], "170": [1, 571], "196": [1, 565] }, { "1": [2, 20], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "1": [2, 28], "33": [2, 28], "55": [2, 28], "69": [2, 28], "89": [2, 28], "92": [2, 28], "97": [2, 28], "98": [2, 28], "100": [2, 28], "101": [2, 28], "108": [2, 28], "113": [2, 28], "115": [2, 28], "118": [2, 28], "122": [2, 28], "123": [2, 28], "125": [2, 28], "127": [2, 28], "134": [2, 28], "135": [2, 28], "145": [2, 28], "147": [2, 28], "149": [2, 28], "150": [2, 28], "151": [2, 28], "153": [2, 28], "154": [2, 28], "155": [2, 28], "158": [2, 28], "167": [2, 28], "170": [2, 28], "172": [2, 28], "175": [2, 28], "178": [2, 28], "179": [2, 28], "181": [2, 28] }, { "4": [1, 578], "5": [1, 581], "44": [1, 576] }, { "34": [1, 582], "38": [1, 509] }, { "1": [2, 26], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [2, 19], "5": [2, 19], "33": [2, 19], "38": [2, 19], "44": [2, 19], "48": [2, 19], "131": [2, 19] }, { "35": [1, 587], "55": [1, 588], "158": [1, 585] }, { "4": [2, 272], "5": [2, 272], "31": [2, 272], "32": [2, 272], "34": [2, 272], "35": [2, 272], "36": [2, 272], "37": [2, 272], "38": [2, 272], "39": [2, 272], "40": [1, 293], "41": [2, 272], "42": [2, 272], "44": [2, 272], "45": [2, 272], "47": [2, 272], "48": [2, 272], "49": [2, 272], "50": [2, 272], "90": [2, 272], "96": [2, 272], "97": [2, 272], "118": [2, 272], "131": [2, 272], "160": [1, 502], "195": [2, 272], "196": [2, 272], "197": [2, 272] }, { "44": [1, 590] }, { "44": [1, 591] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "3": [1, 490], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "34": [1, 600] }, { "90": [1, 601] }, { "90": [1, 602] }, { "44": [1, 603] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [2, 201], "5": [2, 201], "32": [1, 413], "34": [2, 201], "38": [2, 201], "42": [2, 201], "44": [2, 201], "48": [2, 201], "90": [2, 201], "131": [2, 201], "195": [2, 201], "196": [2, 201] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "38": [1, 273], "90": [2, 166] }, { "34": [2, 215], "37": [1, 276], "38": [2, 215], "39": [1, 277], "90": [2, 215], "96": [2, 215], "131": [1, 275] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 484], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "38": [2, 169], "90": [2, 169] }, { "38": [2, 171], "90": [2, 171] }, { "34": [2, 216], "38": [2, 216], "90": [2, 216], "96": [2, 216] }, { "34": [2, 209], "38": [2, 209], "42": [2, 209], "90": [2, 209], "96": [2, 209], "97": [2, 209], "118": [2, 209], "196": [2, 209] }, { "34": [2, 210], "38": [2, 210], "42": [2, 210], "90": [2, 210], "96": [2, 210], "97": [2, 210], "118": [2, 210], "196": [2, 210] }, { "33": [1, 415], "95": [1, 214], "110": [1, 107], "111": [1, 108], "124": [1, 104], "130": [1, 213], "141": [1, 110], "142": [1, 111], "165": [1, 105], "176": [1, 103], "180": [1, 106], "186": [1, 212], "198": [1, 109], "199": [1, 112] }, { "37": [1, 276], "39": [1, 277], "45": [1, 281], "47": [1, 279], "49": [1, 282], "50": [1, 280], "197": [1, 283] }, { "31": [1, 286], "34": [2, 218], "35": [1, 284], "37": [2, 218], "38": [2, 218], "39": [2, 218], "41": [1, 285], "45": [2, 218], "47": [2, 218], "49": [2, 218], "50": [2, 218], "90": [2, 218], "96": [2, 218], "131": [2, 218], "197": [2, 218] }, { "31": [1, 286], "34": [2, 219], "35": [1, 284], "37": [2, 219], "38": [2, 219], "39": [2, 219], "41": [1, 285], "45": [2, 219], "47": [2, 219], "49": [2, 219], "50": [2, 219], "90": [2, 219], "96": [2, 219], "131": [2, 219], "197": [2, 219] }, { "4": [2, 207], "5": [2, 207], "32": [2, 207], "34": [2, 207], "37": [1, 469], "38": [2, 207], "39": [1, 470], "42": [2, 207], "44": [2, 207], "48": [2, 207], "90": [2, 207], "131": [2, 207], "195": [2, 207], "196": [2, 207] }, { "33": [1, 608] }, { "31": [2, 221], "34": [2, 221], "35": [2, 221], "36": [1, 287], "37": [2, 221], "38": [2, 221], "39": [2, 221], "41": [2, 221], "45": [2, 221], "47": [2, 221], "49": [2, 221], "50": [2, 221], "90": [2, 221], "96": [2, 221], "131": [2, 221], "197": [2, 221] }, { "31": [2, 222], "34": [2, 222], "35": [2, 222], "36": [1, 287], "37": [2, 222], "38": [2, 222], "39": [2, 222], "41": [2, 222], "45": [2, 222], "47": [2, 222], "49": [2, 222], "50": [2, 222], "90": [2, 222], "96": [2, 222], "131": [2, 222], "197": [2, 222] }, { "31": [2, 223], "34": [2, 223], "35": [2, 223], "36": [1, 287], "37": [2, 223], "38": [2, 223], "39": [2, 223], "41": [2, 223], "45": [2, 223], "47": [2, 223], "49": [2, 223], "50": [2, 223], "90": [2, 223], "96": [2, 223], "131": [2, 223], "197": [2, 223] }, { "31": [2, 225], "34": [2, 225], "35": [2, 225], "36": [2, 225], "37": [2, 225], "38": [2, 225], "39": [2, 225], "41": [2, 225], "45": [2, 225], "47": [2, 225], "49": [2, 225], "50": [2, 225], "90": [2, 225], "96": [2, 225], "131": [2, 225], "197": [2, 225] }, { "31": [2, 226], "34": [2, 226], "35": [2, 226], "36": [2, 226], "37": [2, 226], "38": [2, 226], "39": [2, 226], "41": [2, 226], "45": [2, 226], "47": [2, 226], "49": [2, 226], "50": [2, 226], "90": [2, 226], "96": [2, 226], "131": [2, 226], "197": [2, 226] }, { "4": [2, 258], "5": [2, 258], "32": [2, 258], "34": [2, 258], "38": [2, 258], "42": [2, 258], "44": [2, 258], "48": [2, 258], "90": [2, 258], "96": [2, 258], "97": [2, 258], "118": [2, 258], "131": [2, 258], "195": [2, 258], "196": [2, 258] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 484], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "4": [2, 274], "5": [2, 274], "31": [2, 274], "32": [2, 274], "34": [2, 274], "35": [2, 274], "36": [2, 274], "37": [2, 274], "38": [2, 274], "39": [2, 274], "40": [2, 274], "41": [2, 274], "42": [2, 274], "44": [2, 274], "45": [2, 274], "46": [2, 274], "47": [2, 274], "48": [2, 274], "49": [2, 274], "50": [2, 274], "85": [2, 274], "90": [2, 274], "96": [2, 274], "97": [2, 274], "118": [2, 274], "131": [2, 274], "160": [2, 274], "195": [2, 274], "196": [2, 274], "197": [2, 274] }, { "38": [1, 610] }, { "34": [1, 611], "38": [1, 509] }, { "34": [1, 612], "38": [1, 509] }, { "34": [1, 613], "38": [1, 509] }, { "4": [2, 300], "5": [2, 300], "31": [2, 300], "32": [2, 300], "34": [2, 300], "35": [2, 300], "36": [2, 300], "37": [2, 300], "38": [2, 300], "39": [2, 300], "41": [2, 300], "42": [2, 300], "44": [2, 300], "45": [2, 300], "47": [2, 300], "48": [2, 300], "49": [2, 300], "50": [2, 300], "90": [2, 300], "96": [2, 300], "97": [2, 300], "118": [2, 300], "131": [2, 300], "195": [2, 300], "196": [2, 300], "197": [2, 300] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "4": [2, 302], "5": [2, 302], "31": [2, 302], "32": [2, 302], "34": [2, 302], "35": [2, 302], "36": [2, 302], "37": [2, 302], "38": [2, 302], "39": [2, 302], "41": [2, 302], "42": [2, 302], "44": [2, 302], "45": [2, 302], "47": [2, 302], "48": [2, 302], "49": [2, 302], "50": [2, 302], "90": [2, 302], "96": [2, 302], "97": [2, 302], "118": [2, 302], "131": [2, 302], "195": [2, 302], "196": [2, 302], "197": [2, 302] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "196": [1, 615], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "40": [1, 293], "42": [1, 619], "85": [1, 620] }, { "34": [1, 621], "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "194": [1, 626] }, { "39": [1, 203], "74": [1, 204] }, { "34": [1, 628], "38": [1, 633], "39": [1, 203], "74": [1, 204], "80": [1, 189], "126": [1, 177] }, { "39": [1, 203], "74": [1, 204], "80": [1, 189] }, { "34": [1, 636] }, { "4": [2, 403], "5": [2, 403], "31": [2, 403], "32": [2, 403], "34": [2, 403], "35": [2, 403], "36": [2, 403], "37": [2, 403], "38": [2, 403], "39": [2, 403], "41": [2, 403], "42": [2, 403], "44": [2, 403], "45": [2, 403], "47": [2, 403], "48": [2, 403], "49": [2, 403], "50": [2, 403], "90": [2, 403], "96": [2, 403], "97": [2, 403], "118": [2, 403], "131": [2, 403], "195": [2, 403], "196": [2, 403], "197": [2, 403] }, { "80": [1, 189] }, { "80": [1, 189] }, { "4": [2, 369], "5": [2, 369], "31": [2, 369], "32": [2, 369], "34": [2, 369], "35": [2, 369], "36": [2, 369], "37": [2, 369], "38": [2, 369], "39": [2, 369], "41": [2, 369], "42": [2, 369], "44": [2, 369], "45": [2, 369], "47": [2, 369], "48": [2, 369], "49": [2, 369], "50": [2, 369], "90": [2, 369], "96": [2, 369], "97": [2, 369], "118": [2, 369], "131": [2, 369], "195": [2, 369], "196": [2, 369], "197": [2, 369] }, { "34": [1, 639] }, { "34": [1, 640] }, { "4": [2, 360], "5": [2, 360], "31": [2, 360], "32": [2, 360], "34": [2, 360], "35": [2, 360], "36": [2, 360], "37": [2, 360], "38": [2, 360], "39": [2, 360], "41": [2, 360], "42": [2, 360], "44": [2, 360], "45": [2, 360], "47": [2, 360], "48": [2, 360], "49": [2, 360], "50": [2, 360], "90": [2, 360], "96": [2, 360], "97": [2, 360], "118": [2, 360], "131": [2, 360], "195": [2, 360], "196": [2, 360], "197": [2, 360] }, { "4": [2, 203], "5": [2, 203], "32": [2, 203], "34": [2, 203], "38": [2, 203], "42": [2, 203], "44": [2, 203], "48": [2, 203], "90": [2, 203], "131": [2, 203], "195": [2, 203], "196": [2, 203] }, { "34": [1, 641], "40": [1, 293] }, { "34": [1, 642] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "4": [2, 179], "5": [2, 179], "34": [2, 179], "38": [2, 179], "44": [2, 179], "48": [2, 179], "96": [2, 179], "131": [2, 179], "196": [2, 179] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "160": [2, 198] }, { "160": [2, 199] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [2, 184], "5": [2, 184], "34": [2, 184], "38": [2, 184], "44": [2, 184], "48": [2, 184], "89": [2, 184], "96": [2, 184], "131": [2, 184], "196": [2, 184] }, { "4": [2, 185], "5": [2, 185], "34": [2, 185], "38": [2, 185], "44": [2, 185], "48": [2, 185], "89": [2, 185], "96": [2, 185], "131": [2, 185], "196": [2, 185] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "4": [2, 193], "5": [2, 193], "34": [2, 193], "38": [2, 193], "40": [2, 193], "44": [2, 193], "48": [2, 193], "89": [2, 193], "96": [2, 193], "131": [2, 193], "196": [2, 193] }, { "69": [1, 95], "92": [1, 57], "94": [1, 443], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "33": [1, 224] }, { "4": [2, 190], "5": [2, 190], "34": [2, 190], "38": [2, 190], "44": [2, 190], "48": [2, 190], "89": [2, 190], "96": [2, 190], "131": [2, 190], "196": [2, 190] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "136": [1, 80], "145": [1, 72], "147": [1, 60], "148": [1, 32], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "173": [1, 658], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "194": [1, 659] }, { "4": [2, 115], "5": [2, 115], "43": [2, 115], "46": [2, 115], "194": [2, 115] }, { "34": [1, 660], "38": [1, 661] }, { "34": [2, 117], "38": [2, 117] }, { "42": [1, 619], "85": [1, 620] }, { "1": [2, 124], "33": [2, 124], "55": [2, 124], "69": [2, 124], "89": [2, 124], "91": [2, 124], "92": [2, 124], "97": [2, 124], "98": [2, 124], "100": [2, 124], "101": [2, 124], "106": [2, 124], "108": [2, 124], "113": [2, 124], "114": [2, 124], "115": [2, 124], "118": [2, 124], "122": [2, 124], "123": [2, 124], "125": [2, 124], "127": [2, 124], "134": [2, 124], "135": [2, 124], "136": [2, 124], "145": [2, 124], "147": [2, 124], "148": [2, 124], "149": [2, 124], "150": [2, 124], "151": [2, 124], "153": [2, 124], "154": [2, 124], "155": [2, 124], "158": [2, 124], "167": [2, 124], "170": [2, 124], "172": [2, 124], "175": [2, 124], "178": [2, 124], "179": [2, 124], "181": [2, 124] }, { "4": [1, 25], "5": [1, 24], "194": [1, 452] }, { "1": [2, 127], "33": [2, 127], "55": [2, 127], "69": [2, 127], "89": [2, 127], "91": [2, 127], "92": [2, 127], "97": [2, 127], "98": [2, 127], "100": [2, 127], "101": [2, 127], "106": [2, 127], "108": [2, 127], "113": [2, 127], "114": [2, 127], "115": [2, 127], "118": [2, 127], "122": [2, 127], "123": [2, 127], "125": [2, 127], "127": [2, 127], "134": [2, 127], "135": [2, 127], "136": [2, 127], "145": [2, 127], "147": [2, 127], "148": [2, 127], "149": [2, 127], "150": [2, 127], "151": [2, 127], "153": [2, 127], "154": [2, 127], "155": [2, 127], "158": [2, 127], "167": [2, 127], "170": [2, 127], "172": [2, 127], "175": [2, 127], "178": [2, 127], "179": [2, 127], "181": [2, 127] }, { "92": [1, 560], "167": [1, 561], "170": [1, 559], "178": [1, 558], "196": [1, 665] }, { "92": [2, 129], "167": [2, 129], "170": [2, 129], "178": [2, 129], "196": [2, 129] }, { "33": [1, 448], "43": [1, 670], "46": [1, 669] }, { "33": [1, 448], "43": [1, 670], "46": [1, 669] }, { "33": [1, 448], "43": [1, 670], "46": [1, 669] }, { "33": [1, 448], "43": [1, 670], "46": [1, 669] }, { "4": [1, 25], "5": [1, 24], "38": [1, 679], "194": [1, 456] }, { "4": [2, 83], "5": [2, 83], "38": [2, 83], "194": [2, 83] }, { "1": [2, 79], "33": [2, 79], "55": [2, 79], "69": [2, 79], "89": [2, 79], "91": [2, 79], "92": [2, 79], "97": [2, 79], "98": [2, 79], "100": [2, 79], "101": [2, 79], "106": [2, 79], "108": [2, 79], "113": [2, 79], "114": [2, 79], "115": [2, 79], "118": [2, 79], "122": [2, 79], "123": [2, 79], "125": [2, 79], "127": [2, 79], "134": [2, 79], "135": [2, 79], "136": [2, 79], "145": [2, 79], "147": [2, 79], "148": [2, 79], "149": [2, 79], "150": [2, 79], "151": [2, 79], "153": [2, 79], "154": [2, 79], "155": [2, 79], "158": [2, 79], "167": [2, 79], "170": [2, 79], "172": [2, 79], "175": [2, 79], "178": [2, 79], "179": [2, 79], "181": [2, 79] }, { "1": [2, 81], "33": [2, 81], "55": [2, 81], "69": [2, 81], "89": [2, 81], "91": [2, 81], "92": [2, 81], "97": [2, 81], "98": [2, 81], "100": [2, 81], "101": [2, 81], "106": [2, 81], "108": [2, 81], "113": [2, 81], "114": [2, 81], "115": [2, 81], "118": [2, 81], "122": [2, 81], "123": [2, 81], "125": [2, 81], "127": [2, 81], "134": [2, 81], "135": [2, 81], "136": [2, 81], "145": [2, 81], "147": [2, 81], "148": [2, 81], "149": [2, 81], "150": [2, 81], "151": [2, 81], "153": [2, 81], "154": [2, 81], "155": [2, 81], "158": [2, 81], "167": [2, 81], "170": [2, 81], "172": [2, 81], "175": [2, 81], "178": [2, 81], "179": [2, 81], "181": [2, 81] }, { "92": [1, 574], "125": [1, 569], "139": [1, 568], "149": [1, 573], "154": [1, 572], "170": [1, 571], "196": [1, 680] }, { "92": [2, 85], "125": [2, 85], "139": [2, 85], "149": [2, 85], "154": [2, 85], "170": [2, 85], "196": [2, 85] }, { "69": [1, 394], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 399], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "69": [1, 394], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 399], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "69": [2, 87], "92": [2, 87], "97": [2, 87], "98": [2, 87], "100": [2, 87], "101": [2, 87], "108": [2, 87], "110": [2, 87], "111": [2, 87], "113": [2, 87], "115": [2, 87], "118": [2, 87], "119": [2, 87], "122": [2, 87], "123": [2, 87], "124": [2, 87], "125": [2, 87], "127": [2, 87], "134": [2, 87], "135": [2, 87], "141": [2, 87], "142": [2, 87], "145": [2, 87], "147": [2, 87], "149": [2, 87], "150": [2, 87], "151": [2, 87], "154": [2, 87], "155": [2, 87], "165": [2, 87], "167": [2, 87], "170": [2, 87], "172": [2, 87], "174": [2, 87], "176": [2, 87], "178": [2, 87], "179": [2, 87], "180": [2, 87], "181": [2, 87], "198": [2, 87], "199": [2, 87] }, { "149": [1, 687], "170": [1, 686] }, { "92": [1, 690], "154": [1, 688], "170": [1, 689] }, { "69": [2, 92], "92": [2, 92], "97": [2, 92], "98": [2, 92], "100": [2, 92], "101": [2, 92], "108": [2, 92], "110": [2, 92], "111": [2, 92], "113": [2, 92], "115": [2, 92], "118": [2, 92], "119": [2, 92], "122": [2, 92], "123": [2, 92], "124": [2, 92], "125": [2, 92], "127": [2, 92], "134": [2, 92], "135": [2, 92], "141": [2, 92], "142": [2, 92], "145": [2, 92], "147": [2, 92], "149": [2, 92], "150": [2, 92], "151": [2, 92], "154": [2, 92], "155": [2, 92], "165": [2, 92], "167": [2, 92], "170": [2, 92], "172": [2, 92], "174": [2, 92], "176": [2, 92], "178": [2, 92], "179": [2, 92], "180": [2, 92], "181": [2, 92], "198": [2, 92], "199": [2, 92] }, { "1": [2, 29], "33": [2, 29], "55": [2, 29], "69": [2, 29], "89": [2, 29], "92": [2, 29], "97": [2, 29], "98": [2, 29], "100": [2, 29], "101": [2, 29], "108": [2, 29], "113": [2, 29], "115": [2, 29], "118": [2, 29], "122": [2, 29], "123": [2, 29], "125": [2, 29], "127": [2, 29], "134": [2, 29], "135": [2, 29], "145": [2, 29], "147": [2, 29], "149": [2, 29], "150": [2, 29], "151": [2, 29], "153": [2, 29], "154": [2, 29], "155": [2, 29], "158": [2, 29], "167": [2, 29], "170": [2, 29], "172": [2, 29], "175": [2, 29], "178": [2, 29], "179": [2, 29], "181": [2, 29] }, { "1": [2, 30], "33": [2, 30], "55": [2, 30], "69": [2, 30], "89": [2, 30], "92": [2, 30], "97": [2, 30], "98": [2, 30], "100": [2, 30], "101": [2, 30], "108": [2, 30], "113": [2, 30], "115": [2, 30], "118": [2, 30], "122": [2, 30], "123": [2, 30], "125": [2, 30], "127": [2, 30], "134": [2, 30], "135": [2, 30], "145": [2, 30], "147": [2, 30], "149": [2, 30], "150": [2, 30], "151": [2, 30], "153": [2, 30], "154": [2, 30], "155": [2, 30], "158": [2, 30], "167": [2, 30], "170": [2, 30], "172": [2, 30], "175": [2, 30], "178": [2, 30], "179": [2, 30], "181": [2, 30] }, { "44": [1, 691] }, { "69": [1, 698], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 692], "110": [1, 107], "111": [1, 108], "113": [1, 694], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 696], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 695], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 693], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "4": [1, 704], "5": [1, 581], "44": [2, 36] }, { "4": [2, 34], "5": [2, 34], "44": [2, 34] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "4": [1, 25], "5": [1, 24], "44": [1, 706] }, { "1": [2, 24], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "44": [1, 708] }, { "48": [1, 709] }, { "48": [1, 710] }, { "44": [2, 51], "48": [2, 51] }, { "4": [2, 18], "5": [2, 18], "33": [2, 18], "38": [2, 18], "40": [1, 711], "44": [2, 18], "48": [2, 18], "131": [2, 18] }, { "38": [1, 713], "44": [2, 53], "48": [2, 53], "131": [1, 712] }, { "1": [2, 14] }, { "1": [2, 15] }, { "34": [2, 276], "38": [2, 276], "42": [2, 276], "90": [2, 276], "97": [2, 276], "118": [2, 276], "196": [2, 276] }, { "34": [2, 279], "38": [2, 279], "42": [2, 279], "90": [2, 279], "97": [2, 279], "118": [2, 279], "196": [2, 279] }, { "4": [2, 281], "5": [2, 281], "31": [1, 473], "32": [2, 281], "34": [2, 281], "35": [1, 471], "37": [2, 281], "38": [2, 281], "39": [2, 281], "41": [1, 472], "42": [2, 281], "44": [2, 281], "48": [2, 281], "90": [2, 281], "97": [2, 281], "118": [2, 281], "131": [2, 281], "195": [2, 281], "196": [2, 281] }, { "4": [2, 282], "5": [2, 282], "31": [1, 473], "32": [2, 282], "34": [2, 282], "35": [1, 471], "37": [2, 282], "38": [2, 282], "39": [2, 282], "41": [1, 472], "42": [2, 282], "44": [2, 282], "48": [2, 282], "90": [2, 282], "97": [2, 282], "118": [2, 282], "131": [2, 282], "195": [2, 282], "196": [2, 282] }, { "4": [2, 284], "5": [2, 284], "31": [2, 284], "32": [2, 284], "34": [2, 284], "35": [2, 284], "36": [1, 474], "37": [2, 284], "38": [2, 284], "39": [2, 284], "41": [2, 284], "42": [2, 284], "44": [2, 284], "48": [2, 284], "90": [2, 284], "97": [2, 284], "118": [2, 284], "131": [2, 284], "195": [2, 284], "196": [2, 284] }, { "4": [2, 285], "5": [2, 285], "31": [2, 285], "32": [2, 285], "34": [2, 285], "35": [2, 285], "36": [1, 474], "37": [2, 285], "38": [2, 285], "39": [2, 285], "41": [2, 285], "42": [2, 285], "44": [2, 285], "48": [2, 285], "90": [2, 285], "97": [2, 285], "118": [2, 285], "131": [2, 285], "195": [2, 285], "196": [2, 285] }, { "4": [2, 286], "5": [2, 286], "31": [2, 286], "32": [2, 286], "34": [2, 286], "35": [2, 286], "36": [1, 474], "37": [2, 286], "38": [2, 286], "39": [2, 286], "41": [2, 286], "42": [2, 286], "44": [2, 286], "48": [2, 286], "90": [2, 286], "97": [2, 286], "118": [2, 286], "131": [2, 286], "195": [2, 286], "196": [2, 286] }, { "4": [2, 288], "5": [2, 288], "31": [2, 288], "32": [2, 288], "34": [2, 288], "35": [2, 288], "36": [2, 288], "37": [2, 288], "38": [2, 288], "39": [2, 288], "41": [2, 288], "42": [2, 288], "44": [2, 288], "48": [2, 288], "90": [2, 288], "97": [2, 288], "118": [2, 288], "131": [2, 288], "195": [2, 288], "196": [2, 288] }, { "4": [2, 263], "5": [2, 263], "31": [2, 263], "32": [2, 263], "34": [2, 263], "35": [2, 263], "36": [2, 263], "37": [2, 263], "38": [2, 263], "39": [2, 263], "41": [2, 263], "42": [2, 263], "44": [2, 263], "48": [2, 263], "90": [2, 263], "96": [2, 263], "97": [2, 263], "118": [2, 263], "131": [2, 263], "195": [2, 263], "196": [2, 263] }, { "4": [2, 57], "5": [2, 57], "33": [2, 57], "34": [2, 57], "38": [2, 57], "44": [2, 57], "55": [2, 57], "69": [2, 57], "89": [2, 57], "91": [2, 57], "92": [2, 57], "97": [2, 57], "98": [2, 57], "100": [2, 57], "101": [2, 57], "106": [2, 57], "108": [2, 57], "113": [2, 57], "114": [2, 57], "115": [2, 57], "118": [2, 57], "122": [2, 57], "123": [2, 57], "125": [2, 57], "127": [2, 57], "134": [2, 57], "135": [2, 57], "136": [2, 57], "145": [2, 57], "147": [2, 57], "148": [2, 57], "149": [2, 57], "150": [2, 57], "151": [2, 57], "153": [2, 57], "154": [2, 57], "155": [2, 57], "158": [2, 57], "167": [2, 57], "170": [2, 57], "172": [2, 57], "175": [2, 57], "178": [2, 57], "179": [2, 57], "181": [2, 57], "194": [2, 57], "196": [2, 57] }, { "4": [2, 58], "5": [2, 58], "33": [2, 58], "34": [2, 58], "38": [2, 58], "44": [2, 58], "55": [2, 58], "69": [2, 58], "89": [2, 58], "91": [2, 58], "92": [2, 58], "97": [2, 58], "98": [2, 58], "100": [2, 58], "101": [2, 58], "106": [2, 58], "108": [2, 58], "113": [2, 58], "114": [2, 58], "115": [2, 58], "118": [2, 58], "122": [2, 58], "123": [2, 58], "125": [2, 58], "127": [2, 58], "134": [2, 58], "135": [2, 58], "136": [2, 58], "145": [2, 58], "147": [2, 58], "148": [2, 58], "149": [2, 58], "150": [2, 58], "151": [2, 58], "153": [2, 58], "154": [2, 58], "155": [2, 58], "158": [2, 58], "167": [2, 58], "170": [2, 58], "172": [2, 58], "175": [2, 58], "178": [2, 58], "179": [2, 58], "181": [2, 58], "194": [2, 58], "196": [2, 58] }, { "1": [2, 74], "33": [2, 74], "55": [2, 74], "69": [2, 74], "89": [2, 74], "91": [2, 74], "92": [2, 74], "97": [2, 74], "98": [2, 74], "100": [2, 74], "101": [2, 74], "106": [2, 74], "108": [2, 74], "113": [2, 74], "114": [2, 74], "115": [2, 74], "118": [2, 74], "122": [2, 74], "123": [2, 74], "125": [2, 74], "127": [2, 74], "134": [2, 74], "135": [2, 74], "136": [2, 74], "145": [2, 74], "147": [2, 74], "148": [2, 74], "149": [2, 74], "150": [2, 74], "151": [2, 74], "153": [2, 74], "154": [2, 74], "155": [2, 74], "158": [2, 74], "167": [2, 74], "170": [2, 74], "172": [2, 74], "173": [2, 74], "175": [2, 74], "178": [2, 74], "179": [2, 74], "181": [2, 74], "196": [2, 74] }, { "4": [2, 163], "5": [2, 163], "34": [2, 163], "38": [1, 118], "44": [2, 163], "48": [2, 163], "131": [1, 119], "196": [2, 163] }, { "4": [2, 164], "5": [2, 164], "34": [2, 164], "44": [2, 164], "48": [2, 164], "196": [2, 164] }, { "34": [2, 211], "38": [2, 211], "42": [2, 211], "90": [2, 211], "96": [2, 211], "97": [2, 211], "118": [2, 211], "196": [2, 211] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "34": [2, 213] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 717], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "4": [2, 290], "5": [2, 290], "31": [2, 290], "32": [2, 290], "34": [2, 290], "35": [2, 290], "36": [2, 290], "37": [2, 290], "38": [2, 290], "39": [2, 290], "41": [2, 290], "42": [2, 290], "44": [2, 290], "45": [2, 290], "47": [2, 290], "48": [2, 290], "49": [2, 290], "50": [2, 290], "90": [2, 290], "96": [2, 290], "97": [2, 290], "118": [2, 290], "131": [2, 290], "195": [2, 290], "196": [2, 290], "197": [2, 290] }, { "4": [2, 289], "5": [2, 289], "31": [2, 289], "32": [2, 289], "34": [2, 289], "35": [2, 289], "36": [2, 289], "37": [2, 289], "38": [2, 289], "39": [2, 289], "41": [2, 289], "42": [2, 289], "44": [2, 289], "45": [2, 289], "47": [2, 289], "48": [2, 289], "49": [2, 289], "50": [2, 289], "90": [2, 289], "96": [2, 289], "97": [2, 289], "118": [2, 289], "131": [2, 289], "195": [2, 289], "196": [2, 289], "197": [2, 289] }, { "4": [2, 291], "5": [2, 291], "31": [2, 291], "32": [2, 291], "34": [2, 291], "35": [2, 291], "36": [2, 291], "37": [2, 291], "38": [2, 291], "39": [2, 291], "41": [2, 291], "42": [2, 291], "44": [2, 291], "45": [2, 291], "47": [2, 291], "48": [2, 291], "49": [2, 291], "50": [2, 291], "90": [2, 291], "96": [2, 291], "97": [2, 291], "118": [2, 291], "131": [2, 291], "195": [2, 291], "196": [2, 291], "197": [2, 291] }, { "34": [2, 345], "38": [2, 345], "90": [2, 345] }, { "4": [2, 303], "5": [2, 303], "31": [2, 303], "32": [2, 303], "34": [2, 303], "35": [2, 303], "36": [2, 303], "37": [2, 303], "38": [2, 303], "39": [2, 303], "41": [2, 303], "42": [2, 303], "44": [2, 303], "45": [2, 303], "47": [2, 303], "48": [2, 303], "49": [2, 303], "50": [2, 303], "90": [2, 303], "96": [2, 303], "97": [2, 303], "118": [2, 303], "131": [2, 303], "195": [2, 303], "196": [2, 303], "197": [2, 303] }, { "40": [1, 293], "46": [1, 718] }, { "38": [2, 304], "196": [2, 304] }, { "4": [2, 298], "5": [2, 298], "31": [2, 298], "32": [2, 298], "34": [2, 298], "35": [2, 298], "36": [2, 298], "37": [2, 298], "38": [2, 298], "39": [2, 298], "41": [2, 298], "42": [2, 298], "44": [2, 298], "45": [2, 298], "47": [2, 298], "48": [2, 298], "49": [2, 298], "50": [2, 298], "90": [2, 298], "96": [2, 298], "97": [2, 298], "118": [2, 298], "131": [2, 298], "195": [2, 298], "196": [2, 298], "197": [2, 298] }, { "52": [1, 720], "53": [1, 731], "54": [1, 721], "58": [1, 722], "62": [1, 723], "65": [1, 732], "66": [1, 734], "69": [1, 730], "71": [1, 724], "73": [1, 733], "75": [1, 725], "76": [1, 728], "81": [1, 729], "83": [1, 727], "86": [1, 726], "194": [1, 735] }, { "4": [2, 409], "5": [2, 409], "31": [2, 409], "32": [2, 409], "34": [2, 409], "35": [2, 409], "36": [2, 409], "37": [2, 409], "38": [2, 409], "39": [2, 409], "41": [2, 409], "42": [2, 409], "44": [2, 409], "45": [2, 409], "47": [2, 409], "48": [2, 409], "49": [2, 409], "50": [2, 409], "90": [2, 409], "96": [2, 409], "97": [2, 409], "118": [2, 409], "131": [2, 409], "195": [2, 409], "196": [2, 409], "197": [2, 409] }, { "4": [2, 404], "5": [2, 404], "31": [2, 404], "32": [2, 404], "34": [2, 404], "35": [2, 404], "36": [2, 404], "37": [2, 404], "38": [2, 404], "39": [2, 404], "41": [2, 404], "42": [2, 404], "44": [2, 404], "45": [2, 404], "47": [2, 404], "48": [2, 404], "49": [2, 404], "50": [2, 404], "90": [2, 404], "96": [2, 404], "97": [2, 404], "118": [2, 404], "131": [2, 404], "195": [2, 404], "196": [2, 404], "197": [2, 404] }, { "34": [1, 736], "38": [1, 737] }, { "42": [1, 619], "85": [1, 620] }, { "34": [1, 739], "38": [1, 740] }, { "34": [2, 307], "38": [2, 307] }, { "100": [1, 743], "101": [1, 745], "115": [1, 749], "122": [1, 744], "123": [1, 746], "134": [1, 747], "145": [1, 748], "179": [1, 750] }, { "38": [1, 751] }, { "4": [2, 371], "5": [2, 371], "31": [2, 371], "32": [2, 371], "34": [2, 371], "35": [2, 371], "36": [2, 371], "37": [2, 371], "38": [2, 371], "39": [2, 371], "41": [2, 371], "42": [2, 371], "44": [2, 371], "45": [2, 371], "47": [2, 371], "48": [2, 371], "49": [2, 371], "50": [2, 371], "90": [2, 371], "96": [2, 371], "97": [2, 371], "118": [2, 371], "131": [2, 371], "195": [2, 371], "196": [2, 371], "197": [2, 371] }, { "34": [1, 752] }, { "34": [1, 753], "38": [1, 754], "39": [1, 203], "74": [1, 204] }, { "34": [1, 756], "38": [1, 757] }, { "34": [2, 373], "38": [2, 373], "39": [2, 373], "74": [2, 373] }, { "34": [2, 374], "38": [2, 374], "39": [2, 374], "74": [2, 374] }, { "38": [1, 758] }, { "34": [1, 759] }, { "4": [2, 389], "5": [2, 389], "31": [2, 389], "32": [2, 389], "34": [2, 389], "35": [2, 389], "36": [2, 389], "37": [2, 389], "38": [2, 389], "39": [2, 389], "41": [2, 389], "42": [2, 389], "44": [2, 389], "45": [2, 389], "47": [2, 389], "48": [2, 389], "49": [2, 389], "50": [2, 389], "90": [2, 389], "96": [2, 389], "97": [2, 389], "118": [2, 389], "131": [2, 389], "195": [2, 389], "196": [2, 389], "197": [2, 389] }, { "34": [1, 760] }, { "34": [1, 761] }, { "4": [2, 386], "5": [2, 386], "31": [2, 386], "32": [2, 386], "34": [2, 386], "35": [2, 386], "36": [2, 386], "37": [2, 386], "38": [2, 386], "39": [2, 386], "41": [2, 386], "42": [2, 386], "44": [2, 386], "45": [2, 386], "47": [2, 386], "48": [2, 386], "49": [2, 386], "50": [2, 386], "90": [2, 386], "96": [2, 386], "97": [2, 386], "118": [2, 386], "131": [2, 386], "195": [2, 386], "196": [2, 386], "197": [2, 386] }, { "4": [2, 387], "5": [2, 387], "31": [2, 387], "32": [2, 387], "34": [2, 387], "35": [2, 387], "36": [2, 387], "37": [2, 387], "38": [2, 387], "39": [2, 387], "41": [2, 387], "42": [2, 387], "44": [2, 387], "45": [2, 387], "47": [2, 387], "48": [2, 387], "49": [2, 387], "50": [2, 387], "90": [2, 387], "96": [2, 387], "97": [2, 387], "118": [2, 387], "131": [2, 387], "195": [2, 387], "196": [2, 387], "197": [2, 387] }, { "4": [2, 259], "5": [2, 259], "32": [2, 259], "34": [2, 259], "38": [2, 259], "42": [2, 259], "44": [2, 259], "48": [2, 259], "90": [2, 259], "96": [2, 259], "97": [2, 259], "118": [2, 259], "131": [2, 259], "195": [2, 259], "196": [2, 259] }, { "4": [2, 262], "5": [2, 262], "32": [2, 262], "34": [2, 262], "38": [2, 262], "42": [2, 262], "44": [2, 262], "48": [2, 262], "90": [2, 262], "96": [2, 262], "97": [2, 262], "118": [2, 262], "131": [2, 262], "195": [2, 262], "196": [2, 262] }, { "90": [1, 762] }, { "34": [1, 763] }, { "34": [1, 764] }, { "34": [1, 765] }, { "34": [1, 766] }, { "38": [2, 480], "40": [1, 293], "160": [2, 480] }, { "34": [2, 195], "38": [2, 195] }, { "34": [2, 196], "38": [2, 196] }, { "34": [2, 197], "38": [2, 197] }, { "4": [2, 191], "5": [2, 191], "34": [2, 191], "38": [2, 191], "44": [2, 191], "48": [2, 191], "89": [2, 191], "96": [2, 191], "131": [2, 191], "196": [2, 191] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "136": [1, 80], "145": [1, 72], "147": [1, 60], "148": [1, 32], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "173": [1, 658], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78], "196": [1, 767] }, { "33": [2, 61], "55": [2, 61], "69": [2, 61], "89": [2, 61], "92": [2, 61], "97": [2, 61], "98": [2, 61], "100": [2, 61], "101": [2, 61], "108": [2, 61], "113": [2, 61], "115": [2, 61], "118": [2, 61], "122": [2, 61], "123": [2, 61], "125": [2, 61], "127": [2, 61], "134": [2, 61], "135": [2, 61], "136": [2, 61], "145": [2, 61], "147": [2, 61], "148": [2, 61], "149": [2, 61], "150": [2, 61], "151": [2, 61], "153": [2, 61], "154": [2, 61], "155": [2, 61], "158": [2, 61], "167": [2, 61], "170": [2, 61], "172": [2, 61], "173": [2, 61], "175": [2, 61], "178": [2, 61], "179": [2, 61], "181": [2, 61], "196": [2, 61] }, { "33": [2, 70], "55": [2, 70], "69": [2, 70], "89": [2, 70], "92": [2, 70], "97": [2, 70], "98": [2, 70], "100": [2, 70], "101": [2, 70], "108": [2, 70], "113": [2, 70], "115": [2, 70], "118": [2, 70], "122": [2, 70], "123": [2, 70], "125": [2, 70], "127": [2, 70], "134": [2, 70], "135": [2, 70], "136": [2, 70], "145": [2, 70], "147": [2, 70], "148": [2, 70], "149": [2, 70], "150": [2, 70], "151": [2, 70], "153": [2, 70], "154": [2, 70], "155": [2, 70], "158": [2, 70], "167": [2, 70], "170": [2, 70], "172": [2, 70], "173": [2, 70], "175": [2, 70], "178": [2, 70], "179": [2, 70], "181": [2, 70], "196": [2, 70] }, { "44": [1, 769] }, { "33": [2, 72], "55": [2, 72], "69": [2, 72], "89": [2, 72], "92": [2, 72], "97": [2, 72], "98": [2, 72], "100": [2, 72], "101": [2, 72], "108": [2, 72], "113": [2, 72], "115": [2, 72], "118": [2, 72], "122": [2, 72], "123": [2, 72], "125": [2, 72], "127": [2, 72], "134": [2, 72], "135": [2, 72], "136": [2, 72], "145": [2, 72], "147": [2, 72], "148": [2, 72], "149": [2, 72], "150": [2, 72], "151": [2, 72], "153": [2, 72], "154": [2, 72], "155": [2, 72], "158": [2, 72], "167": [2, 72], "170": [2, 72], "172": [2, 72], "173": [2, 72], "175": [2, 72], "178": [2, 72], "179": [2, 72], "181": [2, 72], "196": [2, 72] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "136": [1, 80], "145": [1, 72], "147": [1, 60], "148": [1, 32], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "173": [1, 658], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [2, 116], "5": [2, 116], "43": [2, 116], "46": [2, 116], "194": [2, 116] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "34": [2, 119], "38": [2, 119] }, { "1": [2, 125], "33": [2, 125], "55": [2, 125], "69": [2, 125], "89": [2, 125], "91": [2, 125], "92": [2, 125], "97": [2, 125], "98": [2, 125], "100": [2, 125], "101": [2, 125], "106": [2, 125], "108": [2, 125], "113": [2, 125], "114": [2, 125], "115": [2, 125], "118": [2, 125], "122": [2, 125], "123": [2, 125], "125": [2, 125], "127": [2, 125], "134": [2, 125], "135": [2, 125], "136": [2, 125], "145": [2, 125], "147": [2, 125], "148": [2, 125], "149": [2, 125], "150": [2, 125], "151": [2, 125], "153": [2, 125], "154": [2, 125], "155": [2, 125], "158": [2, 125], "167": [2, 125], "170": [2, 125], "172": [2, 125], "175": [2, 125], "178": [2, 125], "179": [2, 125], "181": [2, 125] }, { "194": [1, 452] }, { "1": [2, 128], "33": [2, 128], "55": [2, 128], "69": [2, 128], "89": [2, 128], "91": [2, 128], "92": [2, 128], "97": [2, 128], "98": [2, 128], "100": [2, 128], "101": [2, 128], "106": [2, 128], "108": [2, 128], "113": [2, 128], "114": [2, 128], "115": [2, 128], "118": [2, 128], "122": [2, 128], "123": [2, 128], "125": [2, 128], "127": [2, 128], "134": [2, 128], "135": [2, 128], "136": [2, 128], "145": [2, 128], "147": [2, 128], "148": [2, 128], "149": [2, 128], "150": [2, 128], "151": [2, 128], "153": [2, 128], "154": [2, 128], "155": [2, 128], "158": [2, 128], "167": [2, 128], "170": [2, 128], "172": [2, 128], "175": [2, 128], "178": [2, 128], "179": [2, 128], "181": [2, 128] }, { "92": [2, 130], "167": [2, 130], "170": [2, 130], "178": [2, 130], "196": [2, 130] }, { "43": [1, 670], "46": [1, 669] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "33": [2, 131], "55": [2, 131], "69": [2, 131], "89": [2, 131], "92": [2, 131], "97": [2, 131], "98": [2, 131], "100": [2, 131], "101": [2, 131], "108": [2, 131], "113": [2, 131], "115": [2, 131], "118": [2, 131], "122": [2, 131], "123": [2, 131], "125": [2, 131], "127": [2, 131], "134": [2, 131], "135": [2, 131], "145": [2, 131], "147": [2, 131], "149": [2, 131], "150": [2, 131], "151": [2, 131], "153": [2, 131], "154": [2, 131], "155": [2, 131], "167": [2, 131], "170": [2, 131], "172": [2, 131], "175": [2, 131], "178": [2, 131], "179": [2, 131], "181": [2, 131], "194": [2, 131] }, { "33": [2, 132], "55": [2, 132], "69": [2, 132], "89": [2, 132], "92": [2, 132], "97": [2, 132], "98": [2, 132], "100": [2, 132], "101": [2, 132], "108": [2, 132], "113": [2, 132], "115": [2, 132], "118": [2, 132], "122": [2, 132], "123": [2, 132], "125": [2, 132], "127": [2, 132], "134": [2, 132], "135": [2, 132], "145": [2, 132], "147": [2, 132], "149": [2, 132], "150": [2, 132], "151": [2, 132], "153": [2, 132], "154": [2, 132], "155": [2, 132], "167": [2, 132], "170": [2, 132], "172": [2, 132], "175": [2, 132], "178": [2, 132], "179": [2, 132], "181": [2, 132], "194": [2, 132] }, { "43": [1, 670], "46": [1, 669] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "43": [1, 670], "46": [1, 669] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "43": [1, 670], "46": [1, 669] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78], "194": [1, 782] }, { "1": [2, 78], "33": [2, 78], "55": [2, 78], "69": [2, 78], "89": [2, 78], "91": [2, 78], "92": [2, 78], "97": [2, 78], "98": [2, 78], "100": [2, 78], "101": [2, 78], "106": [2, 78], "108": [2, 78], "113": [2, 78], "114": [2, 78], "115": [2, 78], "118": [2, 78], "122": [2, 78], "123": [2, 78], "125": [2, 78], "127": [2, 78], "134": [2, 78], "135": [2, 78], "136": [2, 78], "145": [2, 78], "147": [2, 78], "148": [2, 78], "149": [2, 78], "150": [2, 78], "151": [2, 78], "153": [2, 78], "154": [2, 78], "155": [2, 78], "158": [2, 78], "167": [2, 78], "170": [2, 78], "172": [2, 78], "175": [2, 78], "178": [2, 78], "179": [2, 78], "181": [2, 78] }, { "194": [1, 456] }, { "55": [1, 784] }, { "1": [2, 82], "33": [2, 82], "55": [2, 82], "69": [2, 82], "89": [2, 82], "91": [2, 82], "92": [2, 82], "97": [2, 82], "98": [2, 82], "100": [2, 82], "101": [2, 82], "106": [2, 82], "108": [2, 82], "113": [2, 82], "114": [2, 82], "115": [2, 82], "118": [2, 82], "122": [2, 82], "123": [2, 82], "125": [2, 82], "127": [2, 82], "134": [2, 82], "135": [2, 82], "136": [2, 82], "145": [2, 82], "147": [2, 82], "148": [2, 82], "149": [2, 82], "150": [2, 82], "151": [2, 82], "153": [2, 82], "154": [2, 82], "155": [2, 82], "158": [2, 82], "167": [2, 82], "170": [2, 82], "172": [2, 82], "175": [2, 82], "178": [2, 82], "179": [2, 82], "181": [2, 82] }, { "92": [2, 86], "125": [2, 86], "139": [2, 86], "149": [2, 86], "154": [2, 86], "170": [2, 86], "196": [2, 86] }, { "38": [1, 786], "135": [1, 785] }, { "33": [2, 434], "38": [2, 434], "135": [2, 434] }, { "4": [1, 25], "5": [1, 24], "44": [1, 787], "129": [1, 789] }, { "33": [1, 792], "129": [1, 791] }, { "69": [2, 88], "92": [2, 88], "97": [2, 88], "98": [2, 88], "100": [2, 88], "101": [2, 88], "108": [2, 88], "110": [2, 88], "111": [2, 88], "113": [2, 88], "115": [2, 88], "118": [2, 88], "119": [2, 88], "122": [2, 88], "123": [2, 88], "124": [2, 88], "125": [2, 88], "127": [2, 88], "134": [2, 88], "135": [2, 88], "141": [2, 88], "142": [2, 88], "145": [2, 88], "147": [2, 88], "149": [2, 88], "150": [2, 88], "151": [2, 88], "154": [2, 88], "155": [2, 88], "165": [2, 88], "167": [2, 88], "170": [2, 88], "172": [2, 88], "174": [2, 88], "176": [2, 88], "178": [2, 88], "179": [2, 88], "180": [2, 88], "181": [2, 88], "198": [2, 88], "199": [2, 88] }, { "170": [1, 793] }, { "170": [1, 794] }, { "69": [2, 91], "92": [2, 91], "97": [2, 91], "98": [2, 91], "100": [2, 91], "101": [2, 91], "108": [2, 91], "110": [2, 91], "111": [2, 91], "113": [2, 91], "115": [2, 91], "118": [2, 91], "119": [2, 91], "122": [2, 91], "123": [2, 91], "124": [2, 91], "125": [2, 91], "127": [2, 91], "134": [2, 91], "135": [2, 91], "141": [2, 91], "142": [2, 91], "145": [2, 91], "147": [2, 91], "149": [2, 91], "150": [2, 91], "151": [2, 91], "154": [2, 91], "155": [2, 91], "165": [2, 91], "167": [2, 91], "170": [2, 91], "172": [2, 91], "174": [2, 91], "176": [2, 91], "178": [2, 91], "179": [2, 91], "180": [2, 91], "181": [2, 91], "198": [2, 91], "199": [2, 91] }, { "69": [2, 93], "92": [2, 93], "97": [2, 93], "98": [2, 93], "100": [2, 93], "101": [2, 93], "108": [2, 93], "110": [2, 93], "111": [2, 93], "113": [2, 93], "115": [2, 93], "118": [2, 93], "119": [2, 93], "122": [2, 93], "123": [2, 93], "124": [2, 93], "125": [2, 93], "127": [2, 93], "134": [2, 93], "135": [2, 93], "141": [2, 93], "142": [2, 93], "145": [2, 93], "147": [2, 93], "149": [2, 93], "150": [2, 93], "151": [2, 93], "154": [2, 93], "155": [2, 93], "165": [2, 93], "167": [2, 93], "170": [2, 93], "172": [2, 93], "174": [2, 93], "176": [2, 93], "178": [2, 93], "179": [2, 93], "180": [2, 93], "181": [2, 93], "198": [2, 93], "199": [2, 93] }, { "1": [2, 31], "33": [2, 31], "55": [2, 31], "69": [2, 31], "89": [2, 31], "92": [2, 31], "97": [2, 31], "98": [2, 31], "100": [2, 31], "101": [2, 31], "108": [2, 31], "113": [2, 31], "115": [2, 31], "118": [2, 31], "122": [2, 31], "123": [2, 31], "125": [2, 31], "127": [2, 31], "134": [2, 31], "135": [2, 31], "145": [2, 31], "147": [2, 31], "149": [2, 31], "150": [2, 31], "151": [2, 31], "153": [2, 31], "154": [2, 31], "155": [2, 31], "158": [2, 31], "167": [2, 31], "170": [2, 31], "172": [2, 31], "175": [2, 31], "178": [2, 31], "179": [2, 31], "181": [2, 31] }, { "46": [1, 795] }, { "46": [1, 796] }, { "46": [1, 797] }, { "46": [1, 798] }, { "46": [1, 799] }, { "46": [1, 800] }, { "46": [2, 466] }, { "46": [2, 467] }, { "46": [2, 468] }, { "46": [2, 469] }, { "46": [2, 470] }, { "4": [2, 35], "5": [2, 35], "44": [2, 35] }, { "69": [1, 698], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "110": [1, 107], "111": [1, 108], "113": [1, 694], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 696], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 695], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 693], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "46": [1, 801] }, { "1": [2, 23], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "44": [1, 803] }, { "196": [1, 804] }, { "35": [1, 587], "55": [1, 588] }, { "35": [1, 587], "55": [1, 588], "157": [1, 806] }, { "35": [1, 808], "69": [1, 394], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 399], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "3": [1, 211], "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 236], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "95": [1, 214], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 213], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 212], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "34": [2, 212], "38": [2, 212], "42": [2, 212], "90": [2, 212], "96": [2, 212], "97": [2, 212], "118": [2, 212], "196": [2, 212] }, { "34": [1, 811] }, { "34": [1, 812] }, { "33": [1, 813] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "4": [2, 408], "5": [2, 408], "31": [2, 408], "32": [2, 408], "34": [2, 408], "35": [2, 408], "36": [2, 408], "37": [2, 408], "38": [2, 408], "39": [2, 408], "41": [2, 408], "42": [2, 408], "44": [2, 408], "45": [2, 408], "47": [2, 408], "48": [2, 408], "49": [2, 408], "50": [2, 408], "90": [2, 408], "96": [2, 408], "97": [2, 408], "118": [2, 408], "131": [2, 408], "195": [2, 408], "196": [2, 408], "197": [2, 408] }, { "4": [2, 410], "5": [2, 410], "31": [2, 410], "32": [2, 410], "34": [2, 410], "35": [2, 410], "36": [2, 410], "37": [2, 410], "38": [2, 410], "39": [2, 410], "41": [2, 410], "42": [2, 410], "44": [2, 410], "45": [2, 410], "47": [2, 410], "48": [2, 410], "49": [2, 410], "50": [2, 410], "90": [2, 410], "96": [2, 410], "97": [2, 410], "118": [2, 410], "131": [2, 410], "195": [2, 410], "196": [2, 410], "197": [2, 410] }, { "4": [2, 411], "5": [2, 411], "31": [2, 411], "32": [2, 411], "34": [2, 411], "35": [2, 411], "36": [2, 411], "37": [2, 411], "38": [2, 411], "39": [2, 411], "41": [2, 411], "42": [2, 411], "44": [2, 411], "45": [2, 411], "47": [2, 411], "48": [2, 411], "49": [2, 411], "50": [2, 411], "90": [2, 411], "96": [2, 411], "97": [2, 411], "118": [2, 411], "131": [2, 411], "195": [2, 411], "196": [2, 411], "197": [2, 411] }, { "4": [2, 412], "5": [2, 412], "31": [2, 412], "32": [2, 412], "34": [2, 412], "35": [2, 412], "36": [2, 412], "37": [2, 412], "38": [2, 412], "39": [2, 412], "41": [2, 412], "42": [2, 412], "44": [2, 412], "45": [2, 412], "47": [2, 412], "48": [2, 412], "49": [2, 412], "50": [2, 412], "90": [2, 412], "96": [2, 412], "97": [2, 412], "118": [2, 412], "131": [2, 412], "195": [2, 412], "196": [2, 412], "197": [2, 412] }, { "4": [2, 413], "5": [2, 413], "31": [2, 413], "32": [2, 413], "34": [2, 413], "35": [2, 413], "36": [2, 413], "37": [2, 413], "38": [2, 413], "39": [2, 413], "41": [2, 413], "42": [2, 413], "44": [2, 413], "45": [2, 413], "47": [2, 413], "48": [2, 413], "49": [2, 413], "50": [2, 413], "90": [2, 413], "96": [2, 413], "97": [2, 413], "118": [2, 413], "131": [2, 413], "195": [2, 413], "196": [2, 413], "197": [2, 413] }, { "4": [2, 414], "5": [2, 414], "31": [2, 414], "32": [2, 414], "34": [2, 414], "35": [2, 414], "36": [2, 414], "37": [2, 414], "38": [2, 414], "39": [2, 414], "41": [2, 414], "42": [2, 414], "44": [2, 414], "45": [2, 414], "47": [2, 414], "48": [2, 414], "49": [2, 414], "50": [2, 414], "90": [2, 414], "96": [2, 414], "97": [2, 414], "118": [2, 414], "131": [2, 414], "195": [2, 414], "196": [2, 414], "197": [2, 414] }, { "4": [2, 415], "5": [2, 415], "31": [2, 415], "32": [2, 415], "34": [2, 415], "35": [2, 415], "36": [2, 415], "37": [2, 415], "38": [2, 415], "39": [2, 415], "41": [2, 415], "42": [2, 415], "44": [2, 415], "45": [2, 415], "47": [2, 415], "48": [2, 415], "49": [2, 415], "50": [2, 415], "90": [2, 415], "96": [2, 415], "97": [2, 415], "118": [2, 415], "131": [2, 415], "195": [2, 415], "196": [2, 415], "197": [2, 415] }, { "4": [2, 416], "5": [2, 416], "31": [2, 416], "32": [2, 416], "34": [2, 416], "35": [2, 416], "36": [2, 416], "37": [2, 416], "38": [2, 416], "39": [2, 416], "41": [2, 416], "42": [2, 416], "44": [2, 416], "45": [2, 416], "47": [2, 416], "48": [2, 416], "49": [2, 416], "50": [2, 416], "90": [2, 416], "96": [2, 416], "97": [2, 416], "118": [2, 416], "131": [2, 416], "195": [2, 416], "196": [2, 416], "197": [2, 416] }, { "4": [2, 417], "5": [2, 417], "31": [2, 417], "32": [2, 417], "34": [2, 417], "35": [2, 417], "36": [2, 417], "37": [2, 417], "38": [2, 417], "39": [2, 417], "41": [2, 417], "42": [2, 417], "44": [2, 417], "45": [2, 417], "47": [2, 417], "48": [2, 417], "49": [2, 417], "50": [2, 417], "90": [2, 417], "96": [2, 417], "97": [2, 417], "118": [2, 417], "131": [2, 417], "195": [2, 417], "196": [2, 417], "197": [2, 417] }, { "4": [2, 418], "5": [2, 418], "31": [2, 418], "32": [2, 418], "34": [2, 418], "35": [2, 418], "36": [2, 418], "37": [2, 418], "38": [2, 418], "39": [2, 418], "41": [2, 418], "42": [2, 418], "44": [2, 418], "45": [2, 418], "47": [2, 418], "48": [2, 418], "49": [2, 418], "50": [2, 418], "90": [2, 418], "96": [2, 418], "97": [2, 418], "118": [2, 418], "131": [2, 418], "195": [2, 418], "196": [2, 418], "197": [2, 418] }, { "4": [2, 419], "5": [2, 419], "31": [2, 419], "32": [2, 419], "34": [2, 419], "35": [2, 419], "36": [2, 419], "37": [2, 419], "38": [2, 419], "39": [2, 419], "41": [2, 419], "42": [2, 419], "44": [2, 419], "45": [2, 419], "47": [2, 419], "48": [2, 419], "49": [2, 419], "50": [2, 419], "90": [2, 419], "96": [2, 419], "97": [2, 419], "118": [2, 419], "131": [2, 419], "195": [2, 419], "196": [2, 419], "197": [2, 419] }, { "4": [2, 420], "5": [2, 420], "31": [2, 420], "32": [2, 420], "34": [2, 420], "35": [2, 420], "36": [2, 420], "37": [2, 420], "38": [2, 420], "39": [2, 420], "41": [2, 420], "42": [2, 420], "44": [2, 420], "45": [2, 420], "47": [2, 420], "48": [2, 420], "49": [2, 420], "50": [2, 420], "90": [2, 420], "96": [2, 420], "97": [2, 420], "118": [2, 420], "131": [2, 420], "195": [2, 420], "196": [2, 420], "197": [2, 420] }, { "33": [1, 815] }, { "33": [1, 816] }, { "33": [1, 817] }, { "33": [1, 818] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "4": [2, 405], "5": [2, 405], "31": [2, 405], "32": [2, 405], "34": [2, 405], "35": [2, 405], "36": [2, 405], "37": [2, 405], "38": [2, 405], "39": [2, 405], "41": [2, 405], "42": [2, 405], "44": [2, 405], "45": [2, 405], "47": [2, 405], "48": [2, 405], "49": [2, 405], "50": [2, 405], "90": [2, 405], "96": [2, 405], "97": [2, 405], "118": [2, 405], "131": [2, 405], "195": [2, 405], "196": [2, 405], "197": [2, 405] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "34": [2, 406], "38": [2, 406] }, { "4": [2, 306], "5": [2, 306], "31": [2, 306], "32": [2, 306], "34": [2, 306], "35": [2, 306], "36": [2, 306], "37": [2, 306], "38": [2, 306], "39": [2, 306], "41": [2, 306], "42": [2, 306], "44": [2, 306], "45": [2, 306], "47": [2, 306], "48": [2, 306], "49": [2, 306], "50": [2, 306], "90": [2, 306], "96": [2, 306], "97": [2, 306], "118": [2, 306], "131": [2, 306], "195": [2, 306], "196": [2, 306], "197": [2, 306] }, { "194": [1, 626] }, { "38": [1, 826], "196": [1, 825] }, { "38": [2, 310], "196": [2, 310] }, { "46": [1, 827] }, { "46": [1, 828] }, { "46": [1, 829] }, { "46": [1, 830] }, { "46": [1, 831] }, { "46": [1, 832] }, { "46": [1, 833] }, { "46": [1, 834] }, { "39": [1, 203], "74": [1, 204] }, { "4": [2, 378], "5": [2, 378], "31": [2, 378], "32": [2, 378], "34": [2, 378], "35": [2, 378], "36": [2, 378], "37": [2, 378], "38": [2, 378], "39": [2, 378], "41": [2, 378], "42": [2, 378], "44": [2, 378], "45": [2, 378], "47": [2, 378], "48": [2, 378], "49": [2, 378], "50": [2, 378], "90": [2, 378], "96": [2, 378], "97": [2, 378], "118": [2, 378], "131": [2, 378], "195": [2, 378], "196": [2, 378], "197": [2, 378] }, { "4": [2, 379], "5": [2, 379], "31": [2, 379], "32": [2, 379], "34": [2, 379], "35": [2, 379], "36": [2, 379], "37": [2, 379], "38": [2, 379], "39": [2, 379], "41": [2, 379], "42": [2, 379], "44": [2, 379], "45": [2, 379], "47": [2, 379], "48": [2, 379], "49": [2, 379], "50": [2, 379], "90": [2, 379], "96": [2, 379], "97": [2, 379], "118": [2, 379], "131": [2, 379], "195": [2, 379], "196": [2, 379], "197": [2, 379] }, { "34": [2, 375], "38": [2, 375], "39": [2, 375], "74": [2, 375], "84": [1, 192], "156": [1, 837] }, { "34": [2, 376], "38": [2, 376], "39": [2, 376], "74": [2, 376] }, { "4": [2, 381], "5": [2, 381], "31": [2, 381], "32": [2, 381], "34": [2, 381], "35": [2, 381], "36": [2, 381], "37": [2, 381], "38": [2, 381], "39": [2, 381], "41": [2, 381], "42": [2, 381], "44": [2, 381], "45": [2, 381], "47": [2, 381], "48": [2, 381], "49": [2, 381], "50": [2, 381], "90": [2, 381], "96": [2, 381], "97": [2, 381], "118": [2, 381], "131": [2, 381], "195": [2, 381], "196": [2, 381], "197": [2, 381] }, { "39": [1, 203], "74": [1, 204], "84": [1, 192], "156": [1, 837] }, { "39": [1, 203], "74": [1, 204] }, { "4": [2, 363], "5": [2, 363], "31": [2, 363], "32": [2, 363], "34": [2, 363], "35": [2, 363], "36": [2, 363], "37": [2, 363], "38": [2, 363], "39": [2, 363], "41": [2, 363], "42": [2, 363], "44": [2, 363], "45": [2, 363], "47": [2, 363], "48": [2, 363], "49": [2, 363], "50": [2, 363], "90": [2, 363], "96": [2, 363], "97": [2, 363], "118": [2, 363], "131": [2, 363], "195": [2, 363], "196": [2, 363], "197": [2, 363] }, { "4": [2, 391], "5": [2, 391], "31": [2, 391], "32": [2, 391], "34": [2, 391], "35": [2, 391], "36": [2, 391], "37": [2, 391], "38": [2, 391], "39": [2, 391], "41": [2, 391], "42": [2, 391], "44": [2, 391], "45": [2, 391], "47": [2, 391], "48": [2, 391], "49": [2, 391], "50": [2, 391], "90": [2, 391], "96": [2, 391], "97": [2, 391], "118": [2, 391], "131": [2, 391], "195": [2, 391], "196": [2, 391], "197": [2, 391] }, { "4": [2, 392], "5": [2, 392], "31": [2, 392], "32": [2, 392], "34": [2, 392], "35": [2, 392], "36": [2, 392], "37": [2, 392], "38": [2, 392], "39": [2, 392], "41": [2, 392], "42": [2, 392], "44": [2, 392], "45": [2, 392], "47": [2, 392], "48": [2, 392], "49": [2, 392], "50": [2, 392], "90": [2, 392], "96": [2, 392], "97": [2, 392], "118": [2, 392], "131": [2, 392], "195": [2, 392], "196": [2, 392], "197": [2, 392] }, { "4": [2, 178], "5": [2, 178], "34": [2, 178], "38": [2, 178], "44": [2, 178], "48": [2, 178], "96": [2, 178], "131": [2, 178], "196": [2, 178] }, { "4": [2, 181], "5": [2, 181], "34": [2, 181], "38": [2, 181], "44": [2, 181], "48": [2, 181], "89": [2, 181], "96": [2, 181], "131": [2, 181], "196": [2, 181] }, { "4": [2, 182], "5": [2, 182], "34": [2, 182], "38": [2, 182], "44": [2, 182], "48": [2, 182], "89": [2, 182], "96": [2, 182], "131": [2, 182], "196": [2, 182] }, { "4": [2, 183], "5": [2, 183], "34": [2, 183], "38": [2, 183], "44": [2, 183], "48": [2, 183], "89": [2, 183], "96": [2, 183], "131": [2, 183], "196": [2, 183] }, { "4": [2, 186], "5": [2, 186], "34": [2, 186], "38": [2, 186], "44": [2, 186], "48": [2, 186], "89": [2, 186], "96": [2, 186], "131": [2, 186], "196": [2, 186] }, { "1": [2, 153], "33": [2, 153], "55": [2, 153], "69": [2, 153], "89": [2, 153], "91": [2, 153], "92": [2, 153], "97": [2, 153], "98": [2, 153], "100": [2, 153], "101": [2, 153], "106": [2, 153], "108": [2, 153], "113": [2, 153], "114": [2, 153], "115": [2, 153], "118": [2, 153], "122": [2, 153], "123": [2, 153], "125": [2, 153], "127": [2, 153], "134": [2, 153], "135": [2, 153], "136": [2, 153], "145": [2, 153], "147": [2, 153], "148": [2, 153], "149": [2, 153], "150": [2, 153], "151": [2, 153], "153": [2, 153], "154": [2, 153], "155": [2, 153], "158": [2, 153], "167": [2, 153], "170": [2, 153], "172": [2, 153], "173": [2, 153], "175": [2, 153], "178": [2, 153], "179": [2, 153], "181": [2, 153], "196": [2, 153] }, { "33": [2, 62], "55": [2, 62], "69": [2, 62], "89": [2, 62], "92": [2, 62], "97": [2, 62], "98": [2, 62], "100": [2, 62], "101": [2, 62], "108": [2, 62], "113": [2, 62], "115": [2, 62], "118": [2, 62], "122": [2, 62], "123": [2, 62], "125": [2, 62], "127": [2, 62], "134": [2, 62], "135": [2, 62], "136": [2, 62], "145": [2, 62], "147": [2, 62], "148": [2, 62], "149": [2, 62], "150": [2, 62], "151": [2, 62], "153": [2, 62], "154": [2, 62], "155": [2, 62], "158": [2, 62], "167": [2, 62], "170": [2, 62], "172": [2, 62], "173": [2, 62], "175": [2, 62], "178": [2, 62], "179": [2, 62], "181": [2, 62], "196": [2, 62] }, { "33": [2, 71], "55": [2, 71], "69": [2, 71], "89": [2, 71], "92": [2, 71], "97": [2, 71], "98": [2, 71], "100": [2, 71], "101": [2, 71], "108": [2, 71], "113": [2, 71], "115": [2, 71], "118": [2, 71], "122": [2, 71], "123": [2, 71], "125": [2, 71], "127": [2, 71], "134": [2, 71], "135": [2, 71], "136": [2, 71], "145": [2, 71], "147": [2, 71], "148": [2, 71], "149": [2, 71], "150": [2, 71], "151": [2, 71], "153": [2, 71], "154": [2, 71], "155": [2, 71], "158": [2, 71], "167": [2, 71], "170": [2, 71], "172": [2, 71], "173": [2, 71], "175": [2, 71], "178": [2, 71], "179": [2, 71], "181": [2, 71], "196": [2, 71] }, { "44": [1, 842] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "136": [1, 80], "145": [1, 72], "147": [1, 60], "148": [1, 32], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "173": [1, 658], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78], "196": [1, 843] }, { "34": [2, 118], "38": [2, 118] }, { "1": [2, 126], "33": [2, 126], "55": [2, 126], "69": [2, 126], "89": [2, 126], "91": [2, 126], "92": [2, 126], "97": [2, 126], "98": [2, 126], "100": [2, 126], "101": [2, 126], "106": [2, 126], "108": [2, 126], "113": [2, 126], "114": [2, 126], "115": [2, 126], "118": [2, 126], "122": [2, 126], "123": [2, 126], "125": [2, 126], "127": [2, 126], "134": [2, 126], "135": [2, 126], "136": [2, 126], "145": [2, 126], "147": [2, 126], "148": [2, 126], "149": [2, 126], "150": [2, 126], "151": [2, 126], "153": [2, 126], "154": [2, 126], "155": [2, 126], "158": [2, 126], "167": [2, 126], "170": [2, 126], "172": [2, 126], "175": [2, 126], "178": [2, 126], "179": [2, 126], "181": [2, 126] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [1, 25], "5": [1, 24], "44": [1, 846] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [1, 25], "5": [1, 24], "44": [1, 849] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [1, 25], "5": [1, 24], "44": [1, 852] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78], "194": [1, 854] }, { "4": [1, 25], "5": [1, 24], "44": [1, 856] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "1": [2, 80], "33": [2, 80], "55": [2, 80], "69": [2, 80], "89": [2, 80], "91": [2, 80], "92": [2, 80], "97": [2, 80], "98": [2, 80], "100": [2, 80], "101": [2, 80], "106": [2, 80], "108": [2, 80], "113": [2, 80], "114": [2, 80], "115": [2, 80], "118": [2, 80], "122": [2, 80], "123": [2, 80], "125": [2, 80], "127": [2, 80], "134": [2, 80], "135": [2, 80], "136": [2, 80], "145": [2, 80], "147": [2, 80], "148": [2, 80], "149": [2, 80], "150": [2, 80], "151": [2, 80], "153": [2, 80], "154": [2, 80], "155": [2, 80], "158": [2, 80], "167": [2, 80], "170": [2, 80], "172": [2, 80], "175": [2, 80], "178": [2, 80], "179": [2, 80], "181": [2, 80] }, { "4": [2, 84], "5": [2, 84], "38": [2, 84], "194": [2, 84] }, { "55": [1, 858] }, { "69": [1, 394], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 399], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "92": [2, 99], "125": [2, 99], "139": [2, 99], "149": [2, 99], "154": [2, 99], "170": [2, 99], "196": [2, 99] }, { "44": [1, 860] }, { "64": [1, 864], "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "4": [1, 25], "5": [1, 24], "44": [1, 865] }, { "69": [1, 394], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 399], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "34": [1, 868], "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "140": [1, 873], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "163": [1, 874], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "69": [2, 89], "92": [2, 89], "97": [2, 89], "98": [2, 89], "100": [2, 89], "101": [2, 89], "108": [2, 89], "110": [2, 89], "111": [2, 89], "113": [2, 89], "115": [2, 89], "118": [2, 89], "119": [2, 89], "122": [2, 89], "123": [2, 89], "124": [2, 89], "125": [2, 89], "127": [2, 89], "134": [2, 89], "135": [2, 89], "141": [2, 89], "142": [2, 89], "145": [2, 89], "147": [2, 89], "149": [2, 89], "150": [2, 89], "151": [2, 89], "154": [2, 89], "155": [2, 89], "165": [2, 89], "167": [2, 89], "170": [2, 89], "172": [2, 89], "174": [2, 89], "176": [2, 89], "178": [2, 89], "179": [2, 89], "180": [2, 89], "181": [2, 89], "198": [2, 89], "199": [2, 89] }, { "69": [2, 90], "92": [2, 90], "97": [2, 90], "98": [2, 90], "100": [2, 90], "101": [2, 90], "108": [2, 90], "110": [2, 90], "111": [2, 90], "113": [2, 90], "115": [2, 90], "118": [2, 90], "119": [2, 90], "122": [2, 90], "123": [2, 90], "124": [2, 90], "125": [2, 90], "127": [2, 90], "134": [2, 90], "135": [2, 90], "141": [2, 90], "142": [2, 90], "145": [2, 90], "147": [2, 90], "149": [2, 90], "150": [2, 90], "151": [2, 90], "154": [2, 90], "155": [2, 90], "165": [2, 90], "167": [2, 90], "170": [2, 90], "172": [2, 90], "174": [2, 90], "176": [2, 90], "178": [2, 90], "179": [2, 90], "180": [2, 90], "181": [2, 90], "198": [2, 90], "199": [2, 90] }, { "126": [1, 177] }, { "89": [1, 877] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "186": [1, 879] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "6": [1, 153], "9": [1, 151], "12": [1, 197], "15": [1, 199], "18": [1, 193], "20": [1, 141], "21": [1, 139], "22": [1, 142], "23": [1, 196], "25": [1, 191], "27": [1, 140], "28": [1, 152], "33": [1, 248], "39": [1, 203], "55": [1, 179], "56": [1, 200], "59": [1, 195], "61": [1, 206], "63": [1, 182], "67": [1, 188], "68": [1, 184], "69": [1, 95], "70": [1, 198], "72": [1, 205], "74": [1, 204], "77": [1, 181], "78": [1, 183], "79": [1, 178], "80": [1, 189], "82": [1, 174], "84": [1, 192], "87": [1, 180], "88": [1, 185], "89": [1, 149], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 148], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 177], "127": [1, 52], "130": [1, 260], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "156": [1, 175], "159": [1, 186], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "186": [1, 259], "194": [1, 150], "198": [1, 109], "199": [1, 112] }, { "1": [2, 21], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "1": [2, 27], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "1": [2, 47] }, { "44": [2, 48] }, { "44": [2, 49] }, { "44": [2, 50] }, { "44": [2, 52], "48": [2, 52] }, { "44": [2, 54], "48": [2, 54], "195": [1, 270] }, { "44": [2, 55], "48": [2, 55], "195": [1, 270] }, { "4": [2, 208], "5": [2, 208], "32": [2, 208], "34": [2, 208], "38": [2, 208], "42": [2, 208], "44": [2, 208], "48": [2, 208], "90": [2, 208], "131": [2, 208], "195": [2, 208], "196": [2, 208] }, { "4": [2, 260], "5": [2, 260], "32": [2, 260], "34": [2, 260], "38": [2, 260], "42": [2, 260], "44": [2, 260], "48": [2, 260], "90": [2, 260], "96": [2, 260], "97": [2, 260], "118": [2, 260], "131": [2, 260], "195": [2, 260], "196": [2, 260] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "38": [2, 305], "196": [2, 305] }, { "52": [1, 720], "53": [1, 731], "54": [1, 721], "58": [1, 722], "62": [1, 723], "65": [1, 732], "66": [1, 734], "69": [1, 730], "71": [1, 724], "73": [1, 733], "75": [1, 725], "76": [1, 728], "81": [1, 729], "83": [1, 727], "86": [1, 726], "194": [1, 735] }, { "64": [1, 886] }, { "69": [1, 409], "140": [1, 411], "151": [1, 410] }, { "51": [1, 378], "52": [1, 379], "53": [1, 380], "54": [1, 381], "57": [1, 382], "58": [1, 383], "62": [1, 384], "65": [1, 385], "66": [1, 386], "69": [1, 394], "71": [1, 387], "73": [1, 388], "75": [1, 389], "76": [1, 390], "81": [1, 391], "83": [1, 392], "86": [1, 393], "91": [1, 324], "92": [1, 57], "93": [1, 372], "94": [1, 325], "95": [1, 326], "96": [1, 327], "97": [1, 74], "98": [1, 79], "99": [1, 339], "100": [1, 66], "101": [1, 67], "102": [1, 373], "103": [1, 340], "104": [1, 341], "105": [1, 342], "106": [1, 313], "107": [1, 374], "108": [1, 53], "109": [1, 343], "110": [1, 107], "111": [1, 108], "112": [1, 344], "113": [1, 399], "114": [1, 328], "115": [1, 68], "116": [1, 345], "117": [1, 346], "118": [1, 75], "119": [1, 101], "120": [1, 347], "121": [1, 375], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 314], "127": [1, 52], "128": [1, 348], "129": [1, 315], "130": [1, 323], "131": [1, 329], "132": [1, 349], "133": [1, 350], "134": [1, 71], "135": [1, 59], "136": [1, 330], "137": [1, 351], "138": [1, 352], "139": [1, 316], "140": [1, 317], "141": [1, 110], "142": [1, 111], "143": [1, 353], "144": [1, 354], "145": [1, 72], "146": [1, 376], "147": [1, 60], "148": [1, 318], "149": [1, 61], "150": [1, 77], "151": [1, 76], "152": [1, 331], "153": [1, 332], "154": [1, 62], "155": [1, 54], "156": [1, 319], "157": [1, 333], "158": [1, 334], "159": [1, 320], "160": [1, 321], "161": [1, 377], "162": [1, 336], "163": [1, 335], "164": [1, 355], "165": [1, 105], "166": [1, 356], "167": [1, 63], "168": [1, 357], "169": [1, 358], "170": [1, 64], "171": [1, 337], "172": [1, 55], "173": [1, 359], "174": [1, 102], "175": [1, 338], "176": [1, 103], "177": [1, 360], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "182": [1, 361], "183": [1, 362], "184": [1, 363], "185": [1, 364], "186": [1, 322], "187": [1, 365], "188": [1, 366], "189": [1, 367], "190": [1, 368], "191": [1, 369], "192": [1, 370], "193": [1, 371], "198": [1, 109], "199": [1, 112] }, { "196": [1, 890] }, { "38": [1, 891], "196": [2, 426] }, { "38": [2, 428], "196": [2, 428] }, { "42": [1, 619], "85": [1, 620] }, { "42": [1, 619], "85": [1, 620] }, { "34": [2, 308], "38": [2, 308] }, { "34": [2, 309], "38": [2, 309] }, { "100": [1, 743], "101": [1, 745], "115": [1, 749], "122": [1, 744], "123": [1, 746], "134": [1, 747], "145": [1, 748], "179": [1, 750] }, { "12": [1, 197], "23": [1, 196], "59": [1, 195], "156": [1, 896] }, { "12": [1, 197], "23": [1, 196], "59": [1, 195], "156": [1, 896] }, { "84": [1, 192], "156": [1, 837] }, { "84": [1, 192], "156": [1, 837] }, { "74": [1, 204] }, { "39": [1, 203], "61": [1, 206], "72": [1, 205], "74": [1, 204] }, { "126": [1, 177] }, { "130": [1, 905], "186": [1, 904] }, { "34": [1, 907], "38": [1, 906] }, { "34": [1, 908] }, { "86": [1, 307] }, { "34": [1, 909] }, { "38": [1, 910] }, { "4": [2, 356], "5": [2, 356], "31": [2, 356], "32": [2, 356], "34": [2, 356], "35": [2, 356], "36": [2, 356], "37": [2, 356], "38": [2, 356], "39": [2, 356], "41": [2, 356], "42": [2, 356], "44": [2, 356], "45": [2, 356], "47": [2, 356], "48": [2, 356], "49": [2, 356], "50": [2, 356], "90": [2, 356], "96": [2, 356], "97": [2, 356], "118": [2, 356], "131": [2, 356], "195": [2, 356], "196": [2, 356], "197": [2, 356] }, { "34": [1, 912], "38": [1, 911] }, { "33": [2, 73], "55": [2, 73], "69": [2, 73], "89": [2, 73], "92": [2, 73], "97": [2, 73], "98": [2, 73], "100": [2, 73], "101": [2, 73], "108": [2, 73], "113": [2, 73], "115": [2, 73], "118": [2, 73], "122": [2, 73], "123": [2, 73], "125": [2, 73], "127": [2, 73], "134": [2, 73], "135": [2, 73], "136": [2, 73], "145": [2, 73], "147": [2, 73], "148": [2, 73], "149": [2, 73], "150": [2, 73], "151": [2, 73], "153": [2, 73], "154": [2, 73], "155": [2, 73], "158": [2, 73], "167": [2, 73], "170": [2, 73], "172": [2, 73], "173": [2, 73], "175": [2, 73], "178": [2, 73], "179": [2, 73], "181": [2, 73], "196": [2, 73] }, { "1": [2, 154], "33": [2, 154], "55": [2, 154], "69": [2, 154], "89": [2, 154], "91": [2, 154], "92": [2, 154], "97": [2, 154], "98": [2, 154], "100": [2, 154], "101": [2, 154], "106": [2, 154], "108": [2, 154], "113": [2, 154], "114": [2, 154], "115": [2, 154], "118": [2, 154], "122": [2, 154], "123": [2, 154], "125": [2, 154], "127": [2, 154], "134": [2, 154], "135": [2, 154], "136": [2, 154], "145": [2, 154], "147": [2, 154], "148": [2, 154], "149": [2, 154], "150": [2, 154], "151": [2, 154], "153": [2, 154], "154": [2, 154], "155": [2, 154], "158": [2, 154], "167": [2, 154], "170": [2, 154], "172": [2, 154], "173": [2, 154], "175": [2, 154], "178": [2, 154], "179": [2, 154], "181": [2, 154], "196": [2, 154] }, { "4": [1, 25], "5": [1, 24], "44": [1, 914] }, { "4": [1, 25], "5": [1, 24], "44": [1, 915] }, { "92": [2, 136], "167": [2, 136], "170": [2, 136], "178": [2, 136], "196": [2, 136] }, { "4": [1, 25], "5": [1, 24], "44": [1, 917] }, { "4": [1, 25], "5": [1, 24], "44": [1, 918] }, { "92": [2, 140], "167": [2, 140], "170": [2, 140], "178": [2, 140], "196": [2, 140] }, { "4": [1, 25], "5": [1, 24], "44": [1, 920] }, { "4": [1, 25], "5": [1, 24], "44": [1, 921] }, { "92": [2, 144], "167": [2, 144], "170": [2, 144], "178": [2, 144], "196": [2, 144] }, { "4": [1, 25], "5": [1, 24], "44": [1, 923] }, { "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "4": [1, 25], "5": [1, 24], "44": [1, 925] }, { "92": [2, 148], "167": [2, 148], "170": [2, 148], "178": [2, 148], "196": [2, 148] }, { "196": [1, 926] }, { "33": [1, 224] }, { "33": [2, 435], "38": [2, 435], "135": [2, 435] }, { "92": [2, 100], "125": [2, 100], "139": [2, 100], "149": [2, 100], "154": [2, 100], "170": [2, 100], "196": [2, 100] }, { "4": [1, 25], "5": [1, 24], "38": [1, 930], "44": [1, 928] }, { "4": [2, 96], "5": [2, 96], "38": [2, 96], "44": [2, 96] }, { "4": [2, 94], "5": [2, 94], "38": [2, 94], "44": [2, 94] }, { "4": [2, 95], "5": [2, 95], "38": [2, 95], "44": [2, 95] }, { "92": [2, 103], "125": [2, 103], "139": [2, 103], "149": [2, 103], "154": [2, 103], "170": [2, 103], "196": [2, 103] }, { "44": [1, 931] }, { "33": [1, 792], "38": [1, 786] }, { "4": [2, 107], "5": [2, 107], "44": [2, 107] }, { "34": [1, 933], "38": [1, 934] }, { "34": [2, 109], "38": [2, 109] }, { "42": [1, 619], "85": [1, 620] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "162": [1, 938], "171": [1, 937] }, { "69": [2, 122], "92": [2, 122], "97": [2, 122], "98": [2, 122], "100": [2, 122], "101": [2, 122], "108": [2, 122], "110": [2, 122], "111": [2, 122], "113": [2, 122], "115": [2, 122], "118": [2, 122], "119": [2, 122], "122": [2, 122], "123": [2, 122], "124": [2, 122], "125": [2, 122], "127": [2, 122], "134": [2, 122], "135": [2, 122], "141": [2, 122], "142": [2, 122], "145": [2, 122], "147": [2, 122], "149": [2, 122], "150": [2, 122], "151": [2, 122], "154": [2, 122], "155": [2, 122], "165": [2, 122], "167": [2, 122], "170": [2, 122], "172": [2, 122], "174": [2, 122], "176": [2, 122], "178": [2, 122], "179": [2, 122], "180": [2, 122], "181": [2, 122], "198": [2, 122], "199": [2, 122] }, { "90": [1, 939] }, { "90": [1, 940] }, { "90": [1, 941], "194": [1, 150] }, { "90": [1, 944] }, { "90": [1, 945] }, { "90": [1, 946] }, { "90": [1, 947] }, { "90": [1, 948] }, { "1": [2, 25], "33": [1, 42], "55": [1, 49], "69": [1, 50], "89": [1, 38], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "113": [1, 45], "115": [1, 68], "118": [1, 75], "122": [1, 69], "123": [1, 70], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "153": [1, 46], "154": [1, 62], "155": [1, 54], "158": [1, 34], "167": [1, 63], "170": [1, 64], "172": [1, 55], "175": [1, 43], "178": [1, 65], "179": [1, 73], "181": [1, 78] }, { "34": [1, 949] }, { "34": [1, 950] }, { "34": [1, 951] }, { "34": [1, 952] }, { "34": [1, 953], "38": [1, 954] }, { "34": [2, 432], "38": [2, 432] }, { "4": [2, 425], "5": [2, 425], "31": [2, 425], "32": [2, 425], "34": [2, 425], "35": [2, 425], "36": [2, 425], "37": [2, 425], "38": [2, 425], "39": [2, 425], "41": [2, 425], "42": [2, 425], "44": [2, 425], "45": [2, 425], "47": [2, 425], "48": [2, 425], "49": [2, 425], "50": [2, 425], "90": [2, 425], "96": [2, 425], "97": [2, 425], "118": [2, 425], "131": [2, 425], "195": [2, 425], "196": [2, 425], "197": [2, 425] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "196": [2, 427], "198": [1, 109], "199": [1, 112] }, { "4": [1, 25], "5": [1, 24], "38": [2, 430], "196": [2, 430] }, { "34": [2, 407], "38": [2, 407] }, { "38": [2, 311], "196": [2, 311] }, { "38": [2, 312], "196": [2, 312] }, { "62": [1, 957] }, { "38": [2, 313], "196": [2, 313] }, { "38": [2, 314], "196": [2, 314] }, { "38": [2, 315], "196": [2, 315] }, { "38": [2, 316], "196": [2, 316] }, { "38": [2, 317], "196": [2, 317] }, { "69": [1, 409], "140": [1, 411], "151": [1, 410] }, { "38": [2, 318], "196": [2, 318] }, { "38": [2, 319], "196": [2, 319] }, { "38": [2, 320], "196": [2, 320] }, { "34": [1, 959], "39": [1, 203], "74": [1, 204] }, { "4": [2, 367], "5": [2, 367], "31": [2, 367], "32": [2, 367], "34": [2, 367], "35": [2, 367], "36": [2, 367], "37": [2, 367], "38": [2, 367], "39": [2, 367], "41": [2, 367], "42": [2, 367], "44": [2, 367], "45": [2, 367], "47": [2, 367], "48": [2, 367], "49": [2, 367], "50": [2, 367], "90": [2, 367], "96": [2, 367], "97": [2, 367], "118": [2, 367], "131": [2, 367], "195": [2, 367], "196": [2, 367], "197": [2, 367] }, { "4": [2, 380], "5": [2, 380], "31": [2, 380], "32": [2, 380], "34": [2, 380], "35": [2, 380], "36": [2, 380], "37": [2, 380], "38": [2, 380], "39": [2, 380], "41": [2, 380], "42": [2, 380], "44": [2, 380], "45": [2, 380], "47": [2, 380], "48": [2, 380], "49": [2, 380], "50": [2, 380], "90": [2, 380], "96": [2, 380], "97": [2, 380], "118": [2, 380], "131": [2, 380], "195": [2, 380], "196": [2, 380], "197": [2, 380] }, { "4": [2, 382], "5": [2, 382], "31": [2, 382], "32": [2, 382], "34": [2, 382], "35": [2, 382], "36": [2, 382], "37": [2, 382], "38": [2, 382], "39": [2, 382], "41": [2, 382], "42": [2, 382], "44": [2, 382], "45": [2, 382], "47": [2, 382], "48": [2, 382], "49": [2, 382], "50": [2, 382], "90": [2, 382], "96": [2, 382], "97": [2, 382], "118": [2, 382], "131": [2, 382], "195": [2, 382], "196": [2, 382], "197": [2, 382] }, { "39": [1, 203], "74": [1, 204] }, { "80": [1, 189] }, { "4": [2, 362], "5": [2, 362], "31": [2, 362], "32": [2, 362], "34": [2, 362], "35": [2, 362], "36": [2, 362], "37": [2, 362], "38": [2, 362], "39": [2, 362], "41": [2, 362], "42": [2, 362], "44": [2, 362], "45": [2, 362], "47": [2, 362], "48": [2, 362], "49": [2, 362], "50": [2, 362], "90": [2, 362], "96": [2, 362], "97": [2, 362], "118": [2, 362], "131": [2, 362], "195": [2, 362], "196": [2, 362], "197": [2, 362] }, { "4": [1, 25], "5": [1, 24], "44": [1, 962] }, { "92": [2, 134], "167": [2, 134], "170": [2, 134], "178": [2, 134], "196": [2, 134] }, { "92": [2, 135], "167": [2, 135], "170": [2, 135], "178": [2, 135], "196": [2, 135] }, { "4": [1, 25], "5": [1, 24], "44": [1, 963] }, { "92": [2, 138], "167": [2, 138], "170": [2, 138], "178": [2, 138], "196": [2, 138] }, { "92": [2, 139], "167": [2, 139], "170": [2, 139], "178": [2, 139], "196": [2, 139] }, { "4": [1, 25], "5": [1, 24], "44": [1, 964] }, { "92": [2, 142], "167": [2, 142], "170": [2, 142], "178": [2, 142], "196": [2, 142] }, { "92": [2, 143], "167": [2, 143], "170": [2, 143], "178": [2, 143], "196": [2, 143] }, { "4": [1, 25], "5": [1, 24], "44": [1, 965] }, { "92": [2, 146], "167": [2, 146], "170": [2, 146], "178": [2, 146], "196": [2, 146] }, { "196": [1, 966] }, { "92": [2, 147], "167": [2, 147], "170": [2, 147], "178": [2, 147], "196": [2, 147] }, { "4": [1, 25], "5": [1, 24], "44": [1, 968] }, { "44": [1, 969] }, { "92": [2, 101], "125": [2, 101], "139": [2, 101], "149": [2, 101], "154": [2, 101], "170": [2, 101], "196": [2, 101] }, { "44": [1, 970] }, { "64": [1, 864], "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "92": [2, 104], "125": [2, 104], "139": [2, 104], "149": [2, 104], "154": [2, 104], "170": [2, 104], "196": [2, 104] }, { "4": [1, 25], "5": [1, 24], "44": [1, 972] }, { "4": [2, 108], "5": [2, 108], "44": [2, 108] }, { "69": [1, 95], "92": [1, 57], "97": [1, 74], "98": [1, 79], "100": [1, 66], "101": [1, 67], "108": [1, 53], "110": [1, 107], "111": [1, 108], "113": [1, 100], "115": [1, 68], "118": [1, 75], "119": [1, 101], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "127": [1, 52], "134": [1, 71], "135": [1, 59], "140": [1, 873], "141": [1, 110], "142": [1, 111], "145": [1, 72], "147": [1, 60], "149": [1, 61], "150": [1, 77], "151": [1, 76], "154": [1, 62], "155": [1, 54], "163": [1, 874], "165": [1, 105], "167": [1, 63], "170": [1, 64], "172": [1, 55], "174": [1, 102], "176": [1, 103], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "198": [1, 109], "199": [1, 112] }, { "4": [1, 25], "5": [1, 24], "34": [2, 111], "38": [2, 111] }, { "42": [1, 619], "85": [1, 620] }, { "69": [2, 120], "92": [2, 120], "97": [2, 120], "98": [2, 120], "100": [2, 120], "101": [2, 120], "108": [2, 120], "110": [2, 120], "111": [2, 120], "113": [2, 120], "115": [2, 120], "118": [2, 120], "119": [2, 120], "122": [2, 120], "123": [2, 120], "124": [2, 120], "125": [2, 120], "127": [2, 120], "134": [2, 120], "135": [2, 120], "141": [2, 120], "142": [2, 120], "145": [2, 120], "147": [2, 120], "149": [2, 120], "150": [2, 120], "151": [2, 120], "154": [2, 120], "155": [2, 120], "165": [2, 120], "167": [2, 120], "170": [2, 120], "172": [2, 120], "174": [2, 120], "176": [2, 120], "178": [2, 120], "179": [2, 120], "180": [2, 120], "181": [2, 120], "198": [2, 120], "199": [2, 120] }, { "69": [2, 121], "92": [2, 121], "97": [2, 121], "98": [2, 121], "100": [2, 121], "101": [2, 121], "108": [2, 121], "110": [2, 121], "111": [2, 121], "113": [2, 121], "115": [2, 121], "118": [2, 121], "119": [2, 121], "122": [2, 121], "123": [2, 121], "124": [2, 121], "125": [2, 121], "127": [2, 121], "134": [2, 121], "135": [2, 121], "141": [2, 121], "142": [2, 121], "145": [2, 121], "147": [2, 121], "149": [2, 121], "150": [2, 121], "151": [2, 121], "154": [2, 121], "155": [2, 121], "165": [2, 121], "167": [2, 121], "170": [2, 121], "172": [2, 121], "174": [2, 121], "176": [2, 121], "178": [2, 121], "179": [2, 121], "180": [2, 121], "181": [2, 121], "198": [2, 121], "199": [2, 121] }, { "4": [1, 25], "5": [1, 24], "44": [1, 977] }, { "4": [2, 37], "5": [2, 37], "44": [2, 37] }, { "90": [2, 43] }, { "38": [1, 980], "90": [1, 979] }, { "38": [2, 45], "90": [2, 45] }, { "4": [2, 38], "5": [2, 38], "44": [2, 38] }, { "4": [2, 39], "5": [2, 39], "44": [2, 39] }, { "4": [2, 40], "5": [2, 40], "44": [2, 40] }, { "4": [2, 41], "5": [2, 41], "44": [2, 41] }, { "4": [2, 42], "5": [2, 42], "44": [2, 42] }, { "34": [1, 981] }, { "4": [2, 421], "5": [2, 421], "31": [2, 421], "32": [2, 421], "34": [2, 421], "35": [2, 421], "36": [2, 421], "37": [2, 421], "38": [2, 421], "39": [2, 421], "41": [2, 421], "42": [2, 421], "44": [2, 421], "45": [2, 421], "47": [2, 421], "48": [2, 421], "49": [2, 421], "50": [2, 421], "90": [2, 421], "96": [2, 421], "97": [2, 421], "118": [2, 421], "131": [2, 421], "195": [2, 421], "196": [2, 421], "197": [2, 421] }, { "4": [2, 422], "5": [2, 422], "31": [2, 422], "32": [2, 422], "34": [2, 422], "35": [2, 422], "36": [2, 422], "37": [2, 422], "38": [2, 422], "39": [2, 422], "41": [2, 422], "42": [2, 422], "44": [2, 422], "45": [2, 422], "47": [2, 422], "48": [2, 422], "49": [2, 422], "50": [2, 422], "90": [2, 422], "96": [2, 422], "97": [2, 422], "118": [2, 422], "131": [2, 422], "195": [2, 422], "196": [2, 422], "197": [2, 422] }, { "4": [2, 423], "5": [2, 423], "31": [2, 423], "32": [2, 423], "34": [2, 423], "35": [2, 423], "36": [2, 423], "37": [2, 423], "38": [2, 423], "39": [2, 423], "41": [2, 423], "42": [2, 423], "44": [2, 423], "45": [2, 423], "47": [2, 423], "48": [2, 423], "49": [2, 423], "50": [2, 423], "90": [2, 423], "96": [2, 423], "97": [2, 423], "118": [2, 423], "131": [2, 423], "195": [2, 423], "196": [2, 423], "197": [2, 423] }, { "4": [2, 424], "5": [2, 424], "31": [2, 424], "32": [2, 424], "34": [2, 424], "35": [2, 424], "36": [2, 424], "37": [2, 424], "38": [2, 424], "39": [2, 424], "41": [2, 424], "42": [2, 424], "44": [2, 424], "45": [2, 424], "47": [2, 424], "48": [2, 424], "49": [2, 424], "50": [2, 424], "90": [2, 424], "96": [2, 424], "97": [2, 424], "118": [2, 424], "131": [2, 424], "195": [2, 424], "196": [2, 424], "197": [2, 424] }, { "51": [1, 378], "52": [1, 379], "53": [1, 380], "54": [1, 381], "57": [1, 382], "58": [1, 383], "62": [1, 384], "65": [1, 385], "66": [1, 386], "69": [1, 394], "71": [1, 387], "73": [1, 388], "75": [1, 389], "76": [1, 390], "81": [1, 391], "83": [1, 392], "86": [1, 393], "91": [1, 324], "92": [1, 57], "93": [1, 372], "94": [1, 325], "95": [1, 326], "96": [1, 327], "97": [1, 74], "98": [1, 79], "99": [1, 339], "100": [1, 66], "101": [1, 67], "102": [1, 373], "103": [1, 340], "104": [1, 341], "105": [1, 342], "106": [1, 313], "107": [1, 374], "108": [1, 53], "109": [1, 343], "110": [1, 107], "111": [1, 108], "112": [1, 344], "113": [1, 399], "114": [1, 328], "115": [1, 68], "116": [1, 345], "117": [1, 346], "118": [1, 75], "119": [1, 101], "120": [1, 347], "121": [1, 375], "122": [1, 69], "123": [1, 70], "124": [1, 104], "125": [1, 58], "126": [1, 314], "127": [1, 52], "128": [1, 348], "129": [1, 315], "130": [1, 323], "131": [1, 329], "132": [1, 349], "133": [1, 350], "134": [1, 71], "135": [1, 59], "136": [1, 330], "137": [1, 351], "138": [1, 352], "139": [1, 316], "140": [1, 317], "141": [1, 110], "142": [1, 111], "143": [1, 353], "144": [1, 354], "145": [1, 72], "146": [1, 376], "147": [1, 60], "148": [1, 318], "149": [1, 61], "150": [1, 77], "151": [1, 76], "152": [1, 331], "153": [1, 332], "154": [1, 62], "155": [1, 54], "156": [1, 319], "157": [1, 333], "158": [1, 334], "159": [1, 320], "160": [1, 321], "161": [1, 377], "162": [1, 336], "163": [1, 335], "164": [1, 355], "165": [1, 105], "166": [1, 356], "167": [1, 63], "168": [1, 357], "169": [1, 358], "170": [1, 64], "171": [1, 337], "172": [1, 55], "173": [1, 359], "174": [1, 102], "175": [1, 338], "176": [1, 103], "177": [1, 360], "178": [1, 65], "179": [1, 73], "180": [1, 106], "181": [1, 78], "182": [1, 361], "183": [1, 362], "184": [1, 363], "185": [1, 364], "186": [1, 322], "187": [1, 365], "188": [1, 366], "189": [1, 367], "190": [1, 368], "191": [1, 369], "192": [1, 370], "193": [1, 371], "198": [1, 109], "199": [1, 112] }, { "38": [2, 429], "196": [2, 429] }, { "38": [2, 431], "196": [2, 431] }, { "33": [1, 983] }, { "34": [1, 984] }, { "4": [2, 366], "5": [2, 366], "31": [2, 366], "32": [2, 366], "34": [2, 366], "35": [2, 366], "36": [2, 366], "37": [2, 366], "38": [2, 366], "39": [2, 366], "41": [2, 366], "42": [2, 366], "44": [2, 366], "45": [2, 366], "47": [2, 366], "48": [2, 366], "49": [2, 366], "50": [2, 366], "90": [2, 366], "96": [2, 366], "97": [2, 366], "118": [2, 366], "131": [2, 366], "195": [2, 366], "196": [2, 366], "197": [2, 366] }, { "34": [1, 986], "38": [1, 985] }, { "34": [1, 987] }, { "92": [2, 133], "167": [2, 133], "170": [2, 133], "178": [2, 133], "196": [2, 133] }, { "92": [2, 137], "167": [2, 137], "170": [2, 137], "178": [2, 137], "196": [2, 137] }, { "92": [2, 141], "167": [2, 141], "170": [2, 141], "178": [2, 141], "196": [2, 141] }, { "92": [2, 145], "167": [2, 145], "170": [2, 145], "178": [2, 145], "196": [2, 145] }, { "4": [1, 25], "5": [1, 24], "44": [1, 989] }, { "4": [1, 25], "5": [1, 24], "44": [1, 990] }, { "92": [2, 152], "167": [2, 152], "170": [2, 152], "178": [2, 152], "196": [2, 152] }, { "92": [2, 98], "125": [2, 98], "139": [2, 98], "149": [2, 98], "154": [2, 98], "170": [2, 98], "196": [2, 98] }, { "92": [2, 102], "125": [2, 102], "139": [2, 102], "149": [2, 102], "154": [2, 102], "170": [2, 102], "196": [2, 102] }, { "4": [2, 97], "5": [2, 97], "38": [2, 97], "44": [2, 97] }, { "92": [2, 105], "125": [2, 105], "139": [2, 105], "149": [2, 105], "154": [2, 105], "170": [2, 105], "196": [2, 105] }, { "44": [1, 991] }, { "34": [2, 110], "38": [2, 110] }, { "34": [2, 113], "38": [2, 113] }, { "4": [1, 25], "5": [1, 24], "34": [2, 112], "38": [2, 112] }, { "1": [2, 32], "33": [2, 32], "55": [2, 32], "69": [2, 32], "89": [2, 32], "92": [2, 32], "97": [2, 32], "98": [2, 32], "100": [2, 32], "101": [2, 32], "108": [2, 32], "113": [2, 32], "115": [2, 32], "118": [2, 32], "122": [2, 32], "123": [2, 32], "125": [2, 32], "127": [2, 32], "134": [2, 32], "135": [2, 32], "145": [2, 32], "147": [2, 32], "149": [2, 32], "150": [2, 32], "151": [2, 32], "153": [2, 32], "154": [2, 32], "155": [2, 32], "158": [2, 32], "167": [2, 32], "170": [2, 32], "172": [2, 32], "175": [2, 32], "178": [2, 32], "179": [2, 32], "181": [2, 32] }, { "44": [1, 993] }, { "90": [2, 44] }, { "194": [1, 150] }, { "4": [2, 261], "5": [2, 261], "32": [2, 261], "34": [2, 261], "38": [2, 261], "42": [2, 261], "44": [2, 261], "48": [2, 261], "90": [2, 261], "96": [2, 261], "97": [2, 261], "118": [2, 261], "131": [2, 261], "195": [2, 261], "196": [2, 261] }, { "34": [2, 433], "38": [2, 433] }, { "38": [1, 633], "39": [1, 203], "74": [1, 204], "80": [1, 189], "126": [1, 177] }, { "4": [2, 365], "5": [2, 365], "31": [2, 365], "32": [2, 365], "34": [2, 365], "35": [2, 365], "36": [2, 365], "37": [2, 365], "38": [2, 365], "39": [2, 365], "41": [2, 365], "42": [2, 365], "44": [2, 365], "45": [2, 365], "47": [2, 365], "48": [2, 365], "49": [2, 365], "50": [2, 365], "90": [2, 365], "96": [2, 365], "97": [2, 365], "118": [2, 365], "131": [2, 365], "195": [2, 365], "196": [2, 365], "197": [2, 365] }, { "34": [1, 996], "39": [1, 203], "74": [1, 204] }, { "4": [2, 385], "5": [2, 385], "31": [2, 385], "32": [2, 385], "34": [2, 385], "35": [2, 385], "36": [2, 385], "37": [2, 385], "38": [2, 385], "39": [2, 385], "41": [2, 385], "42": [2, 385], "44": [2, 385], "45": [2, 385], "47": [2, 385], "48": [2, 385], "49": [2, 385], "50": [2, 385], "90": [2, 385], "96": [2, 385], "97": [2, 385], "118": [2, 385], "131": [2, 385], "195": [2, 385], "196": [2, 385], "197": [2, 385] }, { "4": [2, 361], "5": [2, 361], "31": [2, 361], "32": [2, 361], "34": [2, 361], "35": [2, 361], "36": [2, 361], "37": [2, 361], "38": [2, 361], "39": [2, 361], "41": [2, 361], "42": [2, 361], "44": [2, 361], "45": [2, 361], "47": [2, 361], "48": [2, 361], "49": [2, 361], "50": [2, 361], "90": [2, 361], "96": [2, 361], "97": [2, 361], "118": [2, 361], "131": [2, 361], "195": [2, 361], "196": [2, 361], "197": [2, 361] }, { "4": [1, 25], "5": [1, 24], "44": [1, 997] }, { "92": [2, 150], "167": [2, 150], "170": [2, 150], "178": [2, 150], "196": [2, 150] }, { "92": [2, 151], "167": [2, 151], "170": [2, 151], "178": [2, 151], "196": [2, 151] }, { "92": [2, 106], "125": [2, 106], "139": [2, 106], "149": [2, 106], "154": [2, 106], "170": [2, 106], "196": [2, 106] }, { "34": [2, 114], "38": [2, 114] }, { "1": [2, 33], "33": [2, 33], "55": [2, 33], "69": [2, 33], "89": [2, 33], "92": [2, 33], "97": [2, 33], "98": [2, 33], "100": [2, 33], "101": [2, 33], "108": [2, 33], "113": [2, 33], "115": [2, 33], "118": [2, 33], "122": [2, 33], "123": [2, 33], "125": [2, 33], "127": [2, 33], "134": [2, 33], "135": [2, 33], "145": [2, 33], "147": [2, 33], "149": [2, 33], "150": [2, 33], "151": [2, 33], "153": [2, 33], "154": [2, 33], "155": [2, 33], "158": [2, 33], "167": [2, 33], "170": [2, 33], "172": [2, 33], "175": [2, 33], "178": [2, 33], "179": [2, 33], "181": [2, 33] }, { "38": [2, 46], "90": [2, 46] }, { "34": [1, 998] }, { "4": [2, 384], "5": [2, 384], "31": [2, 384], "32": [2, 384], "34": [2, 384], "35": [2, 384], "36": [2, 384], "37": [2, 384], "38": [2, 384], "39": [2, 384], "41": [2, 384], "42": [2, 384], "44": [2, 384], "45": [2, 384], "47": [2, 384], "48": [2, 384], "49": [2, 384], "50": [2, 384], "90": [2, 384], "96": [2, 384], "97": [2, 384], "118": [2, 384], "131": [2, 384], "195": [2, 384], "196": [2, 384], "197": [2, 384] }, { "92": [2, 149], "167": [2, 149], "170": [2, 149], "178": [2, 149], "196": [2, 149] }, { "4": [2, 383], "5": [2, 383], "31": [2, 383], "32": [2, 383], "34": [2, 383], "35": [2, 383], "36": [2, 383], "37": [2, 383], "38": [2, 383], "39": [2, 383], "41": [2, 383], "42": [2, 383], "44": [2, 383], "45": [2, 383], "47": [2, 383], "48": [2, 383], "49": [2, 383], "50": [2, 383], "90": [2, 383], "96": [2, 383], "97": [2, 383], "118": [2, 383], "131": [2, 383], "195": [2, 383], "196": [2, 383], "197": [2, 383] }];
const SEMANTIC_ACTION = [
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return "$yes"; }),
    (($, $0) => { return "$no"; }),
    (($, $0) => { return "$failed"; }),
    (($, $0) => { return "$train"; }),
    (($, $0) => { return "$debug"; }),
    (($, $0) => { return "$nevermind"; }),
    (($, $0) => { return "$stop"; }),
    (($, $0) => { return "$help"; }),
    (($, $0) => { return "$wakeup"; }),
    (($, special, $0) => new Ast.ControlCommand($.location, new Ast.ControlIntent.Special($.location, special.substring(1)))),
    (($, $0, $1, num, $2, $3) => new Ast.ControlCommand($.location, new Ast.ControlIntent.Choice($.location, num))),
    (($, $0, $1, v, $2, $3) => new Ast.ControlCommand($.location, new Ast.ControlIntent.Answer($.location, v))),
    (($, stmts) => makeInput($, stmts, {})),
    (($, annot, stmts) => makeInput($, stmts, annot)),
    (($, name) => {
        const lastDot = name.value.lastIndexOf('.');
        const className = name.value.substring(0, lastDot);
        const functionName = name.value.substring(lastDot + 1);
        // we check that the function name is a valid identifier because the lexer will greedly parse through things
        // that are not identifiers but are allowed in the class name
        if (!isIdentifier(functionName))
            $.error(`${name.value} is not a valid identifier`); //`
        return { kind: className, name: functionName };
    }),
    (($, kind, $0, name) => {
        return { kind: kind.value, name };
    }),
    (($, $0, dialogueAct, $1, history) => {
        return new Ast.DialogueState($.location, dialogueAct.kind, dialogueAct.name, null, history);
    }),
    (($, $0, dialogueAct, $1, pnames, $2, $3, history) => {
        return new Ast.DialogueState($.location, dialogueAct.kind, dialogueAct.name, pnames.map((p) => p instanceof Ast.VarRefValue ? p.name : p), history);
    }),
    (($, $0, dialogueAct, $1) => {
        return new Ast.DialogueState($.location, dialogueAct.kind, dialogueAct.name, null, []);
    }),
    (($, $0, dialogueAct, $1, pnames, $2, $3) => {
        return new Ast.DialogueState($.location, dialogueAct.kind, dialogueAct.name, pnames.map((p) => p instanceof Ast.VarRefValue ? p.name : p), []);
    }),
    (($, $0, dialogueAct, annot, $1, history) => {
        return new Ast.DialogueState($.location, dialogueAct.kind, dialogueAct.name, null, history, annot);
    }),
    (($, $0, dialogueAct, $1, pnames, $2, annot, $3, history) => {
        return new Ast.DialogueState($.location, dialogueAct.kind, dialogueAct.name, pnames.map((p) => p instanceof Ast.VarRefValue ? p.name : p), history, annot);
    }),
    (($, $0, dialogueAct, annot, $1) => {
        return new Ast.DialogueState($.location, dialogueAct.kind, dialogueAct.name, null, [], annot);
    }),
    (($, $0, dialogueAct, $1, pnames, $2, annot, $3) => {
        return new Ast.DialogueState($.location, dialogueAct.kind, dialogueAct.name, pnames.map((p) => p instanceof Ast.VarRefValue ? p.name : p), [], annot);
    }),
    (($, item) => [item]),
    (($, list, item) => {
        list.push(item);
        return list;
    }),
    (($, rule, $0) => new Ast.DialogueHistoryItem($.location, rule, null, 'accepted')),
    (($, rule, results, $0) => new Ast.DialogueHistoryItem($.location, rule, results[0], 'confirmed', results[1])),
    (($, rule, $0, $1, $2, venum, $3, $4) => new Ast.DialogueHistoryItem($.location, rule, null, venum)),
    (($, rule, $0, $1, $2, venum, $3, annot, $4) => new Ast.DialogueHistoryItem($.location, rule, null, venum, checkDialogueHistoryAnnotationSpec($, annot))),
    (($, ann) => [ann]),
    (($, list, ann) => list.concat([ann])),
    (($, list) => {
        let results, _error, more, count;
        const others = { nl: {}, impl: {} };
        for (const annot of list) {
            if (annot.type === 'impl') {
                switch (annot.key) {
                    case 'results':
                        if (results !== undefined)
                            return $.error(`Duplicate history annotation #[results]`);
                        results = annot.value;
                        break;
                    case 'error':
                        if (_error !== undefined)
                            return $.error(`Duplicate history annotation #[error]`);
                        _error = annot.value;
                        break;
                    case 'more':
                        if (more !== undefined)
                            return $.error(`Duplicate history annotation #[more]`);
                        more = annot.value;
                        break;
                    case 'count':
                        if (count !== undefined)
                            return $.error(`Duplicate history annotation #[count]`);
                        count = annot.value;
                        break;
                    default:
                        others.impl[annot.key] = annot.value;
                }
            }
            else {
                others.nl[annot.key] = annot.value.toJS();
            }
        }
        if (results === undefined)
            return $.error(`Missing history annotation #[results]`);
        if (count === undefined)
            count = new Ast.Value.Number(results.length);
        return [new Ast.DialogueHistoryResultList($.location, results, count, more, _error), others];
    }),
    (($, $0, $1, $2, results, $3) => ({ key: 'results', value: results, type: 'impl' })),
    (($, $0, $1, $2, count, $3) => ({ key: 'count', value: count, type: 'impl' })),
    (($, $0, $1, $2, $3, $4) => ({ key: 'more', value: true, type: 'impl' })),
    (($, $0, $1, $2, error, $3) => ({ key: 'error', value: error, type: 'impl' })),
    (($, $0, key, $1, value, $2) => ({ key, value, type: 'impl' })),
    (($, $0, key, $1, value, $2) => ({ key, value, type: 'nl' })),
    (($, $0, $1) => []),
    (($, $0, list, $1) => list),
    (($, value) => [new Ast.DialogueHistoryResultItem($.location, value.value)]),
    (($, list, $0, value) => {
        list.push(new Ast.DialogueHistoryResultItem(null, value.value));
        return list;
    }),
    (($, $0, $1, user, $2, policy, $3, $4) => {
        const clone = policy.clone();
        clone.principal = user;
        return clone;
    }),
    (($, $0, $1, action) => {
        return new Ast.PermissionRule($.location, Ast.BooleanExpression.True, Ast.PermissionFunction.Builtin, action);
    }),
    (($, query, $0, $1) => {
        return new Ast.PermissionRule($.location, Ast.BooleanExpression.True, query, Ast.PermissionFunction.Builtin);
    }),
    (($, query, $0, action) => {
        return new Ast.PermissionRule($.location, Ast.BooleanExpression.True, query, action);
    }),
    (($, $0) => Ast.PermissionFunction.Star),
    (($, klass, $0, $1) => new Ast.PermissionFunction.ClassStar($.location, klass.value)),
    (($, fn) => new Ast.PermissionFunction.Specified($.location, fn.kind, fn.name, Ast.BooleanExpression.True, null)),
    (($, fn, $0, filter) => new Ast.PermissionFunction.Specified($.location, fn.kind, fn.name, filter, null)),
    (($, fn, $0, filter) => new Ast.PermissionFunction.Specified($.location, fn.kind, fn.name, filter, null)),
    (($, list) => splitAnnotations(list)),
    (($, $0, name, $1, value, $2) => ({ type: 'nl', name, value })),
    (($, $0, name, $1, value, $2) => ({ type: 'impl', name, value })),
    (($, annotation) => [annotation]),
    (($, list, annotation) => list.concat([annotation])),
    (($, stmt) => [stmt]),
    (($, list, stmt) => {
        list.push(stmt);
        return list;
    }),
    (($, stmt) => [stmt]),
    (($, list, stmt) => list.concat([stmt])),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, expr, $0) => expr),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, expr, $0) => expr),
    (($, $0) => { return $0; }),
    (($, $0, expr, $1) => new Ast.ReturnStatement(null, expr)),
    (($, $0, name, $1, expr, $2) => new Ast.Assignment($.location, name, expr, null)),
    (($, $0) => false),
    (($, $0, $1) => true),
    (($, is_abstract, name, class_body) => makeClassDef($, name.value, null, class_body, {}, is_abstract)),
    (($, is_abstract, name, $0, _extends, class_body) => makeClassDef($, name.value, _extends, class_body, {}, is_abstract)),
    (($, is_abstract, name, annot, class_body) => makeClassDef($, name.value, null, class_body, annot, is_abstract)),
    (($, is_abstract, name, $0, _extends, annot, class_body) => makeClassDef($, name.value, _extends, class_body, annot, is_abstract)),
    (($, $0, $1) => []),
    (($, $0, stmts, $1) => stmts),
    (($, name) => [name.value]),
    (($, list, $0, name) => list.concat([name.value])),
    (($, stmt) => [stmt]),
    (($, list, stmt) => list.concat([stmt])),
    (($, $0) => ['query', false, false]),
    (($, $0, $1) => ['query', false, true]),
    (($, $0, $1, $2) => ['query', true, true]),
    (($, $0, $1, $2) => ['query', true, true]),
    (($, $0, $1) => ['query', true, false]),
    (($, $0) => ['action', false, false]),
    (($, $0, $1) => ['action', true, false]),
    (($, $0) => { return $0; }),
    (($, entity) => entity.value),
    (($, entity_name) => [entity_name]),
    (($, entity_name_list, $0, entity_name) => entity_name_list.concat([entity_name])),
    (($, $0, facets, $1, name, in_param_list, $2) => new Ast.MixinImportStmt($.location, facets, name.value, in_param_list)),
    (($, $0, name, $1) => new Ast.EntityDef($.location, name, [], {})),
    (($, $0, name, annotations, $1) => new Ast.EntityDef($.location, name, [], annotations)),
    (($, $0, name, $1, _extends, $2) => new Ast.EntityDef($.location, name, _extends, {})),
    (($, $0, name, $1, _extends, annotations, $2) => new Ast.EntityDef($.location, name, _extends, annotations)),
    (($, fntype, name, args, $0) => {
        const [functionType, is_list, is_monitorable] = fntype;
        return new Ast.FunctionDef($.location, functionType, null, name, [], { is_list, is_monitorable }, args);
    }),
    (($, fntype, name, args, annot, $0) => {
        const [functionType, is_list, is_monitorable] = fntype;
        return new Ast.FunctionDef($.location, functionType, null, name, [], { is_list, is_monitorable }, args, annot);
    }),
    (($, fntype, name, $0, extends_, args, $1) => {
        const [functionType, is_list, is_monitorable] = fntype;
        return new Ast.FunctionDef($.location, functionType, null, name, extends_, { is_list, is_monitorable }, args);
    }),
    (($, fntype, name, $0, extends_, args, annot, $1) => {
        const [functionType, is_list, is_monitorable] = fntype;
        return new Ast.FunctionDef($.location, functionType, null, name, extends_, { is_list, is_monitorable }, args, annot);
    }),
    (($, $0, $1) => []),
    (($, $0, decl_param_list_nonempty, $1) => decl_param_list_nonempty),
    (($, decl_param) => [decl_param]),
    (($, decl_param_list_nonempty, $0, decl_param) => decl_param_list_nonempty.concat([decl_param])),
    (($, name, type) => new Ast.ArgumentDef($.location, Ast.ArgDirection.IN_REQ, name, type)),
    (($, arg_direction, name, type) => new Ast.ArgumentDef($.location, arg_direction, name, type)),
    (($, name, type, annot) => new Ast.ArgumentDef($.location, Ast.ArgDirection.IN_REQ, name, type, annot)),
    (($, arg_direction, name, type, annot) => new Ast.ArgumentDef($.location, arg_direction, name, type, annot)),
    (($, $0, $1) => ({})),
    (($, $0, list, $1) => {
        const typeMap = {};
        for (const [name, type] of list)
            typeMap[name] = type;
        return typeMap;
    }),
    (($, p) => [p]),
    (($, list, $0, p) => {
        list.push(p);
        return list;
    }),
    (($, name, type) => [name, type]),
    (($, $0, $1) => Ast.ArgDirection.IN_REQ),
    (($, $0, $1) => Ast.ArgDirection.IN_OPT),
    (($, $0) => Ast.ArgDirection.OUT),
    (($, $0, name, dataset_body) => new Ast.Dataset($.location, name.value, dataset_body, {})),
    (($, $0, name, annot, dataset_body) => new Ast.Dataset($.location, name.value, dataset_body, annot)),
    (($, $0, name, $1, lang, dataset_body) => new Ast.Dataset($.location, name.value, dataset_body, { impl: { language: new Ast.Value.String(lang) } })),
    (($, $0, name, $1, lang, annot, dataset_body) => {
        annot.impl.language = new Ast.Value.String(lang);
        return new Ast.Dataset($.location, name.value, dataset_body, annot);
    }),
    (($, $0, $1) => []),
    (($, $0, example_list, $1) => example_list),
    (($, example) => [example]),
    (($, example_list, example) => example_list.concat([example])),
    (($, $0) => { return "="; }),
    (($, $0) => { return ":="; }),
    (($, $0, params, assignment_compat, expr, annot, $1) => {
        const [id, utterances, preprocessed, other] = checkExampleAnnotations($, annot);
        return new Ast.Example($.location, id, 'stream', params, expr, utterances, preprocessed, other);
    }),
    (($, $0, params, assignment_compat, expr, $1) => new Ast.Example($.location, -1, 'stream', params, expr, [], [], {})),
    (($, $0, assignment_compat, expr, annot, $1) => {
        const [id, utterances, preprocessed, other] = checkExampleAnnotations($, annot);
        return new Ast.Example($.location, id, 'stream', {}, expr, utterances, preprocessed, other);
    }),
    (($, $0, assignment_compat, expr, $1) => new Ast.Example($.location, -1, 'stream', {}, expr, [], [], {})),
    (($, $0, params, assignment_compat, expr, annot, $1) => {
        const [id, utterances, preprocessed, other] = checkExampleAnnotations($, annot);
        return new Ast.Example($.location, id, 'query', params, expr, utterances, preprocessed, other);
    }),
    (($, $0, params, assignment_compat, expr, $1) => new Ast.Example($.location, -1, 'query', params, expr, [], [], {})),
    (($, $0, assignment_compat, expr, annot, $1) => {
        const [id, utterances, preprocessed, other] = checkExampleAnnotations($, annot);
        return new Ast.Example($.location, id, 'query', {}, expr, utterances, preprocessed, other);
    }),
    (($, $0, assignment_compat, expr, $1) => new Ast.Example($.location, -1, 'query', {}, expr, [], [], {})),
    (($, $0, params, assignment_compat, expr, annot, $1) => {
        const [id, utterances, preprocessed, other] = checkExampleAnnotations($, annot);
        return new Ast.Example($.location, id, 'action', params, expr, utterances, preprocessed, other);
    }),
    (($, $0, params, assignment_compat, expr, $1) => new Ast.Example($.location, -1, 'action', params, expr, [], [], {})),
    (($, $0, assignment_compat, expr, annot, $1) => {
        const [id, utterances, preprocessed, other] = checkExampleAnnotations($, annot);
        return new Ast.Example($.location, id, 'action', {}, expr, utterances, preprocessed, other);
    }),
    (($, $0, assignment_compat, expr, $1) => new Ast.Example($.location, -1, 'action', {}, expr, [], [], {})),
    (($, $0, params, assignment_compat, expr, annot, $1) => {
        const [id, utterances, preprocessed, other] = checkExampleAnnotations($, annot);
        return new Ast.Example($.location, id, 'program', params, expr, utterances, preprocessed, other);
    }),
    (($, $0, params, assignment_compat, expr, $1) => new Ast.Example($.location, -1, 'program', params, expr, [], [], {})),
    (($, $0, assignment_compat, expr, annot, $1) => {
        const [id, utterances, preprocessed, other] = checkExampleAnnotations($, annot);
        return new Ast.Example($.location, id, 'program', {}, expr, utterances, preprocessed, other);
    }),
    (($, $0, assignment_compat, expr, $1) => new Ast.Example($.location, -1, 'program', {}, expr, [], [], {})),
    (($, $0, params, assignment_compat, $1, expr, $2, annot, $3) => {
        const [id, utterances, preprocessed, other] = checkExampleAnnotations($, annot);
        return new Ast.Example($.location, id, 'program', params, expr, utterances, preprocessed, other);
    }),
    (($, $0, params, assignment_compat, $1, expr, $2, $3) => new Ast.Example($.location, -1, 'program', params, expr, [], [], {})),
    (($, $0, assignment_compat, $1, expr, $2, annot, $3) => {
        const [id, utterances, preprocessed, other] = checkExampleAnnotations($, annot);
        return new Ast.Example($.location, id, 'program', {}, expr, utterances, preprocessed, other);
    }),
    (($, $0, assignment_compat, $1, expr, $2, $3) => new Ast.Example($.location, -1, 'program', {}, expr, [], [], {})),
    (($, $0, name, params, $1, stmts, $2) => {
        const [declarations, executable] = splitStatements(stmts);
        return new Ast.FunctionDeclaration($.location, name, params, declarations, executable, {}, null);
    }),
    (($, $0, name, params, annot, $1, stmts, $2) => {
        const [declarations, executable] = splitStatements(stmts);
        return new Ast.FunctionDeclaration($.location, name, params, declarations, executable, annot, null);
    }),
    (($, expr) => new Ast.ExpressionStatement($.location, expr)),
    (($, $0, $1, expr) => new Ast.ExpressionStatement($.location, expr)),
    (($, list) => new Ast.ChainExpression($.location, list, null)),
    (($, list) => new Ast.ChainExpression($.location, list, null)),
    (($, list, $0, $1) => new Ast.ChainExpression($.location, list, null)),
    (($, expr) => [expr]),
    (($, list, $0, expr) => {
        list.push(expr);
        return list;
    }),
    (($, $0) => { return $0; }),
    (($, $0, boolean_expression, $1, $2, expr) => {
        return new Ast.BooleanQuestionExpression($.location, expr, boolean_expression, null);
    }),
    (($, $0, args, $1, $2, expr) => {
        const names = [], computations = [], aliases = [];
        for (const [value, alias] of args) {
            if (value instanceof Ast.VarRefValue &&
                (alias === null || alias === value.name)) {
                names.push(value.name);
            }
            else {
                computations.push(value);
                aliases.push(alias);
            }
        }
        return new Ast.ProjectionExpression($.location, expr, names, computations, aliases, null);
    }),
    (($, $0) => [[new Ast.VarRefValue('*'), null]]),
    (($, $0, $1, list) => {
        const arr = [[new Ast.VarRefValue('*'), null]];
        return arr.concat(list);
    }),
    (($, $0) => { return $0; }),
    (($, p) => [p]),
    (($, list, $0, p) => {
        list.push(p);
        return list;
    }),
    (($, v) => [v, null]),
    (($, v, $0, alias) => [v, alias]),
    (($, $0) => { return $0; }),
    (($, expr, $0, filter) => new Ast.FilterExpression($.location, expr, filter, null)),
    (($, expr, $0, filter) => new Ast.FilterExpression($.location, expr, filter, null)),
    (($, $0) => { return $0; }),
    (($, expr, $0, name) => new Ast.AliasExpression($.location, expr, name, null)),
    (($, $0) => { return $0; }),
    (($, expr, $0, begin, $1, count, $2) => new Ast.SliceExpression($.location, expr, begin, count, null)),
    (($, expr, $0, indices, $1) => new Ast.IndexExpression($.location, expr, indices, null)),
    (($, $0, expr, $1) => expr),
    (($, $0, $1, spec, $2, expr, $3) => new Ast.SortExpression($.location, expr, spec[0], spec[1], null)),
    (($, op, $0, name, $1, expr, $2) => new Ast.AggregationExpression($.location, expr, name, op, null)),
    (($, $0, $1, name, $2, expr, $3) => new Ast.AggregationExpression($.location, expr, name, 'count', null)),
    (($, $0, $1, expr, $2) => new Ast.AggregationExpression($.location, expr, '*', 'count', null)),
    (($, $0, $1, expr, $2) => new Ast.MonitorExpression($.location, expr, null, null)),
    (($, $0, $1, args, $2, expr, $3) => new Ast.MonitorExpression($.location, expr, args, null)),
    (($, name, in_params) => new Ast.FunctionCallExpression($.location, name, in_params, null)),
    (($, call) => new Ast.InvocationExpression($.location, call, null)),
    (($, name, in_params) => {
        const lastDot = name.value.lastIndexOf('.');
        const className = name.value.substring(0, lastDot);
        const functionName = name.value.substring(lastDot + 1);
        // we check that the function name is a valid identifier because the lexer will greedly parse through things
        // that are not identifiers but are allowed in the class name
        if (!isIdentifier(functionName))
            $.error(`${name.value} is not a valid function name`); //`
        return new Ast.Invocation($.location, new Ast.DeviceSelector($.location, className, null, null), functionName, in_params, null);
    }),
    (($, className, $0, functionName, in_params) => new Ast.Invocation($.location, new Ast.DeviceSelector($.location, className.value, null, null), functionName, in_params, null)),
    (($, className, device_params, $0, functionName, in_params) => {
        let id = null, all = false;
        const other = [];
        for (const attr of device_params) {
            if (attr.name === 'id') {
                id = String(attr.value.toJS());
                const value = attr.value;
                if (value instanceof Ast.EntityValue && value.display)
                    other.push(new Ast.InputParam($.location, 'name', new Ast.Value.String(value.display)));
            }
            else if (attr.name === 'all') {
                all = !!attr.value.toJS();
            }
            else {
                other.push(attr);
            }
        }
        const selector = new Ast.DeviceSelector($.location, className.value, id, null, other, all);
        return new Ast.Invocation($.location, selector, functionName, in_params, null);
    }),
    (($, $0, $1) => []),
    (($, $0, in_param_list_nonempty, $1) => in_param_list_nonempty),
    (($, in_param) => [in_param]),
    (($, in_param_list_nonempty, $0, in_param) => in_param_list_nonempty.concat([in_param])),
    (($, $0, $1, value) => new Ast.InputParam($.location, 'all', value)),
    (($, name, $0, value) => new Ast.InputParam($.location, name, value)),
    (($, value, $0) => [value, 'asc']),
    (($, value, $0) => [value, 'desc']),
    (($, $0) => { return $0; }),
    (($, f1, $0, f2) => new Ast.BooleanExpression.Or($.location, [f1, f2])),
    (($, $0) => { return $0; }),
    (($, f1, $0, f2) => new Ast.BooleanExpression.And($.location, [f1, f2])),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0, f) => new Ast.BooleanExpression.Not($.location, f)),
    (($, lhs, op, rhs) => {
        if (lhs instanceof Ast.VarRefValue)
            return new Ast.BooleanExpression.Atom($.location, lhs.name, op, rhs);
        else
            return new Ast.BooleanExpression.Compute($.location, lhs, op, rhs);
    }),
    (($, lhs, op, $0, $1, rhs, $2) => new Ast.BooleanExpression.ComparisonSubquery($.location, lhs, op, rhs)),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0, f) => new Ast.BooleanExpression.Not($.location, f)),
    (($, lhs, op, rhs) => {
        if (lhs instanceof Ast.VarRefValue)
            return new Ast.BooleanExpression.Atom($.location, lhs.name, op, rhs);
        else
            return new Ast.BooleanExpression.Compute($.location, lhs, op, rhs);
    }),
    (($, field, $0, value) => new Ast.Value.ArrayField(value, field)),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, value, $0, filter) => new Ast.Value.Filter(value, filter)),
    (($, $0) => { return $0; }),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('+', [lhs, rhs])),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('-', [lhs, rhs])),
    (($, $0) => { return $0; }),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('*', [lhs, rhs])),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('/', [lhs, rhs])),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('%', [lhs, rhs])),
    (($, $0) => { return $0; }),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('**', [lhs, rhs])),
    (($, $0, value, $1) => value),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return "=="; }),
    (($, $0) => { return ">="; }),
    (($, $0) => { return "<="; }),
    (($, $0) => { return "=~"; }),
    (($, $0) => { return "~="; }),
    (($, $0) => { return "starts_with"; }),
    (($, $0) => { return "ends_with"; }),
    (($, $0) => { return "prefix_of"; }),
    (($, $0) => { return "suffix_of"; }),
    (($, $0) => { return "contains"; }),
    (($, $0) => { return "contains~"; }),
    (($, $0) => { return "~contains"; }),
    (($, $0) => { return "in_array"; }),
    (($, $0) => { return "in_array~"; }),
    (($, $0) => { return "~in_array"; }),
    (($, $0) => { return "min"; }),
    (($, $0) => { return "max"; }),
    (($, $0) => { return "sum"; }),
    (($, $0) => { return "avg"; }),
    (($, $0) => 'distance'),
    (($, $0) => 'set_time'),
    (($, $0) => Ast.BooleanExpression.True),
    (($, $0) => Ast.BooleanExpression.False),
    (($, $0, filter, $1) => filter),
    (($, $0, $1, pname, $2) => new Ast.BooleanExpression.DontCare($.location, pname)),
    (($, op, $0, lhs, $1, rhs, $2) => {
        if (lhs instanceof Ast.VarRefValue)
            return new Ast.BooleanExpression.Atom($.location, lhs.name, op, rhs);
        else
            return new Ast.BooleanExpression.Compute($.location, lhs, op, rhs);
    }),
    (($, op, $0, lhs, $1, $2, $3, rhs, $4, $5) => new Ast.BooleanExpression.ComparisonSubquery($.location, lhs, op, rhs)),
    (($, $0, $1, subquery, $2) => new Ast.BooleanExpression.ExistentialSubquery($.location, subquery)),
    (($, $0, value, $1) => value),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, name) => new Ast.Value.VarRef(name)),
    (($, name) => name),
    (($, prefix, $0, name) => prefix + '.' + name),
    (($, $0) => { return $0; }),
    (($, field, $0, value) => new Ast.Value.ArrayField(value, field)),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, value, $0, filter) => new Ast.Value.Filter(value, filter)),
    (($, $0) => { return $0; }),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('+', [lhs, rhs])),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('-', [lhs, rhs])),
    (($, $0) => { return $0; }),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('*', [lhs, rhs])),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('/', [lhs, rhs])),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('%', [lhs, rhs])),
    (($, $0) => { return $0; }),
    (($, lhs, $0, rhs) => new Ast.Value.Computation('**', [lhs, rhs])),
    (($, op, $0, operands, $1) => new Ast.Value.Computation(op, operands)),
    (($, op, $0, operands, $1) => new Ast.Value.Computation(op, operands)),
    (($, $0, $1, operands, $2) => new Ast.Value.Computation('count', operands)),
    (($, $0) => new Ast.Value.Undefined(true)),
    (($, $0) => new Ast.Value.Undefined(true)),
    (($, $0) => new Ast.Value.Event(null)),
    (($, $0) => new Ast.Value.Event('type')),
    (($, $0) => new Ast.Value.Event('program_id')),
    (($, $0) => new Ast.Value.Event('source')),
    (($, $0, $1, name, type) => new Ast.Value.ContextRef(name, type)),
    (($, $0, $1) => new Ast.Value.Array([])),
    (($, $0, values, $1) => new Ast.Value.Array(values)),
    (($, $0, $1) => new Ast.Value.Object({})),
    (($, $0, values, $1) => new Ast.Value.Object(values)),
    (($, $0, values, $1, $2) => new Ast.Value.Object(values)),
    (($, name, $0, value) => ({ [name]: value })),
    (($, obj, $0, name, $1, value) => {
        obj[name] = value;
        return obj;
    }),
    (($, $0, $1, $2, rules, $3) => new Ast.Value.RecurrentTimeSpecification(rules)),
    (($, first) => [first]),
    (($, head, $0, tail) => head.concat([tail])),
    (($, $0, list, $1) => {
        const obj = {};
        for (const item of list) {
            if (obj[item[0]])
                return $.error(`Duplicate recurrent time rule key ${item[0]}`);
            obj[item[0]] = item[1];
        }
        return new Ast.RecurrentTimeRule(obj);
    }),
    (($, first) => [first]),
    (($, head, $0, tail) => head.concat([tail])),
    (($, $0, $1, v) => ['beginDate', v.value]),
    (($, $0, $1, v) => ['endDate', v.value]),
    (($, $0, $1, v) => ['beginTime', v]),
    (($, $0, $1, v) => ['endTime', v]),
    (($, $0, $1, v) => ['frequency', v]),
    (($, $0, $1, v) => ['interval', v]),
    (($, $0, $1, v) => ['dayOfWeek', v]),
    (($, $0, $1, $2) => ['subtract', true]),
    (($, $0, $1, $2) => ['subtract', false]),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, slot) => (slot.value === undefined ? new Ast.Value.Undefined(true) : slot.value)),
    (($, value) => [value]),
    (($, array, $0, v) => array.concat([v])),
    (($, $0) => new Ast.Value.Boolean(true)),
    (($, $0) => new Ast.Value.Boolean(false)),
    (($, str) => str.value),
    (($, str) => new Ast.Value.String(str)),
    (($, num) => num.value),
    (($, num) => num),
    (($, $0, num) => -num),
    (($, tok) => new Ast.Value.Measure(tok.value.value, tok.value.unit)),
    (($, tok) => new Ast.Value.Measure(tok.value.value, tok.value.unit)),
    (($, num, unit) => new Ast.Value.Measure(num, unit)),
    (($, num) => new Ast.Value.Number(num)),
    (($, tok) => new Ast.Value.Currency(tok.value.value, tok.value.unit)),
    (($, num, tok) => new Ast.Value.Currency(num, tok.value)),
    (($, loc) => new Ast.Value.Location(new Ast.Location.Absolute(loc.value.latitude, loc.value.longitude, loc.value.display || null))),
    (($, $0, $1, name) => new Ast.Value.Location(new Ast.Location.Relative(name))),
    (($, $0, $1, $2, lat, $3, lon, $4, display, $5) => {
        return new Ast.Value.Location(new Ast.Location.Absolute(lat, lon, display));
    }),
    (($, $0, $1, $2, lat, $3, lon, $4) => {
        return new Ast.Value.Location(new Ast.Location.Absolute(lat, lon, null));
    }),
    (($, $0, $1, $2, name, $3) => {
        return new Ast.Value.Location(new Ast.Location.Unresolved(name));
    }),
    (($, time) => new Ast.Time.Absolute(time.value.hour, time.value.minute, time.value.second || 0)),
    (($, $0, $1, $2, hour, $3, minute, $4, second, $5) => new Ast.Time.Absolute(hour, minute, second)),
    (($, $0, $1, $2, hour, $3, minute, $4, $5) => new Ast.Time.Absolute(hour, minute, 0)),
    (($, $0, $1, $2, hour, $3, minute, $4) => new Ast.Time.Absolute(hour, minute, 0)),
    (($, time) => new Ast.Value.Time(time)),
    (($, $0, $1, name) => new Ast.Value.Time(new Ast.Time.Relative(name))),
    (($, $0) => new Ast.Value.Date(null)),
    (($, $0, $1, $2, $3) => new Ast.Value.Date(null)),
    (($, $0) => { return $0; }),
    (($, num) => [num]),
    (($, $0) => [',']),
    (($, list, $0) => {
        list.push(',');
        return list;
    }),
    (($, list, num) => {
        list.push(num);
        return list;
    }),
    (($, abs) => new Ast.Value.Date(date_utils_1.parseDate(abs.value))),
    (($, $0, $1, $2, iso, $3) => {
        return new Ast.Value.Date(new Date(iso));
    }),
    (($, $0, $1, $2, params, $3) => parseIncompleteDate($, params, null)),
    (($, $0, $1, $2, params, $3, time, $4) => parseIncompleteDate($, params, time)),
    (($, $0, $1, $2, weekday, $3) => {
        const d = new Ast.WeekDayDate(weekday, null);
        return new Ast.Value.Date(d);
    }),
    (($, $0, $1, $2, weekday, $3, time, $4) => {
        const d = new Ast.WeekDayDate(weekday, time);
        return new Ast.Value.Date(d);
    }),
    (($, $0, $1, $2, weekday, $3, hours, $4, minutes, $5, seconds, $6) => {
        const d = new Ast.WeekDayDate(weekday, new Ast.Time.Absolute(hours.value, minutes.value, seconds.value));
        return new Ast.Value.Date(d);
    }),
    (($, $0, $1, $2, weekday, $3, hours, $4, minutes, $5, $6) => {
        const d = new Ast.WeekDayDate(weekday, new Ast.Time.Absolute(hours.value, minutes.value, 0));
        return new Ast.Value.Date(d);
    }),
    (($, $0, $1, $2, weekday, $3, hours, $4, minutes, $5) => {
        const d = new Ast.WeekDayDate(weekday, new Ast.Time.Absolute(hours.value, minutes.value, 0));
        return new Ast.Value.Date(d);
    }),
    (($, $0, $1, unit, $2) => new Ast.Value.Date(new Ast.DateEdge('start_of', unit))),
    (($, $0, $1, unit, $2) => new Ast.Value.Date(new Ast.DateEdge('end_of', unit))),
    (($, v) => new Ast.Value.Enum(v)),
    (($, $0, $1, venum, $2) => venum),
    (($, $0, venum) => venum),
    (($, $0, entity, $1, display, $2) => new Ast.Value.Entity(null, entity.value, display)),
    (($, str, entity, $0, display, $1) => new Ast.Value.Entity(str, entity.value, display)),
    (($, str, entity) => new Ast.Value.Entity(str, entity.value, null)),
    (($, entity) => new Ast.Value.Entity(entity.value.value, entity.value.type, entity.value.display)),
    (($, entity) => new Ast.Value.Entity(entity.value, 'tt:username', null)),
    (($, entity) => new Ast.Value.Entity(entity.value, 'tt:hashtag', null)),
    (($, entity) => new Ast.Value.Entity(entity.value, 'tt:url', null)),
    (($, entity) => new Ast.Value.Entity(entity.value, 'tt:phone_number', null)),
    (($, entity) => new Ast.Value.Entity(entity.value, 'tt:email_address', null)),
    (($, entity) => new Ast.Value.Entity(entity.value, 'tt:path_name', null)),
    (($, entity) => new Ast.Value.Entity(entity.value, 'tt:picture', null)),
    (($, className) => new Ast.Value.Entity(className.value, 'tt:device', null)),
    (($, className, $0, name) => new Ast.Value.Entity(className.value + '.' + name, 'tt:function', null)),
    (($, $0, $1, $2, $3) => new Ast.Value.ArgMap({})),
    (($, $0, $1, $2, argmap_list, $3) => new Ast.Value.ArgMap(argmap_list)),
    (($, name, type) => ({ [name]: type })),
    (($, list, $0, name, type) => {
        list[name] = type;
        return list;
    }),
    (($, $0, type_ref) => type_ref),
    (($, type) => type.value),
    (($, $0) => type_1.default.ArgMap),
    (($, $0) => type_1.default.Boolean),
    (($, $0) => type_1.default.Currency),
    (($, $0) => type_1.default.Date),
    (($, $0) => type_1.default.Location),
    (($, $0) => type_1.default.Number),
    (($, $0) => type_1.default.Time),
    (($, $0) => type_1.default.String),
    (($, $0) => type_1.default.Object),
    (($, $0) => type_1.default.RecurrentTimeSpecification),
    (($, name) => new type_1.default.Unknown(name.value)),
    (($, $0, $1, elem, $2) => new type_1.default.Array(elem)),
    (($, $0, $1, name, $2) => new type_1.default.Entity(name.value)),
    (($, $0, $1, name, $2) => new type_1.default.Measure(name)),
    (($, $0, $1, entries, $2) => new type_1.default.Enum(entries)),
    (($, $0, fields, $1) => new type_1.default.Compound(null, fields)),
    (($, list) => {
        const fields = {};
        for (const arg of list)
            fields[arg.name] = arg;
        return fields;
    }),
    (($, list, $0) => {
        const fields = {};
        for (const arg of list)
            fields[arg.name] = arg;
        return fields;
    }),
    (($, field) => [field]),
    (($, list, $0, field) => {
        list.push(field);
        return list;
    }),
    (($, name, type) => new Ast.ArgumentDef($.location, null, name, type, {})),
    (($, name, type, annot) => new Ast.ArgumentDef($.location, null, name, type, annot)),
    (($, v) => [v]),
    (($, enum_list, $0, v) => enum_list.concat([v])),
    (($, name) => [name]),
    (($, list, $0, name) => list.concat([name])),
    (($, $0) => { return "error"; }),
    (($, $0) => { return "confirm"; }),
    (($, $0) => { return "more"; }),
    (($, $0) => { return "results"; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return "action"; }),
    (($, $0) => { return "entity"; }),
    (($, $0) => { return "from"; }),
    (($, $0) => { return "language"; }),
    (($, $0) => { return "list"; }),
    (($, $0) => { return "monitorable"; }),
    (($, $0) => { return "program"; }),
    (($, $0) => { return "query"; }),
    (($, $0) => { return "stream"; }),
    (($, $0) => { return "beginDate"; }),
    (($, $0) => { return "beginTime"; }),
    (($, $0) => { return "dayOfWeek"; }),
    (($, $0) => { return "endDate"; }),
    (($, $0) => { return "endTime"; }),
    (($, $0) => { return "frequency"; }),
    (($, $0) => { return "interval"; }),
    (($, $0) => { return "subtract"; }),
    (($, $0) => { return "asc"; }),
    (($, $0) => { return "desc"; }),
    (($, name) => name.value),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return "count"; }),
    (($, name) => name.value),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, name) => name.value),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return $0; }),
    (($, $0) => { return "count"; }),
    (($, name) => name.value),
    (($, $0) => { return $0; }),
    (($, name) => [name]),
    (($, list, $0, name) => {
        list.push(name);
        return list;
    }),
    (($, name) => name.value),
    (($, $0) => 'min'),
    (($, $0) => 'in'),
    (($, $0) => { return $0; }),
    (($, $0) => { return "class"; }),
    (($, $0) => { return "enum"; }),
    (($, $0) => { return "extends"; }),
    (($, $0) => { return "import"; }),
    (($, $0) => { return "in"; }),
    (($, $0) => { return "let"; }),
    (($, $0) => { return "new"; }),
    (($, $0) => { return "null"; }),
    (($, $0) => { return "of"; }),
    (($, $0) => { return "true"; }),
    (($, $0) => { return "false"; }),
    (($, $0) => { return "abstract"; }),
    (($, $0) => { return "all"; }),
    (($, $0) => { return "any"; }),
    (($, $0) => { return "as"; }),
    (($, $0) => { return "dataset"; }),
    (($, $0) => { return "filter"; }),
    (($, $0) => { return "function"; }),
    (($, $0) => { return "mixin"; }),
    (($, $0) => { return "monitor"; }),
    (($, $0) => { return "notify"; }),
    (($, $0) => { return "now"; }),
    (($, $0) => { return "out"; }),
    (($, $0) => { return "opt"; }),
    (($, $0) => { return "req"; }),
    (($, $0) => { return "sort"; }),
    (($, $0) => { return "await"; }),
    (($, $0) => { return "break"; }),
    (($, $0) => { return "case"; }),
    (($, $0) => { return "catch"; }),
    (($, $0) => { return "const"; }),
    (($, $0) => { return "continue"; }),
    (($, $0) => { return "debugger"; }),
    (($, $0) => { return "delete"; }),
    (($, $0) => { return "do"; }),
    (($, $0) => { return "export"; }),
    (($, $0) => { return "finally"; }),
    (($, $0) => { return "for"; }),
    (($, $0) => { return "if"; }),
    (($, $0) => { return "implements"; }),
    (($, $0) => { return "instanceof"; }),
    (($, $0) => { return "interface"; }),
    (($, $0) => { return "package"; }),
    (($, $0) => { return "private"; }),
    (($, $0) => { return "protected"; }),
    (($, $0) => { return "public"; }),
    (($, $0) => { return "return"; }),
    (($, $0) => { return "static"; }),
    (($, $0) => { return "super"; }),
    (($, $0) => { return "switch"; }),
    (($, $0) => { return "this"; }),
    (($, $0) => { return "throw"; }),
    (($, $0) => { return "try"; }),
    (($, $0) => { return "typeof"; }),
    (($, $0) => { return "var"; }),
    (($, $0) => { return "void"; }),
    (($, $0) => { return "while"; }),
    (($, $0) => { return "with"; }),
    (($, $0) => { return "yield"; }),
    (($, $0) => { return "aggregate"; }),
    (($, $0) => { return "bookkeeping"; }),
    (($, $0) => { return "compute"; }),
    (($, $0) => { return "edge"; }),
    (($, $0) => { return "join"; }),
    (($, $0) => { return "oninput"; }),
    (($, $0) => { return "Any"; }),
    (($, $0) => { return "ArgMap"; }),
    (($, $0) => { return "Array"; }),
    (($, $0) => { return "Boolean"; }),
    (($, $0) => { return "Compound"; }),
    (($, $0) => { return "Currency"; }),
    (($, $0) => { return "Date"; }),
    (($, $0) => { return "Entity"; }),
    (($, $0) => { return "Enum"; }),
    (($, $0) => { return "Location"; }),
    (($, $0) => { return "Measure"; }),
    (($, $0) => { return "Number"; }),
    (($, $0) => { return "Object"; }),
    (($, $0) => { return "RecurrentTimeSpecification"; }),
    (($, $0) => { return "String"; }),
    (($, $0) => { return "Time"; }),
    (($, $0) => $0),
];
const $runtime = __importStar(require("./../utils/sr_parser_runtime"));
exports.default = $runtime.createParser({ TERMINAL_IDS, RULE_NON_TERMINALS, ARITY, GOTO, PARSER_ACTION, SEMANTIC_ACTION });
//# sourceMappingURL=parser.js.map